/* Generated by Cython 0.29.23 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "extra_compile_args": [
            "-std=c++17",
            "-O3",
            "-Wno-deprecated-declarations",
            "-Wno-unreachable-code",
            "-Wno-unused-function"
        ],
        "extra_link_args": [
            "-std=c++17",
            "-lz",
            "-llz4"
        ],
        "language": "c++",
        "name": "fastwarc.warc",
        "sources": [
            "fastwarc/warc.pyx"
        ]
    },
    "module_name": "fastwarc.warc"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_23"
#define CYTHON_HEX_VERSION 0x001D17F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__fastwarc__warc
#define __PYX_HAVE_API__fastwarc__warc
/* Early includes */
#include <stdint.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <vector>
#include <string>
#include <string.h>
#include <stdio.h>
#include <zlib.h>
#include <lz4hc.h>
#include <lz4frame.h>
#include <string_view>
#include <cctype>
#include <cstdlib>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "fastwarc/warc.pyx",
  "stringsource",
  "fastwarc/stream_io.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/*--- Type declarations ---*/
struct __pyx_obj_8fastwarc_9stream_io_IOStream;
struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter;
struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream;
struct __pyx_obj_8fastwarc_9stream_io_FileStream;
struct __pyx_obj_8fastwarc_9stream_io_CompressingStream;
struct __pyx_obj_8fastwarc_9stream_io_GZipStream;
struct __pyx_obj_8fastwarc_9stream_io_LZ4Stream;
struct __pyx_obj_8fastwarc_9stream_io_BufferedReader;
struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap;
struct __pyx_obj_8fastwarc_4warc_WarcRecord;
struct __pyx_obj_8fastwarc_4warc_ArchiveIterator;
struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__;
struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__;
struct __pyx_obj___Pyx_EnumMeta;
struct __pyx_opt_args_8fastwarc_9stream_io_10FileStream_open;
struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_read;
struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_readline;
struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_consume;

/* "fastwarc/stream_io.pxd":96
 *     cdef FILE* fp
 * 
 *     cdef void open(self, const char* path, char* mode=*) except *             # <<<<<<<<<<<<<<
 *     cdef void seek(self, size_t offset)
 * 
 */
struct __pyx_opt_args_8fastwarc_9stream_io_10FileStream_open {
  int __pyx_n;
  char *mode;
};

/* "fastwarc/stream_io.pxd":150
 *     cdef void detect_stream_type(self)
 * 
 *     cpdef string read(self, size_t size=*)             # <<<<<<<<<<<<<<
 *     cpdef string readline(self, bint crlf=*, size_t max_line_len=*)
 *     cpdef size_t tell(self)
 */
struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_read {
  int __pyx_n;
  size_t size;
};

/* "fastwarc/stream_io.pxd":151
 * 
 *     cpdef string read(self, size_t size=*)
 *     cpdef string readline(self, bint crlf=*, size_t max_line_len=*)             # <<<<<<<<<<<<<<
 *     cpdef size_t tell(self)
 *     cpdef void consume(self, size_t size=*)
 */
struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_readline {
  int __pyx_n;
  int crlf;
  size_t max_line_len;
};

/* "fastwarc/stream_io.pxd":153
 *     cpdef string readline(self, bint crlf=*, size_t max_line_len=*)
 *     cpdef size_t tell(self)
 *     cpdef void consume(self, size_t size=*)             # <<<<<<<<<<<<<<
 *     cpdef void close(self)
 *     cpdef string error(self)
 */
struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_consume {
  int __pyx_n;
  size_t size;
};
struct __pyx_ctuple_std__in_string__and_std__in_string;
typedef struct __pyx_ctuple_std__in_string__and_std__in_string __pyx_ctuple_std__in_string__and_std__in_string;
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers;
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write;
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest;
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest;
struct __pyx_opt_args_8fastwarc_4warc_parse_header_block;

/* "fastwarc/warc.pxd":22
 * 
 * 
 * cpdef enum WarcRecordType:             # <<<<<<<<<<<<<<
 *     warcinfo = 2,
 *     response = 4,
 */
enum __pyx_t_8fastwarc_4warc_WarcRecordType {
  __pyx_e_8fastwarc_4warc_warcinfo = 2,
  __pyx_e_8fastwarc_4warc_response = 4,
  __pyx_e_8fastwarc_4warc_resource = 8,
  __pyx_e_8fastwarc_4warc_request = 16,
  __pyx_e_8fastwarc_4warc_metadata = 32,
  __pyx_e_8fastwarc_4warc_revisit = 64,
  __pyx_e_8fastwarc_4warc_conversion = 0x80,
  __pyx_e_8fastwarc_4warc_continuation = 0x100,
  __pyx_e_8fastwarc_4warc_unknown = 0x200,
  __pyx_e_8fastwarc_4warc_any_type = 0xFFFF,
  __pyx_e_8fastwarc_4warc_no_type = 0
};

/* "fastwarc/warc.pxd":55
 * 
 * 
 * cdef enum _NextRecStatus:             # <<<<<<<<<<<<<<
 *     has_next,
 *     skip_next,
 */
enum __pyx_t_8fastwarc_4warc__NextRecStatus {
  __pyx_e_8fastwarc_4warc_has_next,
  __pyx_e_8fastwarc_4warc_skip_next,
  __pyx_e_8fastwarc_4warc_eof,
  __pyx_e_8fastwarc_4warc_error
};

/* "fastwarc/warc.pxd":36
 * 
 * 
 * ctypedef (string, string) str_pair             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_ctuple_std__in_string__and_std__in_string {
  std::string f0;
  std::string f1;
};
typedef __pyx_ctuple_std__in_string__and_std__in_string __pyx_t_8fastwarc_4warc_str_pair;

/* "fastwarc/warc.pxd":73
 *     cdef size_t _stream_pos
 * 
 *     cpdef void init_headers(self, size_t content_length, WarcRecordType record_type=*, bytes record_urn=*)             # <<<<<<<<<<<<<<
 *     cpdef void set_bytes_content(self, bytes b)
 *     cpdef void parse_http(self)
 */
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers {
  int __pyx_n;
  enum __pyx_t_8fastwarc_4warc_WarcRecordType record_type;
  PyObject *record_urn;
};

/* "fastwarc/warc.pxd":76
 *     cpdef void set_bytes_content(self, bytes b)
 *     cpdef void parse_http(self)
 *     cpdef size_t write(self, stream, bint checksum_data=*, size_t chunk_size=*)             # <<<<<<<<<<<<<<
 *     cpdef bint verify_block_digest(self, bint consume=*)
 *     cpdef bint verify_payload_digest(self, bint consume=*)
 */
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write {
  int __pyx_n;
  int checksum_data;
  size_t chunk_size;
};

/* "fastwarc/warc.pxd":77
 *     cpdef void parse_http(self)
 *     cpdef size_t write(self, stream, bint checksum_data=*, size_t chunk_size=*)
 *     cpdef bint verify_block_digest(self, bint consume=*)             # <<<<<<<<<<<<<<
 *     cpdef bint verify_payload_digest(self, bint consume=*)
 * 
 */
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest {
  int __pyx_n;
  int consume;
};

/* "fastwarc/warc.pxd":78
 *     cpdef size_t write(self, stream, bint checksum_data=*, size_t chunk_size=*)
 *     cpdef bint verify_block_digest(self, bint consume=*)
 *     cpdef bint verify_payload_digest(self, bint consume=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume)
 */
struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest {
  int __pyx_n;
  int consume;
};

/* "fastwarc/warc.pyx":596
 * 
 * # noinspection PyProtectedMember
 * cdef size_t parse_header_block(BufferedReader reader, WarcHeaderMap target, bint has_status_line=False):             # <<<<<<<<<<<<<<
 *     """
 *     Helper function for parsing WARC or HTTP header blocks.
 */
struct __pyx_opt_args_8fastwarc_4warc_parse_header_block {
  int __pyx_n;
  int has_status_line;
};

/* "fastwarc/stream_io.pxd":24
 * 
 * 
 * cdef class IOStream:             # <<<<<<<<<<<<<<
 *     cdef string errstr
 * 
 */
struct __pyx_obj_8fastwarc_9stream_io_IOStream {
  PyObject_HEAD
  struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *__pyx_vtab;
  std::string errstr;
};


/* "fastwarc/stream_io.pxd":36
 * 
 * # noinspection PyAttributeOutsideInit
 * cdef class PythonIOStreamAdapter(IOStream):             # <<<<<<<<<<<<<<
 *     """IOStream adapter for Python file-like objects."""
 * 
 */
struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter {
  struct __pyx_obj_8fastwarc_9stream_io_IOStream __pyx_base;
  PyObject *py_stream;
  PyObject *exc;
};


/* "fastwarc/stream_io.pxd":85
 * 
 * 
 * cdef class BytesIOStream(IOStream):             # <<<<<<<<<<<<<<
 *     cdef string buffer
 *     cdef size_t pos
 */
struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream {
  struct __pyx_obj_8fastwarc_9stream_io_IOStream __pyx_base;
  std::string buffer;
  size_t pos;
};


/* "fastwarc/stream_io.pxd":93
 * 
 * 
 * cdef class FileStream(IOStream):             # <<<<<<<<<<<<<<
 *     cdef FILE* fp
 * 
 */
struct __pyx_obj_8fastwarc_9stream_io_FileStream {
  struct __pyx_obj_8fastwarc_9stream_io_IOStream __pyx_base;
  FILE *fp;
};


/* "fastwarc/stream_io.pxd":100
 * 
 * 
 * cdef class CompressingStream(IOStream):             # <<<<<<<<<<<<<<
 *     cdef size_t begin_member(self)
 *     cdef size_t end_member(self)
 */
struct __pyx_obj_8fastwarc_9stream_io_CompressingStream {
  struct __pyx_obj_8fastwarc_9stream_io_IOStream __pyx_base;
};


/* "fastwarc/stream_io.pxd":105
 * 
 * 
 * cdef class GZipStream(CompressingStream):             # <<<<<<<<<<<<<<
 *     cdef IOStream raw_stream
 *     cdef string working_buf
 */
struct __pyx_obj_8fastwarc_9stream_io_GZipStream {
  struct __pyx_obj_8fastwarc_9stream_io_CompressingStream __pyx_base;
  struct __pyx_obj_8fastwarc_9stream_io_IOStream *raw_stream;
  std::string working_buf;
  z_stream zst;
  int stream_read_status;
  char initialized;
  size_t stream_pos;
  int member_started;
  int compression_level;
};


/* "fastwarc/stream_io.pxd":121
 * 
 * 
 * cdef class LZ4Stream(CompressingStream):             # <<<<<<<<<<<<<<
 *     cdef IOStream raw_stream
 *     cdef LZ4F_cctx* cctx
 */
struct __pyx_obj_8fastwarc_9stream_io_LZ4Stream {
  struct __pyx_obj_8fastwarc_9stream_io_CompressingStream __pyx_base;
  struct __pyx_obj_8fastwarc_9stream_io_IOStream *raw_stream;
  LZ4F_cctx *cctx;
  LZ4F_dctx *dctx;
  LZ4F_preferences_t prefs;
  std::string working_buf;
  int frame_started;
  size_t stream_pos;
};


/* "fastwarc/stream_io.pxd":134
 * 
 * 
 * cdef class BufferedReader:             # <<<<<<<<<<<<<<
 *     cdef IOStream stream
 *     cdef string errstr
 */
struct __pyx_obj_8fastwarc_9stream_io_BufferedReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *__pyx_vtab;
  struct __pyx_obj_8fastwarc_9stream_io_IOStream *stream;
  std::string errstr;
  std::string buf;
  size_t buf_size;
  size_t stream_pos;
  size_t limit;
  size_t limit_consumed;
  int negotiate_stream;
  int stream_started;
  int stream_is_compressed;
};


/* "fastwarc/warc.pxd":39
 * 
 * 
 * cdef class WarcHeaderMap:             # <<<<<<<<<<<<<<
 *     cdef string _status_line
 *     cdef vector[str_pair] _headers
 */
struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap {
  PyObject_HEAD
  struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *__pyx_vtab;
  std::string _status_line;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair>  _headers;
  PyObject *_enc;
  PyObject *_dict_cache;
  int _dict_cache_stale;
};


/* "fastwarc/warc.pxd":62
 * 
 * 
 * cdef class WarcRecord:             # <<<<<<<<<<<<<<
 *     cdef WarcRecordType _record_type
 *     cdef WarcHeaderMap _headers
 */
struct __pyx_obj_8fastwarc_4warc_WarcRecord {
  PyObject_HEAD
  struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *__pyx_vtab;
  enum __pyx_t_8fastwarc_4warc_WarcRecordType _record_type;
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *_headers;
  int _is_http;
  int _http_parsed;
  std::string _http_charset;
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *_http_headers;
  size_t _content_length;
  struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *_reader;
  size_t _stream_pos;
};


/* "fastwarc/warc.pxd":84
 * 
 * 
 * cdef class ArchiveIterator:             # <<<<<<<<<<<<<<
 *     cdef BufferedReader reader
 *     cdef WarcRecord record
 */
struct __pyx_obj_8fastwarc_4warc_ArchiveIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_8fastwarc_4warc_ArchiveIterator *__pyx_vtab;
  struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *reader;
  struct __pyx_obj_8fastwarc_4warc_WarcRecord *record;
  int parse_http;
  int verify_digests;
  size_t min_content_length;
  size_t max_content_length;
  PyObject *func_filter;
  uint16_t record_type_filter;
};


/* "fastwarc/warc.pyx":164
 *                         header_value.encode(self._enc, errors='ignore'))
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield from self.items()
 * 
 */
struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ {
  PyObject_HEAD
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self;
};


/* "fastwarc/warc.pyx":683
 *         self.record_type_filter = record_types
 * 
 *     def __iter__(self) -> Iterable[WarcRecord]:             # <<<<<<<<<<<<<<
 *         cdef _NextRecStatus status
 * 
 */
struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ {
  PyObject_HEAD
  struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self;
  enum __pyx_t_8fastwarc_4warc__NextRecStatus __pyx_v_status;
};


/* "EnumBase":15
 * 
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):             # <<<<<<<<<<<<<<
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 */
struct __pyx_obj___Pyx_EnumMeta {
  PyHeapTypeObject __pyx_base;
};



/* "fastwarc/stream_io.pxd":24
 * 
 * 
 * cdef class IOStream:             # <<<<<<<<<<<<<<
 *     cdef string errstr
 * 
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream {
  std::string (*read)(struct __pyx_obj_8fastwarc_9stream_io_IOStream *, size_t);
  size_t (*write)(struct __pyx_obj_8fastwarc_9stream_io_IOStream *, char const *, size_t);
  size_t (*tell)(struct __pyx_obj_8fastwarc_9stream_io_IOStream *);
  void (*flush)(struct __pyx_obj_8fastwarc_9stream_io_IOStream *);
  void (*close)(struct __pyx_obj_8fastwarc_9stream_io_IOStream *);
  std::string (*error)(struct __pyx_obj_8fastwarc_9stream_io_IOStream *);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *__pyx_vtabptr_8fastwarc_9stream_io_IOStream;


/* "fastwarc/stream_io.pxd":36
 * 
 * # noinspection PyAttributeOutsideInit
 * cdef class PythonIOStreamAdapter(IOStream):             # <<<<<<<<<<<<<<
 *     """IOStream adapter for Python file-like objects."""
 * 
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_PythonIOStreamAdapter {
  struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream __pyx_base;
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_vtabptr_8fastwarc_9stream_io_PythonIOStreamAdapter;


/* "fastwarc/stream_io.pxd":85
 * 
 * 
 * cdef class BytesIOStream(IOStream):             # <<<<<<<<<<<<<<
 *     cdef string buffer
 *     cdef size_t pos
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream {
  struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream __pyx_base;
  void (*seek)(struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *, size_t);
  std::string (*getvalue)(struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *__pyx_vtabptr_8fastwarc_9stream_io_BytesIOStream;


/* "fastwarc/stream_io.pxd":93
 * 
 * 
 * cdef class FileStream(IOStream):             # <<<<<<<<<<<<<<
 *     cdef FILE* fp
 * 
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_FileStream {
  struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream __pyx_base;
  void (*open)(struct __pyx_obj_8fastwarc_9stream_io_FileStream *, char const *, struct __pyx_opt_args_8fastwarc_9stream_io_10FileStream_open *__pyx_optional_args);
  void (*seek)(struct __pyx_obj_8fastwarc_9stream_io_FileStream *, size_t);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_FileStream *__pyx_vtabptr_8fastwarc_9stream_io_FileStream;


/* "fastwarc/stream_io.pxd":100
 * 
 * 
 * cdef class CompressingStream(IOStream):             # <<<<<<<<<<<<<<
 *     cdef size_t begin_member(self)
 *     cdef size_t end_member(self)
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream {
  struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream __pyx_base;
  size_t (*begin_member)(struct __pyx_obj_8fastwarc_9stream_io_CompressingStream *);
  size_t (*end_member)(struct __pyx_obj_8fastwarc_9stream_io_CompressingStream *);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream *__pyx_vtabptr_8fastwarc_9stream_io_CompressingStream;


/* "fastwarc/stream_io.pxd":105
 * 
 * 
 * cdef class GZipStream(CompressingStream):             # <<<<<<<<<<<<<<
 *     cdef IOStream raw_stream
 *     cdef string working_buf
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_GZipStream {
  struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream __pyx_base;
  void (*prepopulate)(struct __pyx_obj_8fastwarc_9stream_io_GZipStream *, int, std::string const &);
  void (*_init_z_stream)(struct __pyx_obj_8fastwarc_9stream_io_GZipStream *, int);
  void (*_free_z_stream)(struct __pyx_obj_8fastwarc_9stream_io_GZipStream *);
  int (*_refill_working_buf)(struct __pyx_obj_8fastwarc_9stream_io_GZipStream *, size_t);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_GZipStream *__pyx_vtabptr_8fastwarc_9stream_io_GZipStream;


/* "fastwarc/stream_io.pxd":121
 * 
 * 
 * cdef class LZ4Stream(CompressingStream):             # <<<<<<<<<<<<<<
 *     cdef IOStream raw_stream
 *     cdef LZ4F_cctx* cctx
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_LZ4Stream {
  struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream __pyx_base;
  void (*prepopulate)(struct __pyx_obj_8fastwarc_9stream_io_LZ4Stream *, std::string const &);
  void (*_free_ctx)(struct __pyx_obj_8fastwarc_9stream_io_LZ4Stream *);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_LZ4Stream *__pyx_vtabptr_8fastwarc_9stream_io_LZ4Stream;


/* "fastwarc/stream_io.pxd":134
 * 
 * 
 * cdef class BufferedReader:             # <<<<<<<<<<<<<<
 *     cdef IOStream stream
 *     cdef string errstr
 */

struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader {
  void (*set_limit)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, size_t);
  void (*reset_limit)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *);
  void (*detect_stream_type)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *);
  std::string (*read)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_read *__pyx_optional_args);
  std::string (*readline)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_readline *__pyx_optional_args);
  size_t (*tell)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, int __pyx_skip_dispatch);
  void (*consume)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_consume *__pyx_optional_args);
  void (*close)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, int __pyx_skip_dispatch);
  std::string (*error)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, int __pyx_skip_dispatch);
  int (*_fill_buf)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *);
  std::string_view (*_get_buf)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *);
  void (*_consume_buf)(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, size_t);
};
static struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *__pyx_vtabptr_8fastwarc_9stream_io_BufferedReader;


/* "fastwarc/warc.pyx":143
 * 
 * # noinspection PyAttributeOutsideInit
 * cdef class WarcHeaderMap:             # <<<<<<<<<<<<<<
 *     """Dict-like type representing a WARC or HTTP header block"""
 * 
 */

struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap {
  size_t (*write)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, struct __pyx_obj_8fastwarc_9stream_io_IOStream *);
  void (*clear)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *);
  void (*set_status_line)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &);
  std::string (*find_header)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &);
  void (*set_header)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &);
  void (*append_header)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &);
  void (*add_continuation)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &);
};
static struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *__pyx_vtabptr_8fastwarc_4warc_WarcHeaderMap;
static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_clear(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *);
static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &);
static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &);


/* "fastwarc/warc.pyx":296
 * 
 * # noinspection PyProtectedMember, PyAttributeOutsideInit
 * cdef class WarcRecord:             # <<<<<<<<<<<<<<
 *     """A WARC record."""
 * 
 */

struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord {
  void (*init_headers)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, size_t, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers *__pyx_optional_args);
  void (*set_bytes_content)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, PyObject *, int __pyx_skip_dispatch);
  void (*parse_http)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch);
  size_t (*write)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write *__pyx_optional_args);
  int (*verify_block_digest)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest *__pyx_optional_args);
  int (*verify_payload_digest)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest *__pyx_optional_args);
  int (*_verify_digest)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, std::string const &, int);
  size_t (*_write_impl)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, PyObject *, PyObject *, int, size_t);
};
static struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *__pyx_vtabptr_8fastwarc_4warc_WarcRecord;


/* "fastwarc/warc.pyx":647
 * # noinspection PyProtectedMember, PyAttributeOutsideInit
 * @cython.auto_pickle(False)
 * cdef class ArchiveIterator:             # <<<<<<<<<<<<<<
 *     """WARC record stream iterator."""
 * 
 */

struct __pyx_vtabstruct_8fastwarc_4warc_ArchiveIterator {
  enum __pyx_t_8fastwarc_4warc__NextRecStatus (*_read_next_record)(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *);
  void (*_set_stream)(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_8fastwarc_4warc_ArchiveIterator *__pyx_vtabptr_8fastwarc_4warc_ArchiveIterator;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* GeneratorYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Generator_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* IncludeCppStringH.proto */
#include <string>

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_cpp_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_cpp_string(
         std::string cppstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        cppstring.data(), cppstring.size(), start, stop, encoding, errors, decode_func);
}

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    return (PyObject*) (((PyTypeObject*)type_obj)->tp_new((PyTypeObject*)type_obj, args, kwargs));
}

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CyFunctionClassCell.proto */
static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_8fastwarc_4warc_WarcRecordType __Pyx_PyInt_As_enum____pyx_t_8fastwarc_4warc_WarcRecordType(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(enum __pyx_t_8fastwarc_4warc_WarcRecordType value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static size_t __pyx_f_8fastwarc_4warc_13WarcHeaderMap_write(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, struct __pyx_obj_8fastwarc_9stream_io_IOStream *__pyx_v_stream); /* proto*/
static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_clear(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto*/
static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_status_line); /* proto*/
static std::string __pyx_f_8fastwarc_4warc_13WarcHeaderMap_find_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_key, std::string const &__pyx_v_default); /* proto*/
static void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_key, std::string const &__pyx_v_header_value); /* proto*/
static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_key, std::string const &__pyx_v_header_value); /* proto*/
static void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_add_continuation(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_continuation_value); /* proto*/
static void __pyx_f_8fastwarc_4warc_10WarcRecord_init_headers(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, size_t __pyx_v_content_length, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers *__pyx_optional_args); /* proto*/
static void __pyx_f_8fastwarc_4warc_10WarcRecord_set_bytes_content(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_b, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_8fastwarc_4warc_10WarcRecord_parse_http(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static size_t __pyx_f_8fastwarc_4warc_10WarcRecord_write(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_stream, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write *__pyx_optional_args); /* proto*/
static size_t __pyx_f_8fastwarc_4warc_10WarcRecord__write_impl(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_in_reader, PyObject *__pyx_v_out_stream, int __pyx_v_write_payload_headers, size_t __pyx_v_chunk_size); /* proto*/
static int __pyx_f_8fastwarc_4warc_10WarcRecord__verify_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, std::string const &__pyx_v_base32_digest, int __pyx_v_consume); /* proto*/
static int __pyx_f_8fastwarc_4warc_10WarcRecord_verify_block_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest *__pyx_optional_args); /* proto*/
static int __pyx_f_8fastwarc_4warc_10WarcRecord_verify_payload_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest *__pyx_optional_args); /* proto*/
static enum __pyx_t_8fastwarc_4warc__NextRecStatus __pyx_f_8fastwarc_4warc_15ArchiveIterator__read_next_record(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self); /* proto*/
static void __pyx_f_8fastwarc_4warc_15ArchiveIterator__set_stream(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, PyObject *__pyx_v_stream, int __pyx_skip_dispatch); /* proto*/
static size_t __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_tell(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self); /* proto*/
static std::string __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_read(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self, size_t __pyx_v_size); /* proto*/
static size_t __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_write(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self, char const *__pyx_v_data, size_t __pyx_v_size); /* proto*/
static void __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_flush(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self); /* proto*/
static void __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_close(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'resiliparse_inc.string' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'resiliparse_inc.zlib' */

/* Module declarations from 'resiliparse_inc.lz4hc' */

/* Module declarations from 'resiliparse_inc.lz4frame' */

/* Module declarations from 'resiliparse_inc.string_view' */

/* Module declarations from 'fastwarc.stream_io' */
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_IOStream = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_BytesIOStream = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_FileStream = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_CompressingStream = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_GZipStream = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_LZ4Stream = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_9stream_io_BufferedReader = 0;

/* Module declarations from 'cython' */

/* Module declarations from 'resiliparse_inc.cctype' */

/* Module declarations from 'resiliparse_inc.cstdlib' */

/* Module declarations from 'fastwarc.warc' */
static PyTypeObject *__pyx_ptype_8fastwarc_4warc_WarcHeaderMap = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_4warc_WarcRecord = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_4warc_ArchiveIterator = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_4warc___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_ptype_8fastwarc_4warc___pyx_scope_struct_1___iter__ = 0;
static PyTypeObject *__pyx_ptype___Pyx_EnumMeta = 0;
static PyObject *__Pyx_OrderedDict = 0;
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_globals = 0;
static int __pyx_f_8fastwarc_4warc_is_warc_10(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_warc_11(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_8fastwarc_4warc_has_block_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_8fastwarc_4warc_has_payload_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_http(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_concurrent(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch); /*proto*/
static std::string __pyx_f_8fastwarc_4warc_strip_str(std::string const &); /*proto*/
static std::string __pyx_f_8fastwarc_4warc_str_to_lower(std::string); /*proto*/
static char const *__pyx_f_8fastwarc_4warc__enum_record_type_to_str(enum __pyx_t_8fastwarc_4warc_WarcRecordType); /*proto*/
static enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_f_8fastwarc_4warc__str_record_type_to_enum(std::string); /*proto*/
static size_t __pyx_f_8fastwarc_4warc_parse_header_block(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *, struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, struct __pyx_opt_args_8fastwarc_4warc_parse_header_block *__pyx_optional_args); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyStr_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_std__in_string(std::string const &); /*proto*/
static std::string __pyx_convert_string_from_py_std__in_string(PyObject *); /*proto*/
static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "fastwarc.warc"
extern int __pyx_module_is_main_fastwarc__warc;
int __pyx_module_is_main_fastwarc__warc = 0;

/* Implementation of 'fastwarc.warc' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_reversed;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_LookupError;
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_RuntimeWarning;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_[] = " ";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_v[] = "v";
static const char __pyx_k__5[] = "";
static const char __pyx_k__7[] = "<";
static const char __pyx_k__8[] = ">";
static const char __pyx_k_eq[] = "__eq__";
static const char __pyx_k_it[] = "it";
static const char __pyx_k__10[] = "\".";
static const char __pyx_k__17[] = "'.";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_dct[] = "dct";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_md5[] = "md5";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_s_s[] = "%s.%s";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_urn[] = "urn";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_hash[] = "__hash__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_item[] = "item";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sha1[] = "sha1";
static const char __pyx_k_tell[] = "tell";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_uuid[] = "uuid";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_Tuple[] = "Tuple";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_s_s_d[] = "<%s.%s: %d>";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_uuid4[] = "uuid4";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_asdict[] = "asdict";
static const char __pyx_k_base64[] = "base64";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_digest[] = "digest";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_errors[] = "errors";
static const char __pyx_k_ignore[] = "ignore";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reader[] = "reader";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_sha1_2[] = "sha1:";
static const char __pyx_k_sha256[] = "sha256";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_utcnow[] = "utcnow";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_consume[] = "consume";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_getitem[] = "__getitem__";
static const char __pyx_k_hashlib[] = "hashlib";
static const char __pyx_k_isdigit[] = "isdigit";
static const char __pyx_k_members[] = "__members__";
static const char __pyx_k_no_type[] = "no_type";
static const char __pyx_k_parents[] = "parents";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_request[] = "request";
static const char __pyx_k_revisit[] = "revisit";
static const char __pyx_k_setitem[] = "__setitem__";
static const char __pyx_k_unknown[] = "unknown";
static const char __pyx_k_EnumBase[] = "EnumBase";
static const char __pyx_k_EnumType[] = "EnumType";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_Optional[] = "Optional";
static const char __pyx_k_any_type[] = "any_type";
static const char __pyx_k_casefold[] = "casefold";
static const char __pyx_k_contains[] = "__contains__";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_metadata[] = "metadata";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_resource[] = "resource";
static const char __pyx_k_response[] = "response";
static const char __pyx_k_reversed[] = "reversed";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_strftime[] = "strftime";
static const char __pyx_k_warcinfo[] = "warcinfo";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_WARC_Type[] = "WARC-Type";
static const char __pyx_k_b32decode[] = "b32decode";
static const char __pyx_k_b32encode[] = "b32encode";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_WarcRecord[] = "WarcRecord";
static const char __pyx_k_chunk_size[] = "chunk_size";
static const char __pyx_k_conversion[] = "conversion";
static const char __pyx_k_parse_http[] = "parse_http";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_record_urn[] = "record_urn";
static const char __pyx_k_set_stream[] = "_set_stream";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_LookupError[] = "LookupError";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_StreamError[] = "StreamError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_func_filter[] = "func_filter";
static const char __pyx_k_http_parsed[] = "http_parsed";
static const char __pyx_k_iso_8859_15[] = "iso-8859-15";
static const char __pyx_k_record_type[] = "record_type";
static const char __pyx_k_Pyx_EnumBase[] = "__Pyx_EnumBase";
static const char __pyx_k_continuation[] = "continuation";
static const char __pyx_k_init_headers[] = "init_headers";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_record_types[] = "record_types";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_WarcHeaderMap[] = "WarcHeaderMap";
static const char __pyx_k_Y_m_dT_H_M_SZ[] = "%Y-%m-%dT%H:%M:%SZ";
static const char __pyx_k_checksum_data[] = "checksum_data";
static const char __pyx_k_fastwarc_warc[] = "fastwarc.warc";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_Object_of_type[] = "Object of type '";
static const char __pyx_k_RuntimeWarning[] = "RuntimeWarning";
static const char __pyx_k_WARC_Record_ID[] = "WARC-Record-ID";
static const char __pyx_k_WarcRecordType[] = "WarcRecordType";
static const char __pyx_k_content_length[] = "content_length";
static const char __pyx_k_verify_digests[] = "verify_digests";
static const char __pyx_k_ArchiveIterator[] = "ArchiveIterator";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_has_no_attribute[] = "' has no attribute '";
static const char __pyx_k_fastwarc_warc_pyx[] = "fastwarc/warc.pyx";
static const char __pyx_k_set_bytes_content[] = "set_bytes_content";
static const char __pyx_k_CaseInsensitiveStr[] = "CaseInsensitiveStr";
static const char __pyx_k_Pyx_EnumBase___new[] = "__Pyx_EnumBase.__new__";
static const char __pyx_k_Pyx_EnumBase___str[] = "__Pyx_EnumBase.__str__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_fastwarc_stream_io[] = "fastwarc.stream_io";
static const char __pyx_k_max_content_length[] = "max_content_length";
static const char __pyx_k_min_content_length[] = "min_content_length";
static const char __pyx_k_Pyx_EnumBase___repr[] = "__Pyx_EnumBase.__repr__";
static const char __pyx_k_verify_block_digest[] = "verify_block_digest";
static const char __pyx_k_Unknown_enum_value_s[] = "Unknown enum value: '%s'";
static const char __pyx_k_WarcHeaderMap___iter[] = "WarcHeaderMap.__iter__";
static const char __pyx_k_is_not_a_valid_stream[] = "' is not a valid stream.";
static const char __pyx_k_verify_payload_digest[] = "verify_payload_digest";
static const char __pyx_k_ArchiveIterator___iter[] = "ArchiveIterator.__iter__";
static const char __pyx_k_CaseInsensitiveStrDict[] = "CaseInsensitiveStrDict";
static const char __pyx_k_CaseInsensitiveStr___eq[] = "CaseInsensitiveStr.__eq__";
static const char __pyx_k_CaseInsensitiveStr___hash[] = "CaseInsensitiveStr.__hash__";
static const char __pyx_k_Case_insensitive_str_dict[] = "Case-insensitive str dict.";
static const char __pyx_k_CaseInsensitiveStrDict_get[] = "CaseInsensitiveStrDict.get";
static const char __pyx_k_CaseInsensitiveStrDict_pop[] = "CaseInsensitiveStrDict.pop";
static const char __pyx_k_Unsupported_hash_algorithm[] = "Unsupported hash algorithm \"";
static const char __pyx_k_pyx_unpickle___Pyx_EnumMeta[] = "__pyx_unpickle___Pyx_EnumMeta";
static const char __pyx_k_CaseInsensitiveStrDict_update[] = "CaseInsensitiveStrDict.update";
static const char __pyx_k_CaseInsensitiveStrDict___contain[] = "CaseInsensitiveStrDict.__contains__";
static const char __pyx_k_CaseInsensitiveStrDict___getitem[] = "CaseInsensitiveStrDict.__getitem__";
static const char __pyx_k_CaseInsensitiveStrDict___setitem[] = "CaseInsensitiveStrDict.__setitem__";
static const char __pyx_k_CaseInsensitiveStrDict_setdefaul[] = "CaseInsensitiveStrDict.setdefault";
static const char __pyx_k_Case_insensitive_str_implementat[] = "Case insensitive str implementation for use as dict key.";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_kp_b_;
static PyObject *__pyx_n_s_ArchiveIterator;
static PyObject *__pyx_n_s_ArchiveIterator___iter;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_CaseInsensitiveStr;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict___contain;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict___getitem;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict___setitem;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict_get;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict_pop;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict_setdefaul;
static PyObject *__pyx_n_s_CaseInsensitiveStrDict_update;
static PyObject *__pyx_n_s_CaseInsensitiveStr___eq;
static PyObject *__pyx_n_s_CaseInsensitiveStr___hash;
static PyObject *__pyx_kp_s_Case_insensitive_str_dict;
static PyObject *__pyx_kp_s_Case_insensitive_str_implementat;
static PyObject *__pyx_n_s_EnumBase;
static PyObject *__pyx_n_s_EnumType;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_IntEnum;
static PyObject *__pyx_n_s_Iterable;
static PyObject *__pyx_n_s_List;
static PyObject *__pyx_n_s_LookupError;
static PyObject *__pyx_kp_u_None;
static PyObject *__pyx_kp_u_Object_of_type;
static PyObject *__pyx_n_s_Optional;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Pyx_EnumBase;
static PyObject *__pyx_n_s_Pyx_EnumBase___new;
static PyObject *__pyx_n_s_Pyx_EnumBase___repr;
static PyObject *__pyx_n_s_Pyx_EnumBase___str;
static PyObject *__pyx_n_s_RuntimeWarning;
static PyObject *__pyx_n_s_StreamError;
static PyObject *__pyx_n_s_Tuple;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unknown_enum_value_s;
static PyObject *__pyx_kp_u_Unsupported_hash_algorithm;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u_WARC_Record_ID;
static PyObject *__pyx_kp_u_WARC_Type;
static PyObject *__pyx_n_s_WarcHeaderMap;
static PyObject *__pyx_n_s_WarcHeaderMap___iter;
static PyObject *__pyx_n_s_WarcRecord;
static PyObject *__pyx_n_s_WarcRecordType;
static PyObject *__pyx_kp_u_Y_m_dT_H_M_SZ;
static PyObject *__pyx_kp_u__10;
static PyObject *__pyx_kp_u__17;
static PyObject *__pyx_kp_b__5;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_kp_b__7;
static PyObject *__pyx_kp_b__8;
static PyObject *__pyx_n_s_any_type;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_asdict;
static PyObject *__pyx_n_s_b32decode;
static PyObject *__pyx_n_s_b32encode;
static PyObject *__pyx_n_s_base64;
static PyObject *__pyx_n_s_casefold;
static PyObject *__pyx_n_s_checksum_data;
static PyObject *__pyx_n_s_chunk_size;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_u_close;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_consume;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_content_length;
static PyObject *__pyx_n_s_continuation;
static PyObject *__pyx_n_s_conversion;
static PyObject *__pyx_n_s_datetime;
static PyObject *__pyx_n_s_dct;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_digest;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encoding;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_errors;
static PyObject *__pyx_n_s_fastwarc_stream_io;
static PyObject *__pyx_n_s_fastwarc_warc;
static PyObject *__pyx_kp_s_fastwarc_warc_pyx;
static PyObject *__pyx_n_s_flush;
static PyObject *__pyx_n_s_func_filter;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getitem;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_u_has_no_attribute;
static PyObject *__pyx_n_s_hash;
static PyObject *__pyx_n_s_hashlib;
static PyObject *__pyx_n_s_http_parsed;
static PyObject *__pyx_n_u_ignore;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_init_headers;
static PyObject *__pyx_kp_u_is_not_a_valid_stream;
static PyObject *__pyx_n_s_isdigit;
static PyObject *__pyx_kp_u_iso_8859_15;
static PyObject *__pyx_n_s_it;
static PyObject *__pyx_n_s_item;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_u_items;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max_content_length;
static PyObject *__pyx_n_s_md5;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_metadata;
static PyObject *__pyx_n_s_min_content_length;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_no_type;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_other;
static PyObject *__pyx_n_s_parents;
static PyObject *__pyx_n_s_parse_http;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle___Pyx_EnumMeta;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_u_read;
static PyObject *__pyx_n_s_reader;
static PyObject *__pyx_n_s_record_type;
static PyObject *__pyx_n_s_record_types;
static PyObject *__pyx_n_s_record_urn;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_request;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_resource;
static PyObject *__pyx_n_s_response;
static PyObject *__pyx_n_s_reversed;
static PyObject *__pyx_n_s_revisit;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_kp_s_s_s_d;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set_bytes_content;
static PyObject *__pyx_n_s_set_stream;
static PyObject *__pyx_n_s_setdefault;
static PyObject *__pyx_n_s_setitem;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sha1;
static PyObject *__pyx_kp_b_sha1_2;
static PyObject *__pyx_n_s_sha256;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_strftime;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_tell;
static PyObject *__pyx_n_u_tell;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_typing;
static PyObject *__pyx_n_s_unknown;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_urn;
static PyObject *__pyx_n_s_utcnow;
static PyObject *__pyx_kp_u_utf_8;
static PyObject *__pyx_n_s_uuid;
static PyObject *__pyx_n_s_uuid4;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_verify_block_digest;
static PyObject *__pyx_n_s_verify_digests;
static PyObject *__pyx_n_s_verify_payload_digest;
static PyObject *__pyx_n_s_warcinfo;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_u_write;
static PyObject *__pyx_pf_8fastwarc_4warc_18CaseInsensitiveStr___hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_18CaseInsensitiveStr_2__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict___getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_2__setitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_4__contains__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_6get(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_8setdefault(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_10pop(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_12update(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_it, PyObject *__pyx_v_kwargs); /* proto */
static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap___init__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_2__cinit__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_encoding); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_4__getitem__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_header_key); /* proto */
static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_6__setitem__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_header_key, PyObject *__pyx_v_header_value); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_8__iter__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11__repr__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_13__contains__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_line___get__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_line_2__set__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_status_line); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_code___get__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_15append(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_17get(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_19items(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_21keys(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_23values(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_25asdict(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_27astuples(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_29__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_31__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8fastwarc_4warc_10WarcRecord___cinit__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_9record_id___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_11record_type___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static int __pyx_pf_8fastwarc_4warc_10WarcRecord_11record_type_2__set__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_7headers___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_7is_http___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_14is_http_parsed___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_12http_headers___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_12http_charset___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_14content_length___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_6reader___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_10stream_pos___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_2init_headers(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, size_t __pyx_v_content_length, enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type, PyObject *__pyx_v_record_urn); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_4set_bytes_content(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_6parse_http(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_8write(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_stream, int __pyx_v_checksum_data, size_t __pyx_v_chunk_size); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_10verify_block_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_v_consume); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_12verify_payload_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_v_consume); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8fastwarc_4warc_15ArchiveIterator___init__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static int __pyx_pf_8fastwarc_4warc_15ArchiveIterator_2__cinit__(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, PyObject *__pyx_v_stream, uint16_t __pyx_v_record_types, int __pyx_v_parse_http, size_t __pyx_v_min_content_length, size_t __pyx_v_max_content_length, PyObject *__pyx_v_func_filter, int __pyx_v_verify_digests); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_15ArchiveIterator_4__iter__(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_15ArchiveIterator_7_set_stream(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_is_warc_10(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_2is_warc_11(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_4has_block_digest(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_6has_payload_digest(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_8is_http(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_8fastwarc_4warc_10is_concurrent(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record); /* proto */
static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_8fastwarc_4warc_WarcHeaderMap(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8fastwarc_4warc_WarcRecord(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8fastwarc_4warc_ArchiveIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8fastwarc_4warc___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8fastwarc_4warc___pyx_scope_struct_1___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_222419149;
static enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_k__6;
static uint16_t __pyx_k__13;
static size_t __pyx_k__14;
static size_t __pyx_k__15;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
/* Late includes */

/* "fastwarc/warc.pyx":37
 * 
 * 
 * cdef string strip_str(const string& s) nogil:             # <<<<<<<<<<<<<<
 *     cdef size_t start = 0
 *     cdef size_t end = s.size()
 */

static std::string __pyx_f_8fastwarc_4warc_strip_str(std::string const &__pyx_v_s) {
  size_t __pyx_v_start;
  size_t __pyx_v_end;
  std::string __pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  std::string __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "fastwarc/warc.pyx":38
 * 
 * cdef string strip_str(const string& s) nogil:
 *     cdef size_t start = 0             # <<<<<<<<<<<<<<
 *     cdef size_t end = s.size()
 * 
 */
  __pyx_v_start = 0;

  /* "fastwarc/warc.pyx":39
 * cdef string strip_str(const string& s) nogil:
 *     cdef size_t start = 0
 *     cdef size_t end = s.size()             # <<<<<<<<<<<<<<
 * 
 *     for start in range(0, s.size()):
 */
  __pyx_v_end = __pyx_v_s.size();

  /* "fastwarc/warc.pyx":41
 *     cdef size_t end = s.size()
 * 
 *     for start in range(0, s.size()):             # <<<<<<<<<<<<<<
 *         if not isspace(s[start]):
 *             break
 */
  __pyx_t_1 = __pyx_v_s.size();
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_start = __pyx_t_3;

    /* "fastwarc/warc.pyx":42
 * 
 *     for start in range(0, s.size()):
 *         if not isspace(s[start]):             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_4 = ((!(std::isspace((__pyx_v_s[__pyx_v_start])) != 0)) != 0);
    if (__pyx_t_4) {

      /* "fastwarc/warc.pyx":43
 *     for start in range(0, s.size()):
 *         if not isspace(s[start]):
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for end in reversed(range(s.size())):
 */
      goto __pyx_L4_break;

      /* "fastwarc/warc.pyx":42
 * 
 *     for start in range(0, s.size()):
 *         if not isspace(s[start]):             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L4_break:;

  /* "fastwarc/warc.pyx":45
 *             break
 * 
 *     for end in reversed(range(s.size())):             # <<<<<<<<<<<<<<
 *         if not isspace(s[end]):
 *             break
 */
  for (__pyx_t_1 = __pyx_v_s.size()-1 + 1; __pyx_t_1 >= 0 + 1; ) { __pyx_t_1-=1;
    __pyx_v_end = __pyx_t_1;

    /* "fastwarc/warc.pyx":46
 * 
 *     for end in reversed(range(s.size())):
 *         if not isspace(s[end]):             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_4 = ((!(std::isspace((__pyx_v_s[__pyx_v_end])) != 0)) != 0);
    if (__pyx_t_4) {

      /* "fastwarc/warc.pyx":47
 *     for end in reversed(range(s.size())):
 *         if not isspace(s[end]):
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return s.substr(start, end - start + 1)
 */
      goto __pyx_L7_break;

      /* "fastwarc/warc.pyx":46
 * 
 *     for end in reversed(range(s.size())):
 *         if not isspace(s[end]):             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L7_break:;

  /* "fastwarc/warc.pyx":49
 *             break
 * 
 *     return s.substr(start, end - start + 1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  try {
    __pyx_t_5 = __pyx_v_s.substr(__pyx_v_start, ((__pyx_v_end - __pyx_v_start) + 1));
  } catch(...) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_CppExn2PyErr();
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 49, __pyx_L1_error)
  }
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":37
 * 
 * 
 * cdef string strip_str(const string& s) nogil:             # <<<<<<<<<<<<<<
 *     cdef size_t start = 0
 *     cdef size_t end = s.size()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fastwarc.warc.strip_str", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  return __pyx_r;
}

/* "fastwarc/warc.pyx":52
 * 
 * 
 * cdef string str_to_lower(string s) nogil:             # <<<<<<<<<<<<<<
 *     for i in range(s.size()):
 *         s[i] = tolower(s[i])
 */

static std::string __pyx_f_8fastwarc_4warc_str_to_lower(std::string __pyx_v_s) {
  size_t __pyx_v_i;
  std::string __pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;

  /* "fastwarc/warc.pyx":53
 * 
 * cdef string str_to_lower(string s) nogil:
 *     for i in range(s.size()):             # <<<<<<<<<<<<<<
 *         s[i] = tolower(s[i])
 *     return s
 */
  __pyx_t_1 = __pyx_v_s.size();
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fastwarc/warc.pyx":54
 * cdef string str_to_lower(string s) nogil:
 *     for i in range(s.size()):
 *         s[i] = tolower(s[i])             # <<<<<<<<<<<<<<
 *     return s
 * 
 */
    (__pyx_v_s[__pyx_v_i]) = std::tolower((__pyx_v_s[__pyx_v_i]));
  }

  /* "fastwarc/warc.pyx":55
 *     for i in range(s.size()):
 *         s[i] = tolower(s[i])
 *     return s             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_s;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":52
 * 
 * 
 * cdef string str_to_lower(string s) nogil:             # <<<<<<<<<<<<<<
 *     for i in range(s.size()):
 *         s[i] = tolower(s[i])
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fastwarc/warc.pyx":58
 * 
 * 
 * cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:             # <<<<<<<<<<<<<<
 *     if record_type == warcinfo:
 *         return b'warcinfo'
 */

static char const *__pyx_f_8fastwarc_4warc__enum_record_type_to_str(enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type) {
  char const *__pyx_r;

  /* "fastwarc/warc.pyx":59
 * 
 * cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:
 *     if record_type == warcinfo:             # <<<<<<<<<<<<<<
 *         return b'warcinfo'
 *     elif record_type == response:
 */
  switch (__pyx_v_record_type) {
    case __pyx_e_8fastwarc_4warc_warcinfo:

    /* "fastwarc/warc.pyx":60
 * cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:
 *     if record_type == warcinfo:
 *         return b'warcinfo'             # <<<<<<<<<<<<<<
 *     elif record_type == response:
 *         return b'response'
 */
    __pyx_r = ((char const *)"warcinfo");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":59
 * 
 * cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:
 *     if record_type == warcinfo:             # <<<<<<<<<<<<<<
 *         return b'warcinfo'
 *     elif record_type == response:
 */
    break;
    case __pyx_e_8fastwarc_4warc_response:

    /* "fastwarc/warc.pyx":62
 *         return b'warcinfo'
 *     elif record_type == response:
 *         return b'response'             # <<<<<<<<<<<<<<
 *     elif record_type == resource:
 *         return b'resource'
 */
    __pyx_r = ((char const *)"response");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":61
 *     if record_type == warcinfo:
 *         return b'warcinfo'
 *     elif record_type == response:             # <<<<<<<<<<<<<<
 *         return b'response'
 *     elif record_type == resource:
 */
    break;
    case __pyx_e_8fastwarc_4warc_resource:

    /* "fastwarc/warc.pyx":64
 *         return b'response'
 *     elif record_type == resource:
 *         return b'resource'             # <<<<<<<<<<<<<<
 *     elif record_type == request:
 *         return b'request'
 */
    __pyx_r = ((char const *)"resource");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":63
 *     elif record_type == response:
 *         return b'response'
 *     elif record_type == resource:             # <<<<<<<<<<<<<<
 *         return b'resource'
 *     elif record_type == request:
 */
    break;
    case __pyx_e_8fastwarc_4warc_request:

    /* "fastwarc/warc.pyx":66
 *         return b'resource'
 *     elif record_type == request:
 *         return b'request'             # <<<<<<<<<<<<<<
 *     elif record_type == metadata:
 *         return b'metadata'
 */
    __pyx_r = ((char const *)"request");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":65
 *     elif record_type == resource:
 *         return b'resource'
 *     elif record_type == request:             # <<<<<<<<<<<<<<
 *         return b'request'
 *     elif record_type == metadata:
 */
    break;
    case __pyx_e_8fastwarc_4warc_metadata:

    /* "fastwarc/warc.pyx":68
 *         return b'request'
 *     elif record_type == metadata:
 *         return b'metadata'             # <<<<<<<<<<<<<<
 *     elif record_type == revisit:
 *         return b'revisit'
 */
    __pyx_r = ((char const *)"metadata");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":67
 *     elif record_type == request:
 *         return b'request'
 *     elif record_type == metadata:             # <<<<<<<<<<<<<<
 *         return b'metadata'
 *     elif record_type == revisit:
 */
    break;
    case __pyx_e_8fastwarc_4warc_revisit:

    /* "fastwarc/warc.pyx":70
 *         return b'metadata'
 *     elif record_type == revisit:
 *         return b'revisit'             # <<<<<<<<<<<<<<
 *     elif record_type == conversion:
 *         return b'conversion'
 */
    __pyx_r = ((char const *)"revisit");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":69
 *     elif record_type == metadata:
 *         return b'metadata'
 *     elif record_type == revisit:             # <<<<<<<<<<<<<<
 *         return b'revisit'
 *     elif record_type == conversion:
 */
    break;
    case __pyx_e_8fastwarc_4warc_conversion:

    /* "fastwarc/warc.pyx":72
 *         return b'revisit'
 *     elif record_type == conversion:
 *         return b'conversion'             # <<<<<<<<<<<<<<
 *     elif record_type == continuation:
 *         return b'continuation'
 */
    __pyx_r = ((char const *)"conversion");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":71
 *     elif record_type == revisit:
 *         return b'revisit'
 *     elif record_type == conversion:             # <<<<<<<<<<<<<<
 *         return b'conversion'
 *     elif record_type == continuation:
 */
    break;
    case __pyx_e_8fastwarc_4warc_continuation:

    /* "fastwarc/warc.pyx":74
 *         return b'conversion'
 *     elif record_type == continuation:
 *         return b'continuation'             # <<<<<<<<<<<<<<
 *     else:
 *         return b'unknown'
 */
    __pyx_r = ((char const *)"continuation");
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":73
 *     elif record_type == conversion:
 *         return b'conversion'
 *     elif record_type == continuation:             # <<<<<<<<<<<<<<
 *         return b'continuation'
 *     else:
 */
    break;
    default:

    /* "fastwarc/warc.pyx":76
 *         return b'continuation'
 *     else:
 *         return b'unknown'             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = ((char const *)"unknown");
    goto __pyx_L0;
    break;
  }

  /* "fastwarc/warc.pyx":58
 * 
 * 
 * cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:             # <<<<<<<<<<<<<<
 *     if record_type == warcinfo:
 *         return b'warcinfo'
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fastwarc/warc.pyx":79
 * 
 * 
 * cdef WarcRecordType _str_record_type_to_enum(string record_type) nogil:             # <<<<<<<<<<<<<<
 *     record_type = str_to_lower(record_type)
 *     if record_type == b'warcinfo':
 */

static enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_f_8fastwarc_4warc__str_record_type_to_enum(std::string __pyx_v_record_type) {
  enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_r;
  int __pyx_t_1;

  /* "fastwarc/warc.pyx":80
 * 
 * cdef WarcRecordType _str_record_type_to_enum(string record_type) nogil:
 *     record_type = str_to_lower(record_type)             # <<<<<<<<<<<<<<
 *     if record_type == b'warcinfo':
 *         return warcinfo
 */
  __pyx_v_record_type = __pyx_f_8fastwarc_4warc_str_to_lower(__pyx_v_record_type);

  /* "fastwarc/warc.pyx":81
 * cdef WarcRecordType _str_record_type_to_enum(string record_type) nogil:
 *     record_type = str_to_lower(record_type)
 *     if record_type == b'warcinfo':             # <<<<<<<<<<<<<<
 *         return warcinfo
 *     elif record_type == b'response':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"warcinfo")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":82
 *     record_type = str_to_lower(record_type)
 *     if record_type == b'warcinfo':
 *         return warcinfo             # <<<<<<<<<<<<<<
 *     elif record_type == b'response':
 *         return response
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_warcinfo;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":81
 * cdef WarcRecordType _str_record_type_to_enum(string record_type) nogil:
 *     record_type = str_to_lower(record_type)
 *     if record_type == b'warcinfo':             # <<<<<<<<<<<<<<
 *         return warcinfo
 *     elif record_type == b'response':
 */
  }

  /* "fastwarc/warc.pyx":83
 *     if record_type == b'warcinfo':
 *         return warcinfo
 *     elif record_type == b'response':             # <<<<<<<<<<<<<<
 *         return response
 *     elif record_type == b'resource':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"response")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":84
 *         return warcinfo
 *     elif record_type == b'response':
 *         return response             # <<<<<<<<<<<<<<
 *     elif record_type == b'resource':
 *         return resource
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_response;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":83
 *     if record_type == b'warcinfo':
 *         return warcinfo
 *     elif record_type == b'response':             # <<<<<<<<<<<<<<
 *         return response
 *     elif record_type == b'resource':
 */
  }

  /* "fastwarc/warc.pyx":85
 *     elif record_type == b'response':
 *         return response
 *     elif record_type == b'resource':             # <<<<<<<<<<<<<<
 *         return resource
 *     elif record_type == b'request':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"resource")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":86
 *         return response
 *     elif record_type == b'resource':
 *         return resource             # <<<<<<<<<<<<<<
 *     elif record_type == b'request':
 *         return request
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_resource;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":85
 *     elif record_type == b'response':
 *         return response
 *     elif record_type == b'resource':             # <<<<<<<<<<<<<<
 *         return resource
 *     elif record_type == b'request':
 */
  }

  /* "fastwarc/warc.pyx":87
 *     elif record_type == b'resource':
 *         return resource
 *     elif record_type == b'request':             # <<<<<<<<<<<<<<
 *         return request
 *     elif record_type == b'metadata':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"request")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":88
 *         return resource
 *     elif record_type == b'request':
 *         return request             # <<<<<<<<<<<<<<
 *     elif record_type == b'metadata':
 *         return metadata
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_request;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":87
 *     elif record_type == b'resource':
 *         return resource
 *     elif record_type == b'request':             # <<<<<<<<<<<<<<
 *         return request
 *     elif record_type == b'metadata':
 */
  }

  /* "fastwarc/warc.pyx":89
 *     elif record_type == b'request':
 *         return request
 *     elif record_type == b'metadata':             # <<<<<<<<<<<<<<
 *         return metadata
 *     elif record_type == b'revisit':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"metadata")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":90
 *         return request
 *     elif record_type == b'metadata':
 *         return metadata             # <<<<<<<<<<<<<<
 *     elif record_type == b'revisit':
 *         return revisit
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_metadata;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":89
 *     elif record_type == b'request':
 *         return request
 *     elif record_type == b'metadata':             # <<<<<<<<<<<<<<
 *         return metadata
 *     elif record_type == b'revisit':
 */
  }

  /* "fastwarc/warc.pyx":91
 *     elif record_type == b'metadata':
 *         return metadata
 *     elif record_type == b'revisit':             # <<<<<<<<<<<<<<
 *         return revisit
 *     elif record_type == b'conversion':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"revisit")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":92
 *         return metadata
 *     elif record_type == b'revisit':
 *         return revisit             # <<<<<<<<<<<<<<
 *     elif record_type == b'conversion':
 *         return conversion
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_revisit;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":91
 *     elif record_type == b'metadata':
 *         return metadata
 *     elif record_type == b'revisit':             # <<<<<<<<<<<<<<
 *         return revisit
 *     elif record_type == b'conversion':
 */
  }

  /* "fastwarc/warc.pyx":93
 *     elif record_type == b'revisit':
 *         return revisit
 *     elif record_type == b'conversion':             # <<<<<<<<<<<<<<
 *         return conversion
 *     elif record_type == b'continuation':
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"conversion")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":94
 *         return revisit
 *     elif record_type == b'conversion':
 *         return conversion             # <<<<<<<<<<<<<<
 *     elif record_type == b'continuation':
 *         return continuation
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_conversion;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":93
 *     elif record_type == b'revisit':
 *         return revisit
 *     elif record_type == b'conversion':             # <<<<<<<<<<<<<<
 *         return conversion
 *     elif record_type == b'continuation':
 */
  }

  /* "fastwarc/warc.pyx":95
 *     elif record_type == b'conversion':
 *         return conversion
 *     elif record_type == b'continuation':             # <<<<<<<<<<<<<<
 *         return continuation
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_record_type == ((char const *)"continuation")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":96
 *         return conversion
 *     elif record_type == b'continuation':
 *         return continuation             # <<<<<<<<<<<<<<
 *     else:
 *         return unknown
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_continuation;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":95
 *     elif record_type == b'conversion':
 *         return conversion
 *     elif record_type == b'continuation':             # <<<<<<<<<<<<<<
 *         return continuation
 *     else:
 */
  }

  /* "fastwarc/warc.pyx":98
 *         return continuation
 *     else:
 *         return unknown             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = __pyx_e_8fastwarc_4warc_unknown;
    goto __pyx_L0;
  }

  /* "fastwarc/warc.pyx":79
 * 
 * 
 * cdef WarcRecordType _str_record_type_to_enum(string record_type) nogil:             # <<<<<<<<<<<<<<
 *     record_type = str_to_lower(record_type)
 *     if record_type == b'warcinfo':
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fastwarc/warc.pyx":104
 *     """Case insensitive str implementation for use as dict key."""
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.casefold())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_18CaseInsensitiveStr_1__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_18CaseInsensitiveStr_1__hash__ = {"__hash__", (PyCFunction)__pyx_pw_8fastwarc_4warc_18CaseInsensitiveStr_1__hash__, METH_O, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_18CaseInsensitiveStr_1__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_18CaseInsensitiveStr___hash__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_18CaseInsensitiveStr___hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_hash_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "fastwarc/warc.pyx":105
 * 
 *     def __hash__(self):
 *         return hash(self.casefold())             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_casefold); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_hash_t)-1))) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":104
 *     """Case insensitive str implementation for use as dict key."""
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.casefold())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStr.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":107
 *         return hash(self.casefold())
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self.casefold() == other.casefold()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_18CaseInsensitiveStr_3__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_18CaseInsensitiveStr_3__eq__ = {"__eq__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_18CaseInsensitiveStr_3__eq__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_18CaseInsensitiveStr_3__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, 1); __PYX_ERR(0, 107, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__eq__") < 0)) __PYX_ERR(0, 107, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 107, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStr.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_18CaseInsensitiveStr_2__eq__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_18CaseInsensitiveStr_2__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "fastwarc/warc.pyx":108
 * 
 *     def __eq__(self, other):
 *         return self.casefold() == other.casefold()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_casefold); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_casefold); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":107
 *         return hash(self.casefold())
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self.casefold() == other.casefold()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStr.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":114
 *     """Case-insensitive str dict."""
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_1__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_1__getitem__ = {"__getitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_1__getitem__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_1__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, 1); __PYX_ERR(0, 114, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__getitem__") < 0)) __PYX_ERR(0, 114, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 114, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict___getitem__(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict___getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "fastwarc/warc.pyx":115
 * 
 *     def __getitem__(self, key):
 *         return super().__getitem__(CaseInsensitiveStr(key))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, key, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 115, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_key);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":114
 *     """Case-insensitive str dict."""
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":117
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 *     def __setitem__(self, key, value):             # <<<<<<<<<<<<<<
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_3__setitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_3__setitem__ = {"__setitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_3__setitem__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_3__setitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setitem__", 1, 3, 3, 1); __PYX_ERR(0, 117, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setitem__", 1, 3, 3, 2); __PYX_ERR(0, 117, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setitem__") < 0)) __PYX_ERR(0, 117, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setitem__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 117, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_2__setitem__(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_2__setitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "fastwarc/warc.pyx":118
 * 
 *     def __setitem__(self, key, value):
 *         super().__setitem__(CaseInsensitiveStr(key), value)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 118, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_setitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_key);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_value);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":117
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 *     def __setitem__(self, key, value):             # <<<<<<<<<<<<<<
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":120
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_5__contains__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_5__contains__ = {"__contains__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_5__contains__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_5__contains__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__contains__", 1, 2, 2, 1); __PYX_ERR(0, 120, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__contains__") < 0)) __PYX_ERR(0, 120, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__contains__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 120, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_4__contains__(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_4__contains__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "fastwarc/warc.pyx":121
 * 
 *     def __contains__(self, key):
 *         return super().__contains__(CaseInsensitiveStr(key))             # <<<<<<<<<<<<<<
 * 
 *     def get(self, key, value=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 121, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_contains); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_key);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":120
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":123
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 *     def get(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_7get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_7get = {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_7get, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_7get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 0, 2, 3, 1); __PYX_ERR(0, 123, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 123, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_6get(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_6get(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "fastwarc/warc.pyx":124
 * 
 *     def get(self, key, value=None):
 *         return super().get(CaseInsensitiveStr(key), value)             # <<<<<<<<<<<<<<
 * 
 *     def setdefault(self, key, value=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 124, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_key);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_value);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":123
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 *     def get(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":126
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 *     def setdefault(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_9setdefault(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_9setdefault = {"setdefault", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_9setdefault, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_9setdefault(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setdefault (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setdefault", 0, 2, 3, 1); __PYX_ERR(0, 126, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setdefault") < 0)) __PYX_ERR(0, 126, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setdefault", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 126, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.setdefault", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_8setdefault(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_8setdefault(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setdefault", 0);

  /* "fastwarc/warc.pyx":127
 * 
 *     def setdefault(self, key, value=None):
 *         return super().setdefault(CaseInsensitiveStr(key), value)             # <<<<<<<<<<<<<<
 * 
 *     def pop(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 127, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_setdefault); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_key);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_value);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":126
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 *     def setdefault(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.setdefault", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":129
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 *     def pop(self, key):             # <<<<<<<<<<<<<<
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_11pop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_11pop = {"pop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_11pop, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_11pop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pop", 1, 2, 2, 1); __PYX_ERR(0, 129, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pop") < 0)) __PYX_ERR(0, 129, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 129, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_10pop(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_10pop(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);

  /* "fastwarc/warc.pyx":130
 * 
 *     def pop(self, key):
 *         return super().pop(CaseInsensitiveStr(key))             # <<<<<<<<<<<<<<
 * 
 *     def update(self, it=None, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 130, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_pop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_key);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":129
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 *     def pop(self, key):             # <<<<<<<<<<<<<<
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":132
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 *     def update(self, it=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if hasattr(it, 'items'):
 *             it = it.items()
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_13update(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_13update = {"update", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_13update, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8fastwarc_4warc_22CaseInsensitiveStrDict_13update(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_it = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_it,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_it);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "update") < 0)) __PYX_ERR(0, 132, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_it = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 132, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_12update(__pyx_self, __pyx_v_self, __pyx_v_it, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_22CaseInsensitiveStrDict_12update(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_it, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);
  __Pyx_INCREF(__pyx_v_it);

  /* "fastwarc/warc.pyx":133
 * 
 *     def update(self, it=None, **kwargs):
 *         if hasattr(it, 'items'):             # <<<<<<<<<<<<<<
 *             it = it.items()
 *         if it is not None:
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_it, __pyx_n_u_items); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":134
 *     def update(self, it=None, **kwargs):
 *         if hasattr(it, 'items'):
 *             it = it.items()             # <<<<<<<<<<<<<<
 *         if it is not None:
 *             for k, v in it:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_it, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_it, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":133
 * 
 *     def update(self, it=None, **kwargs):
 *         if hasattr(it, 'items'):             # <<<<<<<<<<<<<<
 *             it = it.items()
 *         if it is not None:
 */
  }

  /* "fastwarc/warc.pyx":135
 *         if hasattr(it, 'items'):
 *             it = it.items()
 *         if it is not None:             # <<<<<<<<<<<<<<
 *             for k, v in it:
 *                 super().__setitem__(CaseInsensitiveStr(k), v)
 */
  __pyx_t_2 = (__pyx_v_it != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":136
 *             it = it.items()
 *         if it is not None:
 *             for k, v in it:             # <<<<<<<<<<<<<<
 *                 super().__setitem__(CaseInsensitiveStr(k), v)
 *         for k, v in kwargs.items():
 */
    if (likely(PyList_CheckExact(__pyx_v_it)) || PyTuple_CheckExact(__pyx_v_it)) {
      __pyx_t_3 = __pyx_v_it; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_it); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 136, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 136, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 136, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 136, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 136, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 136, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 136, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 136, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 136, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 136, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 136, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fastwarc/warc.pyx":137
 *         if it is not None:
 *             for k, v in it:
 *                 super().__setitem__(CaseInsensitiveStr(k), v)             # <<<<<<<<<<<<<<
 *         for k, v in kwargs.items():
 *             super().__setitem__(CaseInsensitiveStr(k), v)
 */
      __pyx_t_8 = __Pyx_CyFunction_GetClassObj(__pyx_self);
      if (!__pyx_t_8) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 137, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self);
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_setitem); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_8 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_11, __pyx_v_k) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_k);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_t_8, __pyx_v_v};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_t_8, __pyx_v_v};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 137, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_12, __pyx_t_8);
        __Pyx_INCREF(__pyx_v_v);
        __Pyx_GIVEREF(__pyx_v_v);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_12, __pyx_v_v);
        __pyx_t_8 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastwarc/warc.pyx":136
 *             it = it.items()
 *         if it is not None:
 *             for k, v in it:             # <<<<<<<<<<<<<<
 *                 super().__setitem__(CaseInsensitiveStr(k), v)
 *         for k, v in kwargs.items():
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":135
 *         if hasattr(it, 'items'):
 *             it = it.items()
 *         if it is not None:             # <<<<<<<<<<<<<<
 *             for k, v in it:
 *                 super().__setitem__(CaseInsensitiveStr(k), v)
 */
  }

  /* "fastwarc/warc.pyx":138
 *             for k, v in it:
 *                 super().__setitem__(CaseInsensitiveStr(k), v)
 *         for k, v in kwargs.items():             # <<<<<<<<<<<<<<
 *             super().__setitem__(CaseInsensitiveStr(k), v)
 * 
 */
  __pyx_t_6 = 0;
  __pyx_t_4 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_n_s_items, (&__pyx_t_13), (&__pyx_t_12)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  while (1) {
    __pyx_t_14 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_13, &__pyx_t_6, &__pyx_t_4, &__pyx_t_5, NULL, __pyx_t_12);
    if (unlikely(__pyx_t_14 == 0)) break;
    if (unlikely(__pyx_t_14 == -1)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "fastwarc/warc.pyx":139
 *                 super().__setitem__(CaseInsensitiveStr(k), v)
 *         for k, v in kwargs.items():
 *             super().__setitem__(CaseInsensitiveStr(k), v)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __Pyx_CyFunction_GetClassObj(__pyx_self);
    if (!__pyx_t_4) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 139, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_self);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_setitem); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_v_k) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_k);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_14 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_14 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_4, __pyx_v_v};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_4, __pyx_v_v};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_14, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_14, __pyx_v_v);
      __pyx_t_4 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fastwarc/warc.pyx":132
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 *     def update(self, it=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if hasattr(it, 'items'):
 *             it = it.items()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("fastwarc.warc.CaseInsensitiveStrDict.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_it);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":146
 *     """Dict-like type representing a WARC or HTTP header block"""
 * 
 *     def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         :param encoding: header source encoding
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap___init__[] = "\n        :param encoding: header source encoding\n        :type encoding: str, optional, default: 'utf-8'\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_8fastwarc_4warc_13WarcHeaderMap___init__;
#endif
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 1))) return -1;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap___init__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap___init__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":152
 *         """
 * 
 *     def __cinit__(self, encoding='utf-8'):             # <<<<<<<<<<<<<<
 *         self._enc = encoding
 *         self._dict_cache = None
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_encoding = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_encoding,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_kp_u_utf_8);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_encoding);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 152, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_encoding = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 152, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_2__cinit__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), __pyx_v_encoding);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_2__cinit__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_encoding) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "fastwarc/warc.pyx":153
 * 
 *     def __cinit__(self, encoding='utf-8'):
 *         self._enc = encoding             # <<<<<<<<<<<<<<
 *         self._dict_cache = None
 *         self._dict_cache_stale = True
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_encoding))||((__pyx_v_encoding) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_encoding)->tp_name), 0))) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_encoding;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_enc);
  __Pyx_DECREF(__pyx_v_self->_enc);
  __pyx_v_self->_enc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":154
 *     def __cinit__(self, encoding='utf-8'):
 *         self._enc = encoding
 *         self._dict_cache = None             # <<<<<<<<<<<<<<
 *         self._dict_cache_stale = True
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_dict_cache);
  __Pyx_DECREF(__pyx_v_self->_dict_cache);
  __pyx_v_self->_dict_cache = ((PyObject*)Py_None);

  /* "fastwarc/warc.pyx":155
 *         self._enc = encoding
 *         self._dict_cache = None
 *         self._dict_cache_stale = True             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, header_key):
 */
  __pyx_v_self->_dict_cache_stale = 1;

  /* "fastwarc/warc.pyx":152
 *         """
 * 
 *     def __cinit__(self, encoding='utf-8'):             # <<<<<<<<<<<<<<
 *         self._enc = encoding
 *         self._dict_cache = None
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":157
 *         self._dict_cache_stale = True
 * 
 *     def __getitem__(self, header_key):             # <<<<<<<<<<<<<<
 *         return self.asdict()[header_key]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_header_key); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_header_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_4__getitem__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), ((PyObject *)__pyx_v_header_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_4__getitem__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_header_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "fastwarc/warc.pyx":158
 * 
 *     def __getitem__(self, header_key):
 *         return self.asdict()[header_key]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, header_key, header_value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_header_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":157
 *         self._dict_cache_stale = True
 * 
 *     def __getitem__(self, header_key):             # <<<<<<<<<<<<<<
 *         return self.asdict()[header_key]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":160
 *         return self.asdict()[header_key]
 * 
 *     def __setitem__(self, header_key, header_value):             # <<<<<<<<<<<<<<
 *         self.set_header(header_key.encode(self._enc, errors='ignore'),
 *                         header_value.encode(self._enc, errors='ignore'))
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_header_key, PyObject *__pyx_v_header_value); /*proto*/
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_header_key, PyObject *__pyx_v_header_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_6__setitem__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), ((PyObject *)__pyx_v_header_key), ((PyObject *)__pyx_v_header_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_6__setitem__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_header_key, PyObject *__pyx_v_header_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::string __pyx_t_5;
  std::string __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "fastwarc/warc.pyx":161
 * 
 *     def __setitem__(self, header_key, header_value):
 *         self.set_header(header_key.encode(self._enc, errors='ignore'),             # <<<<<<<<<<<<<<
 *                         header_value.encode(self._enc, errors='ignore'))
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_header_key, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_enc);
  __Pyx_GIVEREF(__pyx_v_self->_enc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_enc);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_convert_string_from_py_std__in_string(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastwarc/warc.pyx":162
 *     def __setitem__(self, header_key, header_value):
 *         self.set_header(header_key.encode(self._enc, errors='ignore'),
 *                         header_value.encode(self._enc, errors='ignore'))             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_header_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->_enc);
  __Pyx_GIVEREF(__pyx_v_self->_enc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->_enc);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 162, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_convert_string_from_py_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":161
 * 
 *     def __setitem__(self, header_key, header_value):
 *         self.set_header(header_key.encode(self._enc, errors='ignore'),             # <<<<<<<<<<<<<<
 *                         header_value.encode(self._enc, errors='ignore'))
 * 
 */
  ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->__pyx_vtab)->set_header(__pyx_v_self, __pyx_t_5, __pyx_t_6);

  /* "fastwarc/warc.pyx":160
 *         return self.asdict()[header_key]
 * 
 *     def __setitem__(self, header_key, header_value):             # <<<<<<<<<<<<<<
 *         self.set_header(header_key.encode(self._enc, errors='ignore'),
 *                         header_value.encode(self._enc, errors='ignore'))
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8fastwarc_4warc_13WarcHeaderMap_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "fastwarc/warc.pyx":164
 *                         header_value.encode(self._enc, errors='ignore'))
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield from self.items()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_9__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_9__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_8__iter__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_8__iter__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *)__pyx_tp_new_8fastwarc_4warc___pyx_scope_struct____iter__(__pyx_ptype_8fastwarc_4warc___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 164, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8fastwarc_4warc_13WarcHeaderMap_10generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_WarcHeaderMap___iter, __pyx_n_s_fastwarc_warc); if (unlikely(!gen)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8fastwarc_4warc_13WarcHeaderMap_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_yield_from;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 164, __pyx_L1_error)

  /* "fastwarc/warc.pyx":165
 * 
 *     def __iter__(self):
 *         yield from self.items()             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L4_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 165, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_Occurred();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(0, 165, __pyx_L1_error)
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "fastwarc/warc.pyx":164
 *                         header_value.encode(self._enc, errors='ignore'))
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         yield from self.items()
 * 
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":167
 *         yield from self.items()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self.asdict())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_12__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_12__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11__repr__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11__repr__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "fastwarc/warc.pyx":168
 * 
 *     def __repr__(self):
 *         return str(self.asdict())             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":167
 *         yield from self.items()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self.asdict())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":170
 *         return str(self.asdict())
 * 
 *     def __contains__(self, item):             # <<<<<<<<<<<<<<
 *         return item in self.asdict()
 * 
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_14__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_14__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_13__contains__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_13__contains__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_item) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "fastwarc/warc.pyx":171
 * 
 *     def __contains__(self, item):
 *         return item in self.asdict()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_item, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":170
 *         return str(self.asdict())
 * 
 *     def __contains__(self, item):             # <<<<<<<<<<<<<<
 *         return item in self.asdict()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":174
 * 
 *     @property
 *     def status_line(self) -> str:             # <<<<<<<<<<<<<<
 *         """Header status line."""
 *         return self._status_line.decode(self._enc, errors='ignore')
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_line_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_line_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_line___get__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_line___get__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":176
 *     def status_line(self) -> str:
 *         """Header status line."""
 *         return self._status_line.decode(self._enc, errors='ignore')             # <<<<<<<<<<<<<<
 * 
 *     @status_line.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_v_self->_status_line); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_enc);
  __Pyx_GIVEREF(__pyx_v_self->_enc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_enc);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":174
 * 
 *     @property
 *     def status_line(self) -> str:             # <<<<<<<<<<<<<<
 *         """Header status line."""
 *         return self._status_line.decode(self._enc, errors='ignore')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.status_line.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":179
 * 
 *     @status_line.setter
 *     def status_line(self, status_line):             # <<<<<<<<<<<<<<
 *         """Set status line contents."""
 *         self._status_line = status_line.encode(self._enc, errors='ignore')
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_line_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_status_line); /*proto*/
static int __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_line_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_status_line) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_line_2__set__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), ((PyObject *)__pyx_v_status_line));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_line_2__set__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_status_line) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::string __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "fastwarc/warc.pyx":181
 *     def status_line(self, status_line):
 *         """Set status line contents."""
 *         self._status_line = status_line.encode(self._enc, errors='ignore')             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_status_line, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_enc);
  __Pyx_GIVEREF(__pyx_v_self->_enc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_enc);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 181, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_convert_string_from_py_std__in_string(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->_status_line = __pyx_t_5;

  /* "fastwarc/warc.pyx":179
 * 
 *     @status_line.setter
 *     def status_line(self, status_line):             # <<<<<<<<<<<<<<
 *         """Set status line contents."""
 *         self._status_line = status_line.encode(self._enc, errors='ignore')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.status_line.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":184
 * 
 *     @property
 *     def status_code(self) -> int:             # <<<<<<<<<<<<<<
 *         """HTTP status code (unset if header block is not an HTTP header block)."""
 *         if self._status_line.find(<char*>b'HTTP/') != 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_code_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_code___get__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_11status_code___get__(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":186
 *     def status_code(self) -> int:
 *         """HTTP status code (unset if header block is not an HTTP header block)."""
 *         if self._status_line.find(<char*>b'HTTP/') != 0:             # <<<<<<<<<<<<<<
 *             return None
 *         s = self._status_line.split(b' ')
 */
  __pyx_t_1 = ((__pyx_v_self->_status_line.find(((char *)((char *)"HTTP/"))) != 0) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":187
 *         """HTTP status code (unset if header block is not an HTTP header block)."""
 *         if self._status_line.find(<char*>b'HTTP/') != 0:
 *             return None             # <<<<<<<<<<<<<<
 *         s = self._status_line.split(b' ')
 *         if len(s) != 3 or not s[1].isdigit():
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":186
 *     def status_code(self) -> int:
 *         """HTTP status code (unset if header block is not an HTTP header block)."""
 *         if self._status_line.find(<char*>b'HTTP/') != 0:             # <<<<<<<<<<<<<<
 *             return None
 *         s = self._status_line.split(b' ')
 */
  }

  /* "fastwarc/warc.pyx":188
 *         if self._status_line.find(<char*>b'HTTP/') != 0:
 *             return None
 *         s = self._status_line.split(b' ')             # <<<<<<<<<<<<<<
 *         if len(s) != 3 or not s[1].isdigit():
 *             return None
 */
  __pyx_t_3 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_v_self->_status_line); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_split); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_kp_b_) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_b_);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_s = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":189
 *             return None
 *         s = self._status_line.split(b' ')
 *         if len(s) != 3 or not s[1].isdigit():             # <<<<<<<<<<<<<<
 *             return None
 *         return int(s)
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_s); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 189, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 != 3) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_s, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_isdigit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":190
 *         s = self._status_line.split(b' ')
 *         if len(s) != 3 or not s[1].isdigit():
 *             return None             # <<<<<<<<<<<<<<
 *         return int(s)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":189
 *             return None
 *         s = self._status_line.split(b' ')
 *         if len(s) != 3 or not s[1].isdigit():             # <<<<<<<<<<<<<<
 *             return None
 *         return int(s)
 */
  }

  /* "fastwarc/warc.pyx":191
 *         if len(s) != 3 or not s[1].isdigit():
 *             return None
 *         return int(s)             # <<<<<<<<<<<<<<
 * 
 *     def append(self, key, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":184
 * 
 *     @property
 *     def status_code(self) -> int:             # <<<<<<<<<<<<<<
 *         """HTTP status code (unset if header block is not an HTTP header block)."""
 *         if self._status_line.find(<char*>b'HTTP/') != 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.status_code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":193
 *         return int(s)
 * 
 *     def append(self, key, value):             # <<<<<<<<<<<<<<
 *         """Append header (use if header name is not unique)."""
 *         self.append_header(key.encode(self._enc), value.encode(self._enc))
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_16append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_15append[] = "Append header (use if header name is not unique).";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_16append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("append", 1, 2, 2, 1); __PYX_ERR(0, 193, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "append") < 0)) __PYX_ERR(0, 193, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_key = values[0];
    __pyx_v_value = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("append", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 193, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_15append(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_15append(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  std::string __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);

  /* "fastwarc/warc.pyx":195
 *     def append(self, key, value):
 *         """Append header (use if header name is not unique)."""
 *         self.append_header(key.encode(self._enc), value.encode(self._enc))             # <<<<<<<<<<<<<<
 * 
 *     def get(self, item, default=None) -> str:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_self->_enc) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self->_enc);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_self->_enc) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self->_enc);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_convert_string_from_py_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(__pyx_v_self, __pyx_t_4, __pyx_t_5);

  /* "fastwarc/warc.pyx":193
 *         return int(s)
 * 
 *     def append(self, key, value):             # <<<<<<<<<<<<<<
 *         """Append header (use if header name is not unique)."""
 *         self.append_header(key.encode(self._enc), value.encode(self._enc))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":197
 *         self.append_header(key.encode(self._enc), value.encode(self._enc))
 * 
 *     def get(self, item, default=None) -> str:             # <<<<<<<<<<<<<<
 *         """Get header value or ``default``."""
 *         return self.asdict().get(item, default)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_18get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_17get[] = "Get header value or ``default``.";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_18get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_item = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_item,&__pyx_n_s_default,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_item)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 197, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_item = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 197, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_17get(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), __pyx_v_item, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_17get(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "fastwarc/warc.pyx":199
 *     def get(self, item, default=None) -> str:
 *         """Get header value or ``default``."""
 *         return self.asdict().get(item, default)             # <<<<<<<<<<<<<<
 * 
 *     def items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_item, __pyx_v_default};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_item, __pyx_v_default};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_item);
    __Pyx_GIVEREF(__pyx_v_item);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_item);
    __Pyx_INCREF(__pyx_v_default);
    __Pyx_GIVEREF(__pyx_v_default);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_default);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 199, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":197
 *         self.append_header(key.encode(self._enc), value.encode(self._enc))
 * 
 *     def get(self, item, default=None) -> str:             # <<<<<<<<<<<<<<
 *         """Get header value or ``default``."""
 *         return self.asdict().get(item, default)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":201
 *         return self.asdict().get(item, default)
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         """Item view of keys and values."""
 *         return self.asdict().items()
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_20items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_19items[] = "Item view of keys and values.";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_20items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_19items(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_19items(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);

  /* "fastwarc/warc.pyx":203
 *     def items(self):
 *         """Item view of keys and values."""
 *         return self.asdict().items()             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":201
 *         return self.asdict().get(item, default)
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         """Item view of keys and values."""
 *         return self.asdict().items()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":205
 *         return self.asdict().items()
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """Iterable of header keys."""
 *         return self.asdict().keys()
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_22keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_21keys[] = "Iterable of header keys.";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_22keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_21keys(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_21keys(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);

  /* "fastwarc/warc.pyx":207
 *     def keys(self):
 *         """Iterable of header keys."""
 *         return self.asdict().keys()             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_keys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":205
 *         return self.asdict().items()
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """Iterable of header keys."""
 *         return self.asdict().keys()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":209
 *         return self.asdict().keys()
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """Iterable of header values."""
 *         return self.asdict().values()
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_24values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_23values[] = "Iterable of header values.";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_24values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_23values(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_23values(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);

  /* "fastwarc/warc.pyx":211
 *     def values(self):
 *         """Iterable of header values."""
 *         return self.asdict().values()             # <<<<<<<<<<<<<<
 * 
 *     def asdict(self) -> dict:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":209
 *         return self.asdict().keys()
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """Iterable of header values."""
 *         return self.asdict().values()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":213
 *         return self.asdict().values()
 * 
 *     def asdict(self) -> dict:             # <<<<<<<<<<<<<<
 *         """Headers as Python dict."""
 *         cdef str_pair h
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_26asdict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_25asdict[] = "Headers as Python dict.";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_26asdict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("asdict (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_25asdict(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_25asdict(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  __pyx_t_8fastwarc_4warc_str_pair __pyx_7genexpr__pyx_v_h;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair> ::iterator __pyx_t_5;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair>  *__pyx_t_6;
  __pyx_t_8fastwarc_4warc_str_pair __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("asdict", 0);

  /* "fastwarc/warc.pyx":216
 *         """Headers as Python dict."""
 *         cdef str_pair h
 *         if self._dict_cache_stale:             # <<<<<<<<<<<<<<
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 */
  __pyx_t_1 = (__pyx_v_self->_dict_cache_stale != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":217
 *         cdef str_pair h
 *         if self._dict_cache_stale:
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({             # <<<<<<<<<<<<<<
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 *                 for h in self._headers})
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CaseInsensitiveStrDict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { /* enter inner scope */
      __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "fastwarc/warc.pyx":219
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 *                 for h in self._headers})             # <<<<<<<<<<<<<<
 *         return self._dict_cache
 * 
 */
      __pyx_t_6 = &__pyx_v_self->_headers;
      __pyx_t_5 = __pyx_t_6->begin();
      for (;;) {
        if (!(__pyx_t_5 != __pyx_t_6->end())) break;
        __pyx_t_7 = *__pyx_t_5;
        ++__pyx_t_5;
        __pyx_7genexpr__pyx_v_h = __pyx_t_7;

        /* "fastwarc/warc.pyx":218
 *         if self._dict_cache_stale:
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')             # <<<<<<<<<<<<<<
 *                 for h in self._headers})
 *         return self._dict_cache
 */
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_CaseInsensitiveStr); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_7genexpr__pyx_v_h.f0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_decode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_v_self->_enc);
        __Pyx_GIVEREF(__pyx_v_self->_enc);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_self->_enc);
        __pyx_t_12 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 218, __pyx_L1_error)
        __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_10, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_7genexpr__pyx_v_h.f1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_decode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_v_self->_enc);
        __Pyx_GIVEREF(__pyx_v_self->_enc);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_self->_enc);
        __pyx_t_12 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 218, __pyx_L1_error)
        __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_9, __pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_4, (PyObject*)__pyx_t_8, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 218, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "fastwarc/warc.pyx":219
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 *                 for h in self._headers})             # <<<<<<<<<<<<<<
 *         return self._dict_cache
 * 
 */
      }
    } /* exit inner scope */
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":217
 *         cdef str_pair h
 *         if self._dict_cache_stale:
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({             # <<<<<<<<<<<<<<
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 *                 for h in self._headers})
 */
    __pyx_t_3 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_dict_cache);
    __Pyx_DECREF(__pyx_v_self->_dict_cache);
    __pyx_v_self->_dict_cache = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":216
 *         """Headers as Python dict."""
 *         cdef str_pair h
 *         if self._dict_cache_stale:             # <<<<<<<<<<<<<<
 *             self._dict_cache = <dict>CaseInsensitiveStrDict({
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 */
  }

  /* "fastwarc/warc.pyx":220
 *                 CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')): h[1].decode(self._enc, errors='ignore')
 *                 for h in self._headers})
 *         return self._dict_cache             # <<<<<<<<<<<<<<
 * 
 *     def astuples(self) -> List[Tuple[str, str]]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_dict_cache);
  __pyx_r = __pyx_v_self->_dict_cache;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":213
 *         return self.asdict().values()
 * 
 *     def asdict(self) -> dict:             # <<<<<<<<<<<<<<
 *         """Headers as Python dict."""
 *         cdef str_pair h
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.asdict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":222
 *         return self._dict_cache
 * 
 *     def astuples(self) -> List[Tuple[str, str]]:             # <<<<<<<<<<<<<<
 *         """Headers as list of tuples (use if header keys are not necessarily unique)."""
 *         cdef str_pair h
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_28astuples(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_27astuples[] = "Headers as list of tuples (use if header keys are not necessarily unique).";
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_28astuples(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("astuples (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_27astuples(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_27astuples(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  __pyx_t_8fastwarc_4warc_str_pair __pyx_8genexpr1__pyx_v_h;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair> ::iterator __pyx_t_2;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair>  *__pyx_t_3;
  __pyx_t_8fastwarc_4warc_str_pair __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("astuples", 0);

  /* "fastwarc/warc.pyx":225
 *         """Headers as list of tuples (use if header keys are not necessarily unique)."""
 *         cdef str_pair h
 *         return [(h[0].decode(self._enc, errors='ignore'), h[1].decode(self._enc, errors='ignore'))             # <<<<<<<<<<<<<<
 *                 for h in self._headers]
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "fastwarc/warc.pyx":226
 *         cdef str_pair h
 *         return [(h[0].decode(self._enc, errors='ignore'), h[1].decode(self._enc, errors='ignore'))
 *                 for h in self._headers]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = &__pyx_v_self->_headers;
    __pyx_t_2 = __pyx_t_3->begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_t_3->end())) break;
      __pyx_t_4 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_8genexpr1__pyx_v_h = __pyx_t_4;

      /* "fastwarc/warc.pyx":225
 *         """Headers as list of tuples (use if header keys are not necessarily unique)."""
 *         cdef str_pair h
 *         return [(h[0].decode(self._enc, errors='ignore'), h[1].decode(self._enc, errors='ignore'))             # <<<<<<<<<<<<<<
 *                 for h in self._headers]
 * 
 */
      __pyx_t_5 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_8genexpr1__pyx_v_h.f0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_decode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_self->_enc);
      __Pyx_GIVEREF(__pyx_v_self->_enc);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_self->_enc);
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_8genexpr1__pyx_v_h.f1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_self->_enc);
      __Pyx_GIVEREF(__pyx_v_self->_enc);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_self->_enc);
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_errors, __pyx_n_u_ignore) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9);
      __pyx_t_8 = 0;
      __pyx_t_9 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fastwarc/warc.pyx":226
 *         cdef str_pair h
 *         return [(h[0].decode(self._enc, errors='ignore'), h[1].decode(self._enc, errors='ignore'))
 *                 for h in self._headers]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":222
 *         return self._dict_cache
 * 
 *     def astuples(self) -> List[Tuple[str, str]]:             # <<<<<<<<<<<<<<
 *         """Headers as list of tuples (use if header keys are not necessarily unique)."""
 *         cdef str_pair h
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.astuples", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":229
 * 
 * 
 *     cdef size_t write(self, IOStream stream):             # <<<<<<<<<<<<<<
 *         """Write header block into stream."""
 *         cdef size_t bytes_written = 0
 */

static size_t __pyx_f_8fastwarc_4warc_13WarcHeaderMap_write(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, struct __pyx_obj_8fastwarc_9stream_io_IOStream *__pyx_v_stream) {
  size_t __pyx_v_bytes_written;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair> ::iterator __pyx_v_it;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("write", 0);

  /* "fastwarc/warc.pyx":231
 *     cdef size_t write(self, IOStream stream):
 *         """Write header block into stream."""
 *         cdef size_t bytes_written = 0             # <<<<<<<<<<<<<<
 *         if not self._status_line.empty():
 *             bytes_written += stream.write(self._status_line.data(), self._status_line.size())
 */
  __pyx_v_bytes_written = 0;

  /* "fastwarc/warc.pyx":232
 *         """Write header block into stream."""
 *         cdef size_t bytes_written = 0
 *         if not self._status_line.empty():             # <<<<<<<<<<<<<<
 *             bytes_written += stream.write(self._status_line.data(), self._status_line.size())
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 */
  __pyx_t_1 = ((!(__pyx_v_self->_status_line.empty() != 0)) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":233
 *         cdef size_t bytes_written = 0
 *         if not self._status_line.empty():
 *             bytes_written += stream.write(self._status_line.data(), self._status_line.size())             # <<<<<<<<<<<<<<
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 * 
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_stream->__pyx_vtab)->write(__pyx_v_stream, __pyx_v_self->_status_line.data(), __pyx_v_self->_status_line.size()));

    /* "fastwarc/warc.pyx":234
 *         if not self._status_line.empty():
 *             bytes_written += stream.write(self._status_line.data(), self._status_line.size())
 *             bytes_written += stream.write(<char*>b'\r\n', 2)             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_stream->__pyx_vtab)->write(__pyx_v_stream, ((char *)((char *)"\r\n")), 2));

    /* "fastwarc/warc.pyx":232
 *         """Write header block into stream."""
 *         cdef size_t bytes_written = 0
 *         if not self._status_line.empty():             # <<<<<<<<<<<<<<
 *             bytes_written += stream.write(self._status_line.data(), self._status_line.size())
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 */
  }

  /* "fastwarc/warc.pyx":236
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 * 
 *         cdef vector[str_pair].iterator it = self._headers.begin()             # <<<<<<<<<<<<<<
 *         while it != self._headers.end():
 *             if not deref(it)[0].empty():
 */
  __pyx_v_it = __pyx_v_self->_headers.begin();

  /* "fastwarc/warc.pyx":237
 * 
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():             # <<<<<<<<<<<<<<
 *             if not deref(it)[0].empty():
 *                 bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_it != __pyx_v_self->_headers.end()) != 0);
    if (!__pyx_t_1) break;

    /* "fastwarc/warc.pyx":238
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 *             if not deref(it)[0].empty():             # <<<<<<<<<<<<<<
 *                 bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())
 *                 bytes_written += stream.write(<char*>b': ', 2)
 */
    __pyx_t_1 = ((!((*__pyx_v_it).f0.empty() != 0)) != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":239
 *         while it != self._headers.end():
 *             if not deref(it)[0].empty():
 *                 bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())             # <<<<<<<<<<<<<<
 *                 bytes_written += stream.write(<char*>b': ', 2)
 *             bytes_written += stream.write(deref(it)[1].data(), deref(it)[1].size())
 */
      __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_stream->__pyx_vtab)->write(__pyx_v_stream, (*__pyx_v_it).f0.data(), (*__pyx_v_it).f0.size()));

      /* "fastwarc/warc.pyx":240
 *             if not deref(it)[0].empty():
 *                 bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())
 *                 bytes_written += stream.write(<char*>b': ', 2)             # <<<<<<<<<<<<<<
 *             bytes_written += stream.write(deref(it)[1].data(), deref(it)[1].size())
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 */
      __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_stream->__pyx_vtab)->write(__pyx_v_stream, ((char *)((char *)": ")), 2));

      /* "fastwarc/warc.pyx":238
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 *             if not deref(it)[0].empty():             # <<<<<<<<<<<<<<
 *                 bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())
 *                 bytes_written += stream.write(<char*>b': ', 2)
 */
    }

    /* "fastwarc/warc.pyx":241
 *                 bytes_written += stream.write(deref(it)[0].data(), deref(it)[0].size())
 *                 bytes_written += stream.write(<char*>b': ', 2)
 *             bytes_written += stream.write(deref(it)[1].data(), deref(it)[1].size())             # <<<<<<<<<<<<<<
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 *             inc(it)
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_stream->__pyx_vtab)->write(__pyx_v_stream, (*__pyx_v_it).f1.data(), (*__pyx_v_it).f1.size()));

    /* "fastwarc/warc.pyx":242
 *                 bytes_written += stream.write(<char*>b': ', 2)
 *             bytes_written += stream.write(deref(it)[1].data(), deref(it)[1].size())
 *             bytes_written += stream.write(<char*>b'\r\n', 2)             # <<<<<<<<<<<<<<
 *             inc(it)
 *         return bytes_written
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_stream->__pyx_vtab)->write(__pyx_v_stream, ((char *)((char *)"\r\n")), 2));

    /* "fastwarc/warc.pyx":243
 *             bytes_written += stream.write(deref(it)[1].data(), deref(it)[1].size())
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 *             inc(it)             # <<<<<<<<<<<<<<
 *         return bytes_written
 * 
 */
    (void)((++__pyx_v_it));
  }

  /* "fastwarc/warc.pyx":244
 *             bytes_written += stream.write(<char*>b'\r\n', 2)
 *             inc(it)
 *         return bytes_written             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void clear(self):
 */
  __pyx_r = __pyx_v_bytes_written;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":229
 * 
 * 
 *     cdef size_t write(self, IOStream stream):             # <<<<<<<<<<<<<<
 *         """Write header block into stream."""
 *         cdef size_t bytes_written = 0
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":246
 *         return bytes_written
 * 
 *     cdef inline void clear(self):             # <<<<<<<<<<<<<<
 *         """Clear headers."""
 *         self._headers.clear()
 */

static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_clear(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear", 0);

  /* "fastwarc/warc.pyx":248
 *     cdef inline void clear(self):
 *         """Clear headers."""
 *         self._headers.clear()             # <<<<<<<<<<<<<<
 *         self._dict_cache_stale = True
 * 
 */
  __pyx_v_self->_headers.clear();

  /* "fastwarc/warc.pyx":249
 *         """Clear headers."""
 *         self._headers.clear()
 *         self._dict_cache_stale = True             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void set_status_line(self, const string& status_line):
 */
  __pyx_v_self->_dict_cache_stale = 1;

  /* "fastwarc/warc.pyx":246
 *         return bytes_written
 * 
 *     cdef inline void clear(self):             # <<<<<<<<<<<<<<
 *         """Clear headers."""
 *         self._headers.clear()
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "fastwarc/warc.pyx":251
 *         self._dict_cache_stale = True
 * 
 *     cdef inline void set_status_line(self, const string& status_line):             # <<<<<<<<<<<<<<
 *         """Set status line string."""
 *         self._status_line = status_line
 */

static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_status_line) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_status_line", 0);

  /* "fastwarc/warc.pyx":253
 *     cdef inline void set_status_line(self, const string& status_line):
 *         """Set status line string."""
 *         self._status_line = status_line             # <<<<<<<<<<<<<<
 * 
 *     cdef string find_header(self, const string& header_key, const string& default):
 */
  __pyx_v_self->_status_line = __pyx_v_status_line;

  /* "fastwarc/warc.pyx":251
 *         self._dict_cache_stale = True
 * 
 *     cdef inline void set_status_line(self, const string& status_line):             # <<<<<<<<<<<<<<
 *         """Set status line string."""
 *         self._status_line = status_line
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "fastwarc/warc.pyx":255
 *         self._status_line = status_line
 * 
 *     cdef string find_header(self, const string& header_key, const string& default):             # <<<<<<<<<<<<<<
 *         """Return value of first header occurrence or ``default`` (linear complexity)."""
 *         cdef string header_key_lower = str_to_lower(header_key)
 */

static std::string __pyx_f_8fastwarc_4warc_13WarcHeaderMap_find_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_key, std::string const &__pyx_v_default) {
  std::string __pyx_v_header_key_lower;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair> ::iterator __pyx_v_it;
  std::string __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("find_header", 0);

  /* "fastwarc/warc.pyx":257
 *     cdef string find_header(self, const string& header_key, const string& default):
 *         """Return value of first header occurrence or ``default`` (linear complexity)."""
 *         cdef string header_key_lower = str_to_lower(header_key)             # <<<<<<<<<<<<<<
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 */
  __pyx_v_header_key_lower = __pyx_f_8fastwarc_4warc_str_to_lower(__pyx_v_header_key);

  /* "fastwarc/warc.pyx":258
 *         """Return value of first header occurrence or ``default`` (linear complexity)."""
 *         cdef string header_key_lower = str_to_lower(header_key)
 *         cdef vector[str_pair].iterator it = self._headers.begin()             # <<<<<<<<<<<<<<
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 */
  __pyx_v_it = __pyx_v_self->_headers.begin();

  /* "fastwarc/warc.pyx":259
 *         cdef string header_key_lower = str_to_lower(header_key)
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():             # <<<<<<<<<<<<<<
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 *                 return deref(it)[1]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_it != __pyx_v_self->_headers.end()) != 0);
    if (!__pyx_t_1) break;

    /* "fastwarc/warc.pyx":260
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:             # <<<<<<<<<<<<<<
 *                 return deref(it)[1]
 *             inc(it)
 */
    __pyx_t_1 = ((__pyx_f_8fastwarc_4warc_str_to_lower((*__pyx_v_it).f0) == __pyx_v_header_key_lower) != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":261
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 *                 return deref(it)[1]             # <<<<<<<<<<<<<<
 *             inc(it)
 * 
 */
      __pyx_r = (*__pyx_v_it).f1;
      goto __pyx_L0;

      /* "fastwarc/warc.pyx":260
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:             # <<<<<<<<<<<<<<
 *                 return deref(it)[1]
 *             inc(it)
 */
    }

    /* "fastwarc/warc.pyx":262
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 *                 return deref(it)[1]
 *             inc(it)             # <<<<<<<<<<<<<<
 * 
 *         return default
 */
    (void)((++__pyx_v_it));
  }

  /* "fastwarc/warc.pyx":264
 *             inc(it)
 * 
 *         return default             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_header(self, const string& header_key, const string& header_value):
 */
  __pyx_r = __pyx_v_default;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":255
 *         self._status_line = status_line
 * 
 *     cdef string find_header(self, const string& header_key, const string& default):             # <<<<<<<<<<<<<<
 *         """Return value of first header occurrence or ``default`` (linear complexity)."""
 *         cdef string header_key_lower = str_to_lower(header_key)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":266
 *         return default
 * 
 *     cdef void set_header(self, const string& header_key, const string& header_value):             # <<<<<<<<<<<<<<
 *         """Set new header or overwrite existing header if it exists."""
 *         self._dict_cache_stale = True
 */

static void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_key, std::string const &__pyx_v_header_value) {
  std::string __pyx_v_header_key_lower;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair> ::iterator __pyx_v_it;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_8fastwarc_4warc_str_pair __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_header", 0);

  /* "fastwarc/warc.pyx":268
 *     cdef void set_header(self, const string& header_key, const string& header_value):
 *         """Set new header or overwrite existing header if it exists."""
 *         self._dict_cache_stale = True             # <<<<<<<<<<<<<<
 *         cdef string header_key_lower = str_to_lower(header_key)
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 */
  __pyx_v_self->_dict_cache_stale = 1;

  /* "fastwarc/warc.pyx":269
 *         """Set new header or overwrite existing header if it exists."""
 *         self._dict_cache_stale = True
 *         cdef string header_key_lower = str_to_lower(header_key)             # <<<<<<<<<<<<<<
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 */
  __pyx_v_header_key_lower = __pyx_f_8fastwarc_4warc_str_to_lower(__pyx_v_header_key);

  /* "fastwarc/warc.pyx":270
 *         self._dict_cache_stale = True
 *         cdef string header_key_lower = str_to_lower(header_key)
 *         cdef vector[str_pair].iterator it = self._headers.begin()             # <<<<<<<<<<<<<<
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 */
  __pyx_v_it = __pyx_v_self->_headers.begin();

  /* "fastwarc/warc.pyx":271
 *         cdef string header_key_lower = str_to_lower(header_key)
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():             # <<<<<<<<<<<<<<
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 *                 deref(it)[1] = header_value
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_it != __pyx_v_self->_headers.end()) != 0);
    if (!__pyx_t_1) break;

    /* "fastwarc/warc.pyx":272
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:             # <<<<<<<<<<<<<<
 *                 deref(it)[1] = header_value
 *                 return
 */
    __pyx_t_1 = ((__pyx_f_8fastwarc_4warc_str_to_lower((*__pyx_v_it).f0) == __pyx_v_header_key_lower) != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":273
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 *                 deref(it)[1] = header_value             # <<<<<<<<<<<<<<
 *                 return
 *             inc(it)
 */
      (*__pyx_v_it).f1 = __pyx_v_header_value;

      /* "fastwarc/warc.pyx":274
 *             if str_to_lower(deref(it)[0]) == header_key_lower:
 *                 deref(it)[1] = header_value
 *                 return             # <<<<<<<<<<<<<<
 *             inc(it)
 * 
 */
      goto __pyx_L0;

      /* "fastwarc/warc.pyx":272
 *         cdef vector[str_pair].iterator it = self._headers.begin()
 *         while it != self._headers.end():
 *             if str_to_lower(deref(it)[0]) == header_key_lower:             # <<<<<<<<<<<<<<
 *                 deref(it)[1] = header_value
 *                 return
 */
    }

    /* "fastwarc/warc.pyx":275
 *                 deref(it)[1] = header_value
 *                 return
 *             inc(it)             # <<<<<<<<<<<<<<
 * 
 *         self._headers.push_back((header_key, header_value))
 */
    (void)((++__pyx_v_it));
  }

  /* "fastwarc/warc.pyx":277
 *             inc(it)
 * 
 *         self._headers.push_back((header_key, header_value))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void append_header(self, const string& header_key, const string& header_value):
 */
  __pyx_t_2.f0 = __pyx_v_header_key;
  __pyx_t_2.f1 = __pyx_v_header_value;
  try {
    __pyx_v_self->_headers.push_back(__pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 277, __pyx_L1_error)
  }

  /* "fastwarc/warc.pyx":266
 *         return default
 * 
 *     cdef void set_header(self, const string& header_key, const string& header_value):             # <<<<<<<<<<<<<<
 *         """Set new header or overwrite existing header if it exists."""
 *         self._dict_cache_stale = True
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fastwarc.warc.WarcHeaderMap.set_header", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fastwarc/warc.pyx":279
 *         self._headers.push_back((header_key, header_value))
 * 
 *     cdef inline void append_header(self, const string& header_key, const string& header_value):             # <<<<<<<<<<<<<<
 *         """Append header (use if header key is not unique)."""
 *         self._headers.push_back((header_key, header_value))
 */

static CYTHON_INLINE void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_key, std::string const &__pyx_v_header_value) {
  __Pyx_RefNannyDeclarations
  __pyx_t_8fastwarc_4warc_str_pair __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append_header", 0);

  /* "fastwarc/warc.pyx":281
 *     cdef inline void append_header(self, const string& header_key, const string& header_value):
 *         """Append header (use if header key is not unique)."""
 *         self._headers.push_back((header_key, header_value))             # <<<<<<<<<<<<<<
 *         self._dict_cache_stale = True
 * 
 */
  __pyx_t_1.f0 = __pyx_v_header_key;
  __pyx_t_1.f1 = __pyx_v_header_value;
  try {
    __pyx_v_self->_headers.push_back(__pyx_t_1);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 281, __pyx_L1_error)
  }

  /* "fastwarc/warc.pyx":282
 *         """Append header (use if header key is not unique)."""
 *         self._headers.push_back((header_key, header_value))
 *         self._dict_cache_stale = True             # <<<<<<<<<<<<<<
 * 
 *     cdef void add_continuation(self, const string& header_continuation_value):
 */
  __pyx_v_self->_dict_cache_stale = 1;

  /* "fastwarc/warc.pyx":279
 *         self._headers.push_back((header_key, header_value))
 * 
 *     cdef inline void append_header(self, const string& header_key, const string& header_value):             # <<<<<<<<<<<<<<
 *         """Append header (use if header key is not unique)."""
 *         self._headers.push_back((header_key, header_value))
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fastwarc.warc.WarcHeaderMap.append_header", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fastwarc/warc.pyx":284
 *         self._dict_cache_stale = True
 * 
 *     cdef void add_continuation(self, const string& header_continuation_value):             # <<<<<<<<<<<<<<
 *         """Append value to previous header."""
 *         if not self._headers.empty():
 */

static void __pyx_f_8fastwarc_4warc_13WarcHeaderMap_add_continuation(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, std::string const &__pyx_v_header_continuation_value) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_8fastwarc_4warc_str_pair __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_continuation", 0);

  /* "fastwarc/warc.pyx":286
 *     cdef void add_continuation(self, const string& header_continuation_value):
 *         """Append value to previous header."""
 *         if not self._headers.empty():             # <<<<<<<<<<<<<<
 *             self._headers.back()[1].append(<char*>b' ')
 *             self._headers.back()[1].append(header_continuation_value)
 */
  __pyx_t_1 = ((!(__pyx_v_self->_headers.empty() != 0)) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":287
 *         """Append value to previous header."""
 *         if not self._headers.empty():
 *             self._headers.back()[1].append(<char*>b' ')             # <<<<<<<<<<<<<<
 *             self._headers.back()[1].append(header_continuation_value)
 *         else:
 */
    try {
      __pyx_v_self->_headers.back().f1.append(((char *)((char *)" ")));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 287, __pyx_L1_error)
    }

    /* "fastwarc/warc.pyx":288
 *         if not self._headers.empty():
 *             self._headers.back()[1].append(<char*>b' ')
 *             self._headers.back()[1].append(header_continuation_value)             # <<<<<<<<<<<<<<
 *         else:
 *             # This should no happen, but what can we do?!
 */
    try {
      __pyx_v_self->_headers.back().f1.append(__pyx_v_header_continuation_value);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 288, __pyx_L1_error)
    }

    /* "fastwarc/warc.pyx":286
 *     cdef void add_continuation(self, const string& header_continuation_value):
 *         """Append value to previous header."""
 *         if not self._headers.empty():             # <<<<<<<<<<<<<<
 *             self._headers.back()[1].append(<char*>b' ')
 *             self._headers.back()[1].append(header_continuation_value)
 */
    goto __pyx_L3;
  }

  /* "fastwarc/warc.pyx":291
 *         else:
 *             # This should no happen, but what can we do?!
 *             self._headers.push_back((<char*>b'', header_continuation_value))             # <<<<<<<<<<<<<<
 *         self._dict_cache_stale = True
 * 
 */
  /*else*/ {
    __pyx_t_2.f0 = ((char *)((char *)""));
    __pyx_t_2.f1 = __pyx_v_header_continuation_value;
    try {
      __pyx_v_self->_headers.push_back(__pyx_t_2);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 291, __pyx_L1_error)
    }
  }
  __pyx_L3:;

  /* "fastwarc/warc.pyx":292
 *             # This should no happen, but what can we do?!
 *             self._headers.push_back((<char*>b'', header_continuation_value))
 *         self._dict_cache_stale = True             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_dict_cache_stale = 1;

  /* "fastwarc/warc.pyx":284
 *         self._dict_cache_stale = True
 * 
 *     cdef void add_continuation(self, const string& header_continuation_value):             # <<<<<<<<<<<<<<
 *         """Append value to previous header."""
 *         if not self._headers.empty():
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fastwarc.warc.WarcHeaderMap.add_continuation", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_30__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_30__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_29__reduce_cython__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_29__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_32__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_32__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_13WarcHeaderMap_31__setstate_cython__(((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_13WarcHeaderMap_31__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcHeaderMap.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":299
 *     """A WARC record."""
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._record_type = unknown
 *         self._is_http = False
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_10WarcRecord_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8fastwarc_4warc_10WarcRecord_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord___cinit__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_10WarcRecord___cinit__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "fastwarc/warc.pyx":300
 * 
 *     def __cinit__(self):
 *         self._record_type = unknown             # <<<<<<<<<<<<<<
 *         self._is_http = False
 *         self._http_parsed = False
 */
  __pyx_v_self->_record_type = __pyx_e_8fastwarc_4warc_unknown;

  /* "fastwarc/warc.pyx":301
 *     def __cinit__(self):
 *         self._record_type = unknown
 *         self._is_http = False             # <<<<<<<<<<<<<<
 *         self._http_parsed = False
 *         self._content_length = 0
 */
  __pyx_v_self->_is_http = 0;

  /* "fastwarc/warc.pyx":302
 *         self._record_type = unknown
 *         self._is_http = False
 *         self._http_parsed = False             # <<<<<<<<<<<<<<
 *         self._content_length = 0
 *         self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')
 */
  __pyx_v_self->_http_parsed = 0;

  /* "fastwarc/warc.pyx":303
 *         self._is_http = False
 *         self._http_parsed = False
 *         self._content_length = 0             # <<<<<<<<<<<<<<
 *         self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')
 *         self._http_headers = None
 */
  __pyx_v_self->_content_length = 0;

  /* "fastwarc/warc.pyx":304
 *         self._http_parsed = False
 *         self._content_length = 0
 *         self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')             # <<<<<<<<<<<<<<
 *         self._http_headers = None
 *         self._stream_pos = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8fastwarc_4warc_WarcHeaderMap(((PyTypeObject *)__pyx_ptype_8fastwarc_4warc_WarcHeaderMap), __pyx_tuple__4, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_GOTREF(__pyx_v_self->_headers);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_headers));
  __pyx_v_self->_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":305
 *         self._content_length = 0
 *         self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')
 *         self._http_headers = None             # <<<<<<<<<<<<<<
 *         self._stream_pos = 0
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_http_headers);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_http_headers));
  __pyx_v_self->_http_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)Py_None);

  /* "fastwarc/warc.pyx":306
 *         self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')
 *         self._http_headers = None
 *         self._stream_pos = 0             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->_stream_pos = 0;

  /* "fastwarc/warc.pyx":299
 *     """A WARC record."""
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._record_type = unknown
 *         self._is_http = False
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":309
 * 
 *     @property
 *     def record_id(self) -> str:             # <<<<<<<<<<<<<<
 *         """Record ID (same as ``headers['WARC-Record'ID']``."""
 *         return self._headers['WARC-Record-ID']
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_9record_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_9record_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_9record_id___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_9record_id___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":311
 *     def record_id(self) -> str:
 *         """Record ID (same as ``headers['WARC-Record'ID']``."""
 *         return self._headers['WARC-Record-ID']             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->_headers), __pyx_kp_u_WARC_Record_ID); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":309
 * 
 *     @property
 *     def record_id(self) -> str:             # <<<<<<<<<<<<<<
 *         """Record ID (same as ``headers['WARC-Record'ID']``."""
 *         return self._headers['WARC-Record-ID']
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.record_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":314
 * 
 *     @property
 *     def record_type(self) -> WarcRecordType:             # <<<<<<<<<<<<<<
 *         """Record type (same as ``headers['WARC-Type']``."""
 *         return self._record_type
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_11record_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_11record_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_11record_type___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_11record_type___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":316
 *     def record_type(self) -> WarcRecordType:
 *         """Record type (same as ``headers['WARC-Type']``."""
 *         return self._record_type             # <<<<<<<<<<<<<<
 * 
 *     @record_type.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_v_self->_record_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":314
 * 
 *     @property
 *     def record_type(self) -> WarcRecordType:             # <<<<<<<<<<<<<<
 *         """Record type (same as ``headers['WARC-Type']``."""
 *         return self._record_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.record_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":319
 * 
 *     @record_type.setter
 *     def record_type(self, WarcRecordType record_type):             # <<<<<<<<<<<<<<
 *         self._record_type = record_type
 *         self._headers['WARC-Type'] = _enum_record_type_to_str(record_type)
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_10WarcRecord_11record_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_record_type); /*proto*/
static int __pyx_pw_8fastwarc_4warc_10WarcRecord_11record_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_record_type) {
  enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_record_type); {
    __pyx_v_record_type = ((enum __pyx_t_8fastwarc_4warc_WarcRecordType)__Pyx_PyInt_As_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_arg_record_type)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 319, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.record_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_11record_type_2__set__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), ((enum __pyx_t_8fastwarc_4warc_WarcRecordType)__pyx_v_record_type));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_10WarcRecord_11record_type_2__set__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "fastwarc/warc.pyx":320
 *     @record_type.setter
 *     def record_type(self, WarcRecordType record_type):
 *         self._record_type = record_type             # <<<<<<<<<<<<<<
 *         self._headers['WARC-Type'] = _enum_record_type_to_str(record_type)
 * 
 */
  __pyx_v_self->_record_type = __pyx_v_record_type;

  /* "fastwarc/warc.pyx":321
 *     def record_type(self, WarcRecordType record_type):
 *         self._record_type = record_type
 *         self._headers['WARC-Type'] = _enum_record_type_to_str(record_type)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_f_8fastwarc_4warc__enum_record_type_to_str(__pyx_v_record_type)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self->_headers), __pyx_kp_u_WARC_Type, __pyx_t_1) < 0)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":319
 * 
 *     @record_type.setter
 *     def record_type(self, WarcRecordType record_type):             # <<<<<<<<<<<<<<
 *         self._record_type = record_type
 *         self._headers['WARC-Type'] = _enum_record_type_to_str(record_type)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.record_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":324
 * 
 *     @property
 *     def headers(self) -> WarcHeaderMap:             # <<<<<<<<<<<<<<
 *         """WARC record headers."""
 *         return self._headers
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7headers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7headers_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_7headers___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_7headers___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":326
 *     def headers(self) -> WarcHeaderMap:
 *         """WARC record headers."""
 *         return self._headers             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_headers));
  __pyx_r = ((PyObject *)__pyx_v_self->_headers);
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":324
 * 
 *     @property
 *     def headers(self) -> WarcHeaderMap:             # <<<<<<<<<<<<<<
 *         """WARC record headers."""
 *         return self._headers
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":329
 * 
 *     @property
 *     def is_http(self) -> bool:             # <<<<<<<<<<<<<<
 *         """Whether record is an HTTP record"""
 *         return self._is_http
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7is_http_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7is_http_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_7is_http___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_7is_http___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":331
 *     def is_http(self) -> bool:
 *         """Whether record is an HTTP record"""
 *         return self._is_http             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_is_http); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":329
 * 
 *     @property
 *     def is_http(self) -> bool:             # <<<<<<<<<<<<<<
 *         """Whether record is an HTTP record"""
 *         return self._is_http
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.is_http.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":334
 * 
 *     @property
 *     def is_http_parsed(self) -> bool:             # <<<<<<<<<<<<<<
 *         """Whether HTTP headers have been parsed."""
 *         return self._http_parsed
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_14is_http_parsed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_14is_http_parsed_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_14is_http_parsed___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_14is_http_parsed___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":336
 *     def is_http_parsed(self) -> bool:
 *         """Whether HTTP headers have been parsed."""
 *         return self._http_parsed             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_http_parsed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":334
 * 
 *     @property
 *     def is_http_parsed(self) -> bool:             # <<<<<<<<<<<<<<
 *         """Whether HTTP headers have been parsed."""
 *         return self._http_parsed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.is_http_parsed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":339
 * 
 *     @property
 *     def http_headers(self) -> Optional[WarcHeaderMap]:             # <<<<<<<<<<<<<<
 *         """HTTP headers if record is an HTTP record and HTTP headers have been parsed yet."""
 *         return self._http_headers
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_12http_headers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_12http_headers_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_12http_headers___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_12http_headers___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":341
 *     def http_headers(self) -> Optional[WarcHeaderMap]:
 *         """HTTP headers if record is an HTTP record and HTTP headers have been parsed yet."""
 *         return self._http_headers             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_http_headers));
  __pyx_r = ((PyObject *)__pyx_v_self->_http_headers);
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":339
 * 
 *     @property
 *     def http_headers(self) -> Optional[WarcHeaderMap]:             # <<<<<<<<<<<<<<
 *         """HTTP headers if record is an HTTP record and HTTP headers have been parsed yet."""
 *         return self._http_headers
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":344
 * 
 *     @property
 *     def http_charset(self) -> Optional[str]:             # <<<<<<<<<<<<<<
 *         """
 *         HTTP charset/encoding as returned by the server or ``None`` if no valid charset is set.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_12http_charset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_12http_charset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_12http_charset___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_12http_charset___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  std::string __pyx_v_content_type;
  size_t __pyx_v_pos;
  size_t __pyx_v_pos_end;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":349
 *         A returned string is guaranteed to be a valid Python encoding name.
 *         """
 *         if not self._http_parsed or self._http_charset == <char*>b'_':             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->_http_parsed != 0)) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->_http_charset == ((char *)((char *)"_"))) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":350
 *         """
 *         if not self._http_parsed or self._http_charset == <char*>b'_':
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         if not self._http_charset.empty():
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":349
 *         A returned string is guaranteed to be a valid Python encoding name.
 *         """
 *         if not self._http_parsed or self._http_charset == <char*>b'_':             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  }

  /* "fastwarc/warc.pyx":352
 *             return None
 * 
 *         if not self._http_charset.empty():             # <<<<<<<<<<<<<<
 *             return self._http_charset.decode()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_http_charset.empty() != 0)) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":353
 * 
 *         if not self._http_charset.empty():
 *             return self._http_charset.decode()             # <<<<<<<<<<<<<<
 * 
 *         cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_decode_cpp_string(__pyx_v_self->_http_charset, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":352
 *             return None
 * 
 *         if not self._http_charset.empty():             # <<<<<<<<<<<<<<
 *             return self._http_charset.decode()
 * 
 */
  }

  /* "fastwarc/warc.pyx":355
 *             return self._http_charset.decode()
 * 
 *         cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')             # <<<<<<<<<<<<<<
 *         cdef size_t pos = content_type.find(<char*>b'charset=')
 *         if pos == strnpos:
 */
  __pyx_v_content_type = ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_http_headers->__pyx_vtab)->find_header(__pyx_v_self->_http_headers, ((char *)((char *)"content-type")), ((char *)((char *)"")));

  /* "fastwarc/warc.pyx":356
 * 
 *         cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
 *         cdef size_t pos = content_type.find(<char*>b'charset=')             # <<<<<<<<<<<<<<
 *         if pos == strnpos:
 *             return None
 */
  __pyx_v_pos = __pyx_v_content_type.find(((char *)((char *)"charset=")));

  /* "fastwarc/warc.pyx":357
 *         cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
 *         cdef size_t pos = content_type.find(<char*>b'charset=')
 *         if pos == strnpos:             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  __pyx_t_1 = ((__pyx_v_pos == std::string::npos) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":358
 *         cdef size_t pos = content_type.find(<char*>b'charset=')
 *         if pos == strnpos:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         pos += 8
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":357
 *         cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
 *         cdef size_t pos = content_type.find(<char*>b'charset=')
 *         if pos == strnpos:             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  }

  /* "fastwarc/warc.pyx":360
 *             return None
 * 
 *         pos += 8             # <<<<<<<<<<<<<<
 *         cdef size_t pos_end = content_type.find(<char*>b';', pos)
 *         if pos_end != strnpos:
 */
  __pyx_v_pos = (__pyx_v_pos + 8);

  /* "fastwarc/warc.pyx":361
 * 
 *         pos += 8
 *         cdef size_t pos_end = content_type.find(<char*>b';', pos)             # <<<<<<<<<<<<<<
 *         if pos_end != strnpos:
 *             pos_end = pos_end - pos
 */
  __pyx_v_pos_end = __pyx_v_content_type.find(((char *)((char *)";")), __pyx_v_pos);

  /* "fastwarc/warc.pyx":362
 *         pos += 8
 *         cdef size_t pos_end = content_type.find(<char*>b';', pos)
 *         if pos_end != strnpos:             # <<<<<<<<<<<<<<
 *             pos_end = pos_end - pos
 * 
 */
  __pyx_t_1 = ((__pyx_v_pos_end != std::string::npos) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":363
 *         cdef size_t pos_end = content_type.find(<char*>b';', pos)
 *         if pos_end != strnpos:
 *             pos_end = pos_end - pos             # <<<<<<<<<<<<<<
 * 
 *         self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))
 */
    __pyx_v_pos_end = (__pyx_v_pos_end - __pyx_v_pos);

    /* "fastwarc/warc.pyx":362
 *         pos += 8
 *         cdef size_t pos_end = content_type.find(<char*>b';', pos)
 *         if pos_end != strnpos:             # <<<<<<<<<<<<<<
 *             pos_end = pos_end - pos
 * 
 */
  }

  /* "fastwarc/warc.pyx":365
 *             pos_end = pos_end - pos
 * 
 *         self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))             # <<<<<<<<<<<<<<
 *         try:
 *             ''.encode(encoding=self._http_charset.decode())
 */
  try {
    __pyx_t_4 = __pyx_v_content_type.substr(__pyx_v_pos, __pyx_v_pos_end);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 365, __pyx_L1_error)
  }
  __pyx_v_self->_http_charset = __pyx_f_8fastwarc_4warc_str_to_lower(__pyx_f_8fastwarc_4warc_strip_str(__pyx_t_4));

  /* "fastwarc/warc.pyx":366
 * 
 *         self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))
 *         try:             # <<<<<<<<<<<<<<
 *             ''.encode(encoding=self._http_charset.decode())
 *         except LookupError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "fastwarc/warc.pyx":367
 *         self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))
 *         try:
 *             ''.encode(encoding=self._http_charset.decode())             # <<<<<<<<<<<<<<
 *         except LookupError:
 *             self._http_charset = <char*>b'_'
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__5, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 367, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_decode_cpp_string(__pyx_v_self->_http_charset, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 367, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_encoding, __pyx_t_9) < 0) __PYX_ERR(0, 367, __pyx_L9_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 367, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fastwarc/warc.pyx":366
 * 
 *         self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))
 *         try:             # <<<<<<<<<<<<<<
 *             ''.encode(encoding=self._http_charset.decode())
 *         except LookupError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "fastwarc/warc.pyx":368
 *         try:
 *             ''.encode(encoding=self._http_charset.decode())
 *         except LookupError:             # <<<<<<<<<<<<<<
 *             self._http_charset = <char*>b'_'
 *             return None
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_LookupError);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("fastwarc.warc.WarcRecord.http_charset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(0, 368, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_3);

      /* "fastwarc/warc.pyx":369
 *             ''.encode(encoding=self._http_charset.decode())
 *         except LookupError:
 *             self._http_charset = <char*>b'_'             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
      __pyx_v_self->_http_charset = ((char *)((char *)"_"));

      /* "fastwarc/warc.pyx":370
 *         except LookupError:
 *             self._http_charset = <char*>b'_'
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         return self._http_charset.decode()
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L12_except_return;
    }
    goto __pyx_L11_except_error;
    __pyx_L11_except_error:;

    /* "fastwarc/warc.pyx":366
 * 
 *         self._http_charset = str_to_lower(strip_str(content_type.substr(pos, pos_end)))
 *         try:             # <<<<<<<<<<<<<<
 *             ''.encode(encoding=self._http_charset.decode())
 *         except LookupError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L12_except_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L0;
    __pyx_L14_try_end:;
  }

  /* "fastwarc/warc.pyx":372
 *             return None
 * 
 *         return self._http_charset.decode()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_decode_cpp_string(__pyx_v_self->_http_charset, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":344
 * 
 *     @property
 *     def http_charset(self) -> Optional[str]:             # <<<<<<<<<<<<<<
 *         """
 *         HTTP charset/encoding as returned by the server or ``None`` if no valid charset is set.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.http_charset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":375
 * 
 *     @property
 *     def content_length(self) -> int:             # <<<<<<<<<<<<<<
 *         """Remaining WARC length in bytes (not necessarily the same as the ``Content-Length`` header)."""
 *         return self._content_length
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_14content_length_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_14content_length_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_14content_length___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_14content_length___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":377
 *     def content_length(self) -> int:
 *         """Remaining WARC length in bytes (not necessarily the same as the ``Content-Length`` header)."""
 *         return self._content_length             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_content_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":375
 * 
 *     @property
 *     def content_length(self) -> int:             # <<<<<<<<<<<<<<
 *         """Remaining WARC length in bytes (not necessarily the same as the ``Content-Length`` header)."""
 *         return self._content_length
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.content_length.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":380
 * 
 *     @property
 *     def reader(self) -> BufferedReader:             # <<<<<<<<<<<<<<
 *         """Reader for the remaining WARC record content."""
 *         return self._reader
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_6reader_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_6reader_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_6reader___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_6reader___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":382
 *     def reader(self) -> BufferedReader:
 *         """Reader for the remaining WARC record content."""
 *         return self._reader             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_reader));
  __pyx_r = ((PyObject *)__pyx_v_self->_reader);
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":380
 * 
 *     @property
 *     def reader(self) -> BufferedReader:             # <<<<<<<<<<<<<<
 *         """Reader for the remaining WARC record content."""
 *         return self._reader
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":385
 * 
 *     @property
 *     def stream_pos(self) -> int:             # <<<<<<<<<<<<<<
 *         """WARC record start offset in the original (uncompressed) stream."""
 *         return self._stream_pos
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_10stream_pos_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_10stream_pos_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_10stream_pos___get__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_10stream_pos___get__(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fastwarc/warc.pyx":387
 *     def stream_pos(self) -> int:
 *         """WARC record start offset in the original (uncompressed) stream."""
 *         return self._stream_pos             # <<<<<<<<<<<<<<
 * 
 *     cpdef void init_headers(self, size_t content_length, WarcRecordType record_type=no_type, bytes record_urn=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_stream_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":385
 * 
 *     @property
 *     def stream_pos(self) -> int:             # <<<<<<<<<<<<<<
 *         """WARC record start offset in the original (uncompressed) stream."""
 *         return self._stream_pos
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.stream_pos.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":389
 *         return self._stream_pos
 * 
 *     cpdef void init_headers(self, size_t content_length, WarcRecordType record_type=no_type, bytes record_urn=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize mandatory headers in a fresh :class:`WarcRecord` instance.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_3init_headers(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_8fastwarc_4warc_10WarcRecord_init_headers(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, size_t __pyx_v_content_length, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers *__pyx_optional_args) {
  enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type = __pyx_k__6;
  PyObject *__pyx_v_record_urn = ((PyObject*)Py_None);
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_t_11;
  std::string __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_headers", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_record_type = __pyx_optional_args->record_type;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_record_urn = __pyx_optional_args->record_urn;
      }
    }
  }
  __Pyx_INCREF(__pyx_v_record_urn);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_init_headers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_10WarcRecord_3init_headers)) {
        __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_content_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_v_record_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_v_record_urn};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_v_record_urn};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 389, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __Pyx_INCREF(__pyx_v_record_urn);
          __Pyx_GIVEREF(__pyx_v_record_urn);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_v_record_urn);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":400
 *         :type record_urn: bytes, optional
 *         """
 *         if record_urn is None:             # <<<<<<<<<<<<<<
 *             record_urn = uuid.uuid4().urn.encode()
 * 
 */
  __pyx_t_9 = (__pyx_v_record_urn == ((PyObject*)Py_None));
  __pyx_t_10 = (__pyx_t_9 != 0);
  if (__pyx_t_10) {

    /* "fastwarc/warc.pyx":401
 *         """
 *         if record_urn is None:
 *             record_urn = uuid.uuid4().urn.encode()             # <<<<<<<<<<<<<<
 * 
 *         if record_type == no_type:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_uuid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_uuid4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_urn); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_record_urn, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "fastwarc/warc.pyx":400
 *         :type record_urn: bytes, optional
 *         """
 *         if record_urn is None:             # <<<<<<<<<<<<<<
 *             record_urn = uuid.uuid4().urn.encode()
 * 
 */
  }

  /* "fastwarc/warc.pyx":403
 *             record_urn = uuid.uuid4().urn.encode()
 * 
 *         if record_type == no_type:             # <<<<<<<<<<<<<<
 *             record_type = self.record_type
 *         if record_type == any_type or record_type == no_type:
 */
  __pyx_t_10 = ((__pyx_v_record_type == __pyx_e_8fastwarc_4warc_no_type) != 0);
  if (__pyx_t_10) {

    /* "fastwarc/warc.pyx":404
 * 
 *         if record_type == no_type:
 *             record_type = self.record_type             # <<<<<<<<<<<<<<
 *         if record_type == any_type or record_type == no_type:
 *             record_type = unknown
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_record_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = ((enum __pyx_t_8fastwarc_4warc_WarcRecordType)__Pyx_PyInt_As_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_record_type = __pyx_t_11;

    /* "fastwarc/warc.pyx":403
 *             record_urn = uuid.uuid4().urn.encode()
 * 
 *         if record_type == no_type:             # <<<<<<<<<<<<<<
 *             record_type = self.record_type
 *         if record_type == any_type or record_type == no_type:
 */
  }

  /* "fastwarc/warc.pyx":405
 *         if record_type == no_type:
 *             record_type = self.record_type
 *         if record_type == any_type or record_type == no_type:             # <<<<<<<<<<<<<<
 *             record_type = unknown
 *         self.record_type = record_type
 */
  switch (__pyx_v_record_type) {
    case __pyx_e_8fastwarc_4warc_any_type:
    case __pyx_e_8fastwarc_4warc_no_type:

    /* "fastwarc/warc.pyx":406
 *             record_type = self.record_type
 *         if record_type == any_type or record_type == no_type:
 *             record_type = unknown             # <<<<<<<<<<<<<<
 *         self.record_type = record_type
 * 
 */
    __pyx_v_record_type = __pyx_e_8fastwarc_4warc_unknown;

    /* "fastwarc/warc.pyx":405
 *         if record_type == no_type:
 *             record_type = self.record_type
 *         if record_type == any_type or record_type == no_type:             # <<<<<<<<<<<<<<
 *             record_type = unknown
 *         self.record_type = record_type
 */
    break;
    default: break;
  }

  /* "fastwarc/warc.pyx":407
 *         if record_type == any_type or record_type == no_type:
 *             record_type = unknown
 *         self.record_type = record_type             # <<<<<<<<<<<<<<
 * 
 *         self._headers.clear()
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_v_record_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_record_type, __pyx_t_1) < 0) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":409
 *         self.record_type = record_type
 * 
 *         self._headers.clear()             # <<<<<<<<<<<<<<
 *         self._headers.set_status_line(<char*>b'WARC/1.1')
 *         self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))
 */
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_clear(__pyx_v_self->_headers);

  /* "fastwarc/warc.pyx":410
 * 
 *         self._headers.clear()
 *         self._headers.set_status_line(<char*>b'WARC/1.1')             # <<<<<<<<<<<<<<
 *         self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))
 *         self._headers.append_header(<char*>b'WARC-Date', datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())
 */
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line(__pyx_v_self->_headers, ((char *)((char *)"WARC/1.1")));

  /* "fastwarc/warc.pyx":411
 *         self._headers.clear()
 *         self._headers.set_status_line(<char*>b'WARC/1.1')
 *         self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))             # <<<<<<<<<<<<<<
 *         self._headers.append_header(<char*>b'WARC-Date', datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())
 *         self._headers.append_header(<char*>b'WARC-Record-ID', b''.join((b'<', record_urn, b'>')))
 */
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Type")), __pyx_f_8fastwarc_4warc__enum_record_type_to_str(__pyx_v_record_type));

  /* "fastwarc/warc.pyx":412
 *         self._headers.set_status_line(<char*>b'WARC/1.1')
 *         self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))
 *         self._headers.append_header(<char*>b'WARC-Date', datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())             # <<<<<<<<<<<<<<
 *         self._headers.append_header(<char*>b'WARC-Record-ID', b''.join((b'<', record_urn, b'>')))
 *         self._headers.append_header(<char*>b'Content-Length', to_string(content_length))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_utcnow); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_strftime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_8, __pyx_kp_u_Y_m_dT_H_M_SZ) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_Y_m_dT_H_M_SZ);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __pyx_convert_string_from_py_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Date")), __pyx_t_12);

  /* "fastwarc/warc.pyx":413
 *         self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))
 *         self._headers.append_header(<char*>b'WARC-Date', datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())
 *         self._headers.append_header(<char*>b'WARC-Record-ID', b''.join((b'<', record_urn, b'>')))             # <<<<<<<<<<<<<<
 *         self._headers.append_header(<char*>b'Content-Length', to_string(content_length))
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_kp_b__7);
  __Pyx_GIVEREF(__pyx_kp_b__7);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_b__7);
  __Pyx_INCREF(__pyx_v_record_urn);
  __Pyx_GIVEREF(__pyx_v_record_urn);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_record_urn);
  __Pyx_INCREF(__pyx_kp_b__8);
  __Pyx_GIVEREF(__pyx_kp_b__8);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_b__8);
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__5, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __pyx_convert_string_from_py_std__in_string(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Record-ID")), __pyx_t_12);

  /* "fastwarc/warc.pyx":414
 *         self._headers.append_header(<char*>b'WARC-Date', datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())
 *         self._headers.append_header(<char*>b'WARC-Record-ID', b''.join((b'<', record_urn, b'>')))
 *         self._headers.append_header(<char*>b'Content-Length', to_string(content_length))             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_bytes_content(self, bytes b):
 */
  __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(__pyx_v_self->_headers, ((char *)((char *)"Content-Length")), std::to_string(__pyx_v_content_length));

  /* "fastwarc/warc.pyx":389
 *         return self._stream_pos
 * 
 *     cpdef void init_headers(self, size_t content_length, WarcRecordType record_type=no_type, bytes record_urn=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize mandatory headers in a fresh :class:`WarcRecord` instance.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord.init_headers", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_record_urn);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_3init_headers(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10WarcRecord_2init_headers[] = "\n        Initialize mandatory headers in a fresh :class:`WarcRecord` instance.\n        \n        :param content_length: WARC record body length in bytes\n        :type content_length: int\n        :param record_type: WARC-Type\n        :type record_type: WarcRecordType, optional, default: no_type\n        :param record_urn: WARC-Record-ID as URN without ``'<'``, ``'>'`` (if unset, a random URN will be generated)\n        :type record_urn: bytes, optional\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_3init_headers(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_content_length;
  enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type;
  PyObject *__pyx_v_record_urn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_headers (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_content_length,&__pyx_n_s_record_type,&__pyx_n_s_record_urn,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_content_length)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_record_type);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_record_urn);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "init_headers") < 0)) __PYX_ERR(0, 389, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_content_length = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_content_length == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_record_type = ((enum __pyx_t_8fastwarc_4warc_WarcRecordType)__Pyx_PyInt_As_enum____pyx_t_8fastwarc_4warc_WarcRecordType(values[1])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    } else {
      __pyx_v_record_type = __pyx_k__6;
    }
    __pyx_v_record_urn = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("init_headers", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 389, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.init_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record_urn), (&PyBytes_Type), 1, "record_urn", 1))) __PYX_ERR(0, 389, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_2init_headers(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), __pyx_v_content_length, __pyx_v_record_type, __pyx_v_record_urn);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_2init_headers(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, size_t __pyx_v_content_length, enum __pyx_t_8fastwarc_4warc_WarcRecordType __pyx_v_record_type, PyObject *__pyx_v_record_urn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_headers", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1.__pyx_n = 2;
  __pyx_t_1.record_type = __pyx_v_record_type;
  __pyx_t_1.record_urn = __pyx_v_record_urn;
  __pyx_vtabptr_8fastwarc_4warc_WarcRecord->init_headers(__pyx_v_self, __pyx_v_content_length, 1, &__pyx_t_1); 
  __pyx_t_2 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.init_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":416
 *         self._headers.append_header(<char*>b'Content-Length', to_string(content_length))
 * 
 *     cpdef void set_bytes_content(self, bytes b):             # <<<<<<<<<<<<<<
 *         """
 *         Set WARC body.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_5set_bytes_content(PyObject *__pyx_v_self, PyObject *__pyx_v_b); /*proto*/
static void __pyx_f_8fastwarc_4warc_10WarcRecord_set_bytes_content(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_b, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_bytes_content", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_bytes_content); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_10WarcRecord_5set_bytes_content)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_b) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_b);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":423
 *         :type b: bytes
 *         """
 *         self._reader = BufferedReader.__new__(BufferedReader, BytesIOStream(b))             # <<<<<<<<<<<<<<
 *         self._content_length = len(b)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_BytesIOStream), __pyx_v_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_BufferedReader), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8fastwarc_9stream_io_BufferedReader)))) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_reader));
  __pyx_v_self->_reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":424
 *         """
 *         self._reader = BufferedReader.__new__(BufferedReader, BytesIOStream(b))
 *         self._content_length = len(b)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void parse_http(self):
 */
  if (unlikely(__pyx_v_b == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 424, __pyx_L1_error)
  }
  __pyx_t_5 = PyBytes_GET_SIZE(__pyx_v_b); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 424, __pyx_L1_error)
  __pyx_v_self->_content_length = __pyx_t_5;

  /* "fastwarc/warc.pyx":416
 *         self._headers.append_header(<char*>b'Content-Length', to_string(content_length))
 * 
 *     cpdef void set_bytes_content(self, bytes b):             # <<<<<<<<<<<<<<
 *         """
 *         Set WARC body.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord.set_bytes_content", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_5set_bytes_content(PyObject *__pyx_v_self, PyObject *__pyx_v_b); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10WarcRecord_4set_bytes_content[] = "\n        Set WARC body.\n        \n        :param b: body as bytes\n        :type b: bytes\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_5set_bytes_content(PyObject *__pyx_v_self, PyObject *__pyx_v_b) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_bytes_content (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_b), (&PyBytes_Type), 1, "b", 1))) __PYX_ERR(0, 416, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_4set_bytes_content(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), ((PyObject*)__pyx_v_b));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_4set_bytes_content(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_bytes_content", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_8fastwarc_4warc_10WarcRecord_set_bytes_content(__pyx_v_self, __pyx_v_b, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.set_bytes_content", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":426
 *         self._content_length = len(b)
 * 
 *     cpdef void parse_http(self):             # <<<<<<<<<<<<<<
 *         """
 *         Parse HTTP headers and advance content reader.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7parse_http(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_8fastwarc_4warc_10WarcRecord_parse_http(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_skip_dispatch) {
  size_t __pyx_v_num_bytes;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  size_t __pyx_t_7;
  struct __pyx_opt_args_8fastwarc_4warc_parse_header_block __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_http", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_parse_http); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_10WarcRecord_7parse_http)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":432
 *         It is safe to call this method multiple times, even if the record is not an HTTP record.
 *         """
 *         if self._http_parsed or not self._is_http:             # <<<<<<<<<<<<<<
 *             return
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
 */
  __pyx_t_6 = (__pyx_v_self->_http_parsed != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = ((!(__pyx_v_self->_is_http != 0)) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "fastwarc/warc.pyx":433
 *         """
 *         if self._http_parsed or not self._is_http:
 *             return             # <<<<<<<<<<<<<<
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
 *         cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)
 */
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":432
 *         It is safe to call this method multiple times, even if the record is not an HTTP record.
 *         """
 *         if self._http_parsed or not self._is_http:             # <<<<<<<<<<<<<<
 *             return
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
 */
  }

  /* "fastwarc/warc.pyx":434
 *         if self._http_parsed or not self._is_http:
 *             return
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')             # <<<<<<<<<<<<<<
 *         cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)
 *         self._content_length = self._content_length - num_bytes
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_8fastwarc_4warc_WarcHeaderMap(((PyTypeObject *)__pyx_ptype_8fastwarc_4warc_WarcHeaderMap), __pyx_tuple__9, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_GOTREF(__pyx_v_self->_http_headers);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_http_headers));
  __pyx_v_self->_http_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":435
 *             return
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
 *         cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)             # <<<<<<<<<<<<<<
 *         self._content_length = self._content_length - num_bytes
 *         self._http_parsed = True
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reader); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8fastwarc_9stream_io_BufferedReader))))) __PYX_ERR(0, 435, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_http_headers);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.has_status_line = 1;
  __pyx_t_7 = __pyx_f_8fastwarc_4warc_parse_header_block(((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)__pyx_t_1), ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_t_2), &__pyx_t_8); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_num_bytes = __pyx_t_7;

  /* "fastwarc/warc.pyx":436
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
 *         cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)
 *         self._content_length = self._content_length - num_bytes             # <<<<<<<<<<<<<<
 *         self._http_parsed = True
 * 
 */
  __pyx_v_self->_content_length = (__pyx_v_self->_content_length - __pyx_v_num_bytes);

  /* "fastwarc/warc.pyx":437
 *         cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)
 *         self._content_length = self._content_length - num_bytes
 *         self._http_parsed = True             # <<<<<<<<<<<<<<
 * 
 *     # noinspection PyTypeChecker
 */
  __pyx_v_self->_http_parsed = 1;

  /* "fastwarc/warc.pyx":426
 *         self._content_length = len(b)
 * 
 *     cpdef void parse_http(self):             # <<<<<<<<<<<<<<
 *         """
 *         Parse HTTP headers and advance content reader.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord.parse_http", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7parse_http(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10WarcRecord_6parse_http[] = "\n        Parse HTTP headers and advance content reader.\n        \n        It is safe to call this method multiple times, even if the record is not an HTTP record.\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_7parse_http(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse_http (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_6parse_http(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_6parse_http(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_http", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_8fastwarc_4warc_10WarcRecord_parse_http(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.parse_http", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":440
 * 
 *     # noinspection PyTypeChecker
 *     cpdef size_t write(self, stream, bint checksum_data=False, size_t chunk_size=16384):             # <<<<<<<<<<<<<<
 *         """
 *         Write WARC record onto a stream.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_9write(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_8fastwarc_4warc_10WarcRecord_write(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_stream, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write *__pyx_optional_args) {
  int __pyx_v_checksum_data = ((int)0);
  size_t __pyx_v_chunk_size = ((size_t)0x4000);
  struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *__pyx_v_block_buf = 0;
  PyObject *__pyx_v_block_digest = NULL;
  PyObject *__pyx_v_payload_digest = NULL;
  std::string __pyx_v_payload_data;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  std::string __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_checksum_data = __pyx_optional_args->checksum_data;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_chunk_size = __pyx_optional_args->chunk_size;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_10WarcRecord_9write)) {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_checksum_data); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_chunk_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_stream, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_stream, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_v_stream);
          __Pyx_GIVEREF(__pyx_v_stream);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_stream);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_9;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":452
 *         """
 *         # If the raw byte content hasn't been parsed, we can simply pass it through
 *         if not checksum_data and not self.http_parsed:             # <<<<<<<<<<<<<<
 *             return self._write_impl(self.reader, stream, True, chunk_size)
 * 
 */
  __pyx_t_11 = ((!(__pyx_v_checksum_data != 0)) != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_10 = __pyx_t_11;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_http_parsed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = ((!__pyx_t_11) != 0);
  __pyx_t_10 = __pyx_t_12;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_10) {

    /* "fastwarc/warc.pyx":453
 *         # If the raw byte content hasn't been parsed, we can simply pass it through
 *         if not checksum_data and not self.http_parsed:
 *             return self._write_impl(self.reader, stream, True, chunk_size)             # <<<<<<<<<<<<<<
 * 
 *         # Otherwise read everything into memory for content-length correction and checksumming
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reader); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *)__pyx_v_self->__pyx_vtab)->_write_impl(__pyx_v_self, __pyx_t_1, __pyx_v_stream, 1, __pyx_v_chunk_size);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":452
 *         """
 *         # If the raw byte content hasn't been parsed, we can simply pass it through
 *         if not checksum_data and not self.http_parsed:             # <<<<<<<<<<<<<<
 *             return self._write_impl(self.reader, stream, True, chunk_size)
 * 
 */
  }

  /* "fastwarc/warc.pyx":456
 * 
 *         # Otherwise read everything into memory for content-length correction and checksumming
 *         cdef BytesIOStream block_buf = BytesIOStream()             # <<<<<<<<<<<<<<
 * 
 *         block_digest = hashlib.sha1() if checksum_data else None
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_BytesIOStream)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_block_buf = ((struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":458
 *         cdef BytesIOStream block_buf = BytesIOStream()
 * 
 *         block_digest = hashlib.sha1() if checksum_data else None             # <<<<<<<<<<<<<<
 *         payload_digest = hashlib.sha1() if checksum_data and self._http_parsed else None
 * 
 */
  if ((__pyx_v_checksum_data != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sha1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_v_block_digest = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":459
 * 
 *         block_digest = hashlib.sha1() if checksum_data else None
 *         payload_digest = hashlib.sha1() if checksum_data and self._http_parsed else None             # <<<<<<<<<<<<<<
 * 
 *         if self._http_parsed:
 */
  __pyx_t_12 = (__pyx_v_checksum_data != 0);
  if (__pyx_t_12) {
  } else {
    __pyx_t_10 = __pyx_t_12;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_12 = (__pyx_v_self->_http_parsed != 0);
  __pyx_t_10 = __pyx_t_12;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_10) {
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_sha1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_v_payload_digest = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":461
 *         payload_digest = hashlib.sha1() if checksum_data and self._http_parsed else None
 * 
 *         if self._http_parsed:             # <<<<<<<<<<<<<<
 *             self._http_headers.write(block_buf)
 *             block_buf.write(<char*>b'\r\n', 2)
 */
  __pyx_t_10 = (__pyx_v_self->_http_parsed != 0);
  if (__pyx_t_10) {

    /* "fastwarc/warc.pyx":462
 * 
 *         if self._http_parsed:
 *             self._http_headers.write(block_buf)             # <<<<<<<<<<<<<<
 *             block_buf.write(<char*>b'\r\n', 2)
 * 
 */
    (void)(((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_http_headers->__pyx_vtab)->write(__pyx_v_self->_http_headers, ((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_v_block_buf)));

    /* "fastwarc/warc.pyx":463
 *         if self._http_parsed:
 *             self._http_headers.write(block_buf)
 *             block_buf.write(<char*>b'\r\n', 2)             # <<<<<<<<<<<<<<
 * 
 *             if checksum_data:
 */
    (void)(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_block_buf->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_v_block_buf), ((char *)((char *)"\r\n")), 2));

    /* "fastwarc/warc.pyx":465
 *             block_buf.write(<char*>b'\r\n', 2)
 * 
 *             if checksum_data:             # <<<<<<<<<<<<<<
 *                 block_digest.update(block_buf.getvalue())
 * 
 */
    __pyx_t_10 = (__pyx_v_checksum_data != 0);
    if (__pyx_t_10) {

      /* "fastwarc/warc.pyx":466
 * 
 *             if checksum_data:
 *                 block_digest.update(block_buf.getvalue())             # <<<<<<<<<<<<<<
 * 
 *         cdef string payload_data
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_block_digest, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __pyx_convert_PyBytes_string_to_py_std__in_string(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_block_buf->__pyx_base.__pyx_vtab)->getvalue(__pyx_v_block_buf)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fastwarc/warc.pyx":465
 *             block_buf.write(<char*>b'\r\n', 2)
 * 
 *             if checksum_data:             # <<<<<<<<<<<<<<
 *                 block_digest.update(block_buf.getvalue())
 * 
 */
    }

    /* "fastwarc/warc.pyx":461
 *         payload_digest = hashlib.sha1() if checksum_data and self._http_parsed else None
 * 
 *         if self._http_parsed:             # <<<<<<<<<<<<<<
 *             self._http_headers.write(block_buf)
 *             block_buf.write(<char*>b'\r\n', 2)
 */
  }

  /* "fastwarc/warc.pyx":469
 * 
 *         cdef string payload_data
 *         while True:             # <<<<<<<<<<<<<<
 *             payload_data = self.reader.read(chunk_size)
 *             if payload_data.empty():
 */
  while (1) {

    /* "fastwarc/warc.pyx":470
 *         cdef string payload_data
 *         while True:
 *             payload_data = self.reader.read(chunk_size)             # <<<<<<<<<<<<<<
 *             if payload_data.empty():
 *                 break
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reader); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_chunk_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_13 = __pyx_convert_string_from_py_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_payload_data = __pyx_t_13;

    /* "fastwarc/warc.pyx":471
 *         while True:
 *             payload_data = self.reader.read(chunk_size)
 *             if payload_data.empty():             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_10 = (__pyx_v_payload_data.empty() != 0);
    if (__pyx_t_10) {

      /* "fastwarc/warc.pyx":472
 *             payload_data = self.reader.read(chunk_size)
 *             if payload_data.empty():
 *                 break             # <<<<<<<<<<<<<<
 * 
 *             if checksum_data:
 */
      goto __pyx_L11_break;

      /* "fastwarc/warc.pyx":471
 *         while True:
 *             payload_data = self.reader.read(chunk_size)
 *             if payload_data.empty():             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    }

    /* "fastwarc/warc.pyx":474
 *                 break
 * 
 *             if checksum_data:             # <<<<<<<<<<<<<<
 *                 block_digest.update(payload_data.data()[:payload_data.size()])
 *                 if payload_digest is not None:
 */
    __pyx_t_10 = (__pyx_v_checksum_data != 0);
    if (__pyx_t_10) {

      /* "fastwarc/warc.pyx":475
 * 
 *             if checksum_data:
 *                 block_digest.update(payload_data.data()[:payload_data.size()])             # <<<<<<<<<<<<<<
 *                 if payload_digest is not None:
 *                     payload_digest.update(payload_data.data()[:payload_data.size()])
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_block_digest, __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_payload_data.data() + 0, __pyx_v_payload_data.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fastwarc/warc.pyx":476
 *             if checksum_data:
 *                 block_digest.update(payload_data.data()[:payload_data.size()])
 *                 if payload_digest is not None:             # <<<<<<<<<<<<<<
 *                     payload_digest.update(payload_data.data()[:payload_data.size()])
 *             block_buf.write(payload_data.data(), payload_data.size())
 */
      __pyx_t_10 = (__pyx_v_payload_digest != Py_None);
      __pyx_t_12 = (__pyx_t_10 != 0);
      if (__pyx_t_12) {

        /* "fastwarc/warc.pyx":477
 *                 block_digest.update(payload_data.data()[:payload_data.size()])
 *                 if payload_digest is not None:
 *                     payload_digest.update(payload_data.data()[:payload_data.size()])             # <<<<<<<<<<<<<<
 *             block_buf.write(payload_data.data(), payload_data.size())
 * 
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_payload_digest, __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 477, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_payload_data.data() + 0, __pyx_v_payload_data.size() - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 477, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "fastwarc/warc.pyx":476
 *             if checksum_data:
 *                 block_digest.update(payload_data.data()[:payload_data.size()])
 *                 if payload_digest is not None:             # <<<<<<<<<<<<<<
 *                     payload_digest.update(payload_data.data()[:payload_data.size()])
 *             block_buf.write(payload_data.data(), payload_data.size())
 */
      }

      /* "fastwarc/warc.pyx":474
 *                 break
 * 
 *             if checksum_data:             # <<<<<<<<<<<<<<
 *                 block_digest.update(payload_data.data()[:payload_data.size()])
 *                 if payload_digest is not None:
 */
    }

    /* "fastwarc/warc.pyx":478
 *                 if payload_digest is not None:
 *                     payload_digest.update(payload_data.data()[:payload_data.size()])
 *             block_buf.write(payload_data.data(), payload_data.size())             # <<<<<<<<<<<<<<
 * 
 *         self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))
 */
    (void)(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_block_buf->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_v_block_buf), __pyx_v_payload_data.data(), __pyx_v_payload_data.size()));
  }
  __pyx_L11_break:;

  /* "fastwarc/warc.pyx":480
 *             block_buf.write(payload_data.data(), payload_data.size())
 * 
 *         self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))             # <<<<<<<<<<<<<<
 *         if checksum_data:
 *             if payload_digest is not None:
 */
  ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_headers->__pyx_vtab)->set_header(__pyx_v_self->_headers, ((char *)((char *)"Content-Length")), std::to_string(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_block_buf->__pyx_base.__pyx_vtab)->__pyx_base.tell(((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_v_block_buf))));

  /* "fastwarc/warc.pyx":481
 * 
 *         self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))
 *         if checksum_data:             # <<<<<<<<<<<<<<
 *             if payload_digest is not None:
 *                 self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))
 */
  __pyx_t_12 = (__pyx_v_checksum_data != 0);
  if (__pyx_t_12) {

    /* "fastwarc/warc.pyx":482
 *         self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))
 *         if checksum_data:
 *             if payload_digest is not None:             # <<<<<<<<<<<<<<
 *                 self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))
 *             self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_digest.digest()))
 */
    __pyx_t_12 = (__pyx_v_payload_digest != Py_None);
    __pyx_t_10 = (__pyx_t_12 != 0);
    if (__pyx_t_10) {

      /* "fastwarc/warc.pyx":483
 *         if checksum_data:
 *             if payload_digest is not None:
 *                 self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))             # <<<<<<<<<<<<<<
 *             self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_digest.digest()))
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_base64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_b32encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_payload_digest, __pyx_n_s_digest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_kp_b_sha1_2, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = __pyx_convert_string_from_py_std__in_string(__pyx_t_2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_headers->__pyx_vtab)->set_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Payload-Digest")), __pyx_t_13);

      /* "fastwarc/warc.pyx":482
 *         self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))
 *         if checksum_data:
 *             if payload_digest is not None:             # <<<<<<<<<<<<<<
 *                 self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))
 *             self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_digest.digest()))
 */
    }

    /* "fastwarc/warc.pyx":484
 *             if payload_digest is not None:
 *                 self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))
 *             self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_digest.digest()))             # <<<<<<<<<<<<<<
 * 
 *         block_buf.seek(0)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_base64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_b32encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_block_digest, __pyx_n_s_digest); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_kp_b_sha1_2, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = __pyx_convert_string_from_py_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_headers->__pyx_vtab)->set_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Block-Digest")), __pyx_t_13);

    /* "fastwarc/warc.pyx":481
 * 
 *         self._headers.set_header(<char*>b'Content-Length', to_string(block_buf.tell()))
 *         if checksum_data:             # <<<<<<<<<<<<<<
 *             if payload_digest is not None:
 *                 self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest.digest()))
 */
  }

  /* "fastwarc/warc.pyx":486
 *             self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_digest.digest()))
 * 
 *         block_buf.seek(0)             # <<<<<<<<<<<<<<
 *         return self._write_impl(block_buf, stream, False, chunk_size)
 * 
 */
  ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_block_buf->__pyx_base.__pyx_vtab)->seek(__pyx_v_block_buf, 0);

  /* "fastwarc/warc.pyx":487
 * 
 *         block_buf.seek(0)
 *         return self._write_impl(block_buf, stream, False, chunk_size)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _write_impl(self, in_reader, out_stream, bint write_payload_headers, size_t chunk_size):
 */
  __pyx_r = ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *)__pyx_v_self->__pyx_vtab)->_write_impl(__pyx_v_self, ((PyObject *)__pyx_v_block_buf), __pyx_v_stream, 0, __pyx_v_chunk_size);
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":440
 * 
 *     # noinspection PyTypeChecker
 *     cpdef size_t write(self, stream, bint checksum_data=False, size_t chunk_size=16384):             # <<<<<<<<<<<<<<
 *         """
 *         Write WARC record onto a stream.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord.write", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_block_buf);
  __Pyx_XDECREF(__pyx_v_block_digest);
  __Pyx_XDECREF(__pyx_v_payload_digest);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_9write(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10WarcRecord_8write[] = "\n        Write WARC record onto a stream.\n        \n        :param stream: output stream\n        :param checksum_data: add block and payload digest headers\n        :type checksum_data: bool, optional, default: False\n        :param chunk_size: write block size\n        :type chunk_size: int, optional, default: 16384\n        :return: number of bytes written\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_9write(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_stream = 0;
  int __pyx_v_checksum_data;
  size_t __pyx_v_chunk_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_stream,&__pyx_n_s_checksum_data,&__pyx_n_s_chunk_size,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_checksum_data);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chunk_size);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "write") < 0)) __PYX_ERR(0, 440, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_stream = values[0];
    if (values[1]) {
      __pyx_v_checksum_data = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_checksum_data == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L3_error)
    } else {
      __pyx_v_checksum_data = ((int)0);
    }
    if (values[2]) {
      __pyx_v_chunk_size = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_chunk_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L3_error)
    } else {
      __pyx_v_chunk_size = ((size_t)0x4000);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 440, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_8write(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), __pyx_v_stream, __pyx_v_checksum_data, __pyx_v_chunk_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_8write(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_stream, int __pyx_v_checksum_data, size_t __pyx_v_chunk_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.checksum_data = __pyx_v_checksum_data;
  __pyx_t_2.chunk_size = __pyx_v_chunk_size;
  __pyx_t_1 = __pyx_vtabptr_8fastwarc_4warc_WarcRecord->write(__pyx_v_self, __pyx_v_stream, 1, &__pyx_t_2); 
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":489
 *         return self._write_impl(block_buf, stream, False, chunk_size)
 * 
 *     cdef size_t _write_impl(self, in_reader, out_stream, bint write_payload_headers, size_t chunk_size):             # <<<<<<<<<<<<<<
 *         cdef IOStream out_stream_wrapped
 *         cdef size_t bytes_written = 0
 */

static size_t __pyx_f_8fastwarc_4warc_10WarcRecord__write_impl(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, PyObject *__pyx_v_in_reader, PyObject *__pyx_v_out_stream, int __pyx_v_write_payload_headers, size_t __pyx_v_chunk_size) {
  struct __pyx_obj_8fastwarc_9stream_io_IOStream *__pyx_v_out_stream_wrapped = 0;
  size_t __pyx_v_bytes_written;
  int __pyx_v_compress_member_started;
  std::string __pyx_v_data;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  std::string __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_write_impl", 0);

  /* "fastwarc/warc.pyx":491
 *     cdef size_t _write_impl(self, in_reader, out_stream, bint write_payload_headers, size_t chunk_size):
 *         cdef IOStream out_stream_wrapped
 *         cdef size_t bytes_written = 0             # <<<<<<<<<<<<<<
 *         cdef bint compress_member_started = False
 * 
 */
  __pyx_v_bytes_written = 0;

  /* "fastwarc/warc.pyx":492
 *         cdef IOStream out_stream_wrapped
 *         cdef size_t bytes_written = 0
 *         cdef bint compress_member_started = False             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(out_stream, IOStream):
 */
  __pyx_v_compress_member_started = 0;

  /* "fastwarc/warc.pyx":494
 *         cdef bint compress_member_started = False
 * 
 *         if isinstance(out_stream, IOStream):             # <<<<<<<<<<<<<<
 *             out_stream_wrapped = <IOStream>out_stream
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_out_stream, __pyx_ptype_8fastwarc_9stream_io_IOStream); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":495
 * 
 *         if isinstance(out_stream, IOStream):
 *             out_stream_wrapped = <IOStream>out_stream             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(out_stream, CompressingStream):
 */
    __pyx_t_3 = __pyx_v_out_stream;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_out_stream_wrapped = ((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":497
 *             out_stream_wrapped = <IOStream>out_stream
 * 
 *             if isinstance(out_stream, CompressingStream):             # <<<<<<<<<<<<<<
 *                 bytes_written = (<CompressingStream>out_stream_wrapped).begin_member()
 *                 compress_member_started = True
 */
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_out_stream, __pyx_ptype_8fastwarc_9stream_io_CompressingStream); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":498
 * 
 *             if isinstance(out_stream, CompressingStream):
 *                 bytes_written = (<CompressingStream>out_stream_wrapped).begin_member()             # <<<<<<<<<<<<<<
 *                 compress_member_started = True
 * 
 */
      __pyx_v_bytes_written = ((struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream *)((struct __pyx_obj_8fastwarc_9stream_io_CompressingStream *)__pyx_v_out_stream_wrapped)->__pyx_base.__pyx_vtab)->begin_member(((struct __pyx_obj_8fastwarc_9stream_io_CompressingStream *)__pyx_v_out_stream_wrapped));

      /* "fastwarc/warc.pyx":499
 *             if isinstance(out_stream, CompressingStream):
 *                 bytes_written = (<CompressingStream>out_stream_wrapped).begin_member()
 *                 compress_member_started = True             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(out_stream, object) and hasattr(out_stream, 'write'):
 */
      __pyx_v_compress_member_started = 1;

      /* "fastwarc/warc.pyx":497
 *             out_stream_wrapped = <IOStream>out_stream
 * 
 *             if isinstance(out_stream, CompressingStream):             # <<<<<<<<<<<<<<
 *                 bytes_written = (<CompressingStream>out_stream_wrapped).begin_member()
 *                 compress_member_started = True
 */
    }

    /* "fastwarc/warc.pyx":494
 *         cdef bint compress_member_started = False
 * 
 *         if isinstance(out_stream, IOStream):             # <<<<<<<<<<<<<<
 *             out_stream_wrapped = <IOStream>out_stream
 * 
 */
    goto __pyx_L3;
  }

  /* "fastwarc/warc.pyx":501
 *                 compress_member_started = True
 * 
 *         elif isinstance(out_stream, object) and hasattr(out_stream, 'write'):             # <<<<<<<<<<<<<<
 *             out_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, out_stream)
 *         else:
 */
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_out_stream, __pyx_builtin_object); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(__pyx_v_out_stream, __pyx_n_u_write); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_4 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":502
 * 
 *         elif isinstance(out_stream, object) and hasattr(out_stream, 'write'):
 *             out_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, out_stream)             # <<<<<<<<<<<<<<
 *         else:
 *             warnings.warn(f"Object of type '{type(out_stream).__name__}' is not a valid stream.", RuntimeWarning)
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_out_stream);
    __Pyx_GIVEREF(__pyx_v_out_stream);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_out_stream);
    __pyx_t_5 = __Pyx_tp_new(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter), ((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_8fastwarc_9stream_io_IOStream)))) __PYX_ERR(0, 502, __pyx_L1_error)
    __pyx_v_out_stream_wrapped = ((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "fastwarc/warc.pyx":501
 *                 compress_member_started = True
 * 
 *         elif isinstance(out_stream, object) and hasattr(out_stream, 'write'):             # <<<<<<<<<<<<<<
 *             out_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, out_stream)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "fastwarc/warc.pyx":504
 *             out_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, out_stream)
 *         else:
 *             warnings.warn(f"Object of type '{type(out_stream).__name__}' is not a valid stream.", RuntimeWarning)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_Object_of_type);
    __pyx_t_7 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_Object_of_type);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Object_of_type);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_out_stream)), __pyx_n_s_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_10);
    __pyx_t_10 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_stream);
    __pyx_t_7 += 24;
    __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_stream);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_is_not_a_valid_stream);
    __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_11 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_11 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_10, __pyx_builtin_RuntimeWarning};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_10, __pyx_builtin_RuntimeWarning};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_11, __pyx_t_10);
      __Pyx_INCREF(__pyx_builtin_RuntimeWarning);
      __Pyx_GIVEREF(__pyx_builtin_RuntimeWarning);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_11, __pyx_builtin_RuntimeWarning);
      __pyx_t_10 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "fastwarc/warc.pyx":505
 *         else:
 *             warnings.warn(f"Object of type '{type(out_stream).__name__}' is not a valid stream.", RuntimeWarning)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         bytes_written += self._headers.write(out_stream_wrapped)
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "fastwarc/warc.pyx":507
 *             return 0
 * 
 *         bytes_written += self._headers.write(out_stream_wrapped)             # <<<<<<<<<<<<<<
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 */
  __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_headers->__pyx_vtab)->write(__pyx_v_self->_headers, __pyx_v_out_stream_wrapped));

  /* "fastwarc/warc.pyx":508
 * 
 *         bytes_written += self._headers.write(out_stream_wrapped)
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)             # <<<<<<<<<<<<<<
 * 
 *         if write_payload_headers and self._http_parsed:
 */
  __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_out_stream_wrapped->__pyx_vtab)->write(__pyx_v_out_stream_wrapped, ((char const *)"\r\n"), 2));

  /* "fastwarc/warc.pyx":510
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 *         if write_payload_headers and self._http_parsed:             # <<<<<<<<<<<<<<
 *             bytes_written += self._http_headers.write(out_stream_wrapped)
 *             bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 */
  __pyx_t_2 = (__pyx_v_write_payload_headers != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->_http_parsed != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":511
 * 
 *         if write_payload_headers and self._http_parsed:
 *             bytes_written += self._http_headers.write(out_stream_wrapped)             # <<<<<<<<<<<<<<
 *             bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_http_headers->__pyx_vtab)->write(__pyx_v_self->_http_headers, __pyx_v_out_stream_wrapped));

    /* "fastwarc/warc.pyx":512
 *         if write_payload_headers and self._http_parsed:
 *             bytes_written += self._http_headers.write(out_stream_wrapped)
 *             bytes_written += out_stream_wrapped.write(b'\r\n', 2)             # <<<<<<<<<<<<<<
 * 
 *         cdef string data
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_out_stream_wrapped->__pyx_vtab)->write(__pyx_v_out_stream_wrapped, ((char const *)"\r\n"), 2));

    /* "fastwarc/warc.pyx":510
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 *         if write_payload_headers and self._http_parsed:             # <<<<<<<<<<<<<<
 *             bytes_written += self._http_headers.write(out_stream_wrapped)
 *             bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 */
  }

  /* "fastwarc/warc.pyx":515
 * 
 *         cdef string data
 *         while True:             # <<<<<<<<<<<<<<
 *             data = in_reader.read(chunk_size)
 *             if data.empty():
 */
  while (1) {

    /* "fastwarc/warc.pyx":516
 *         cdef string data
 *         while True:
 *             data = in_reader.read(chunk_size)             # <<<<<<<<<<<<<<
 *             if data.empty():
 *                 break
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_in_reader, __pyx_n_s_read); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_PyInt_FromSize_t(__pyx_v_chunk_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_12 = __pyx_convert_string_from_py_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_data = __pyx_t_12;

    /* "fastwarc/warc.pyx":517
 *         while True:
 *             data = in_reader.read(chunk_size)
 *             if data.empty():             # <<<<<<<<<<<<<<
 *                 break
 *             bytes_written += out_stream_wrapped.write(data.data(), data.size())
 */
    __pyx_t_1 = (__pyx_v_data.empty() != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":518
 *             data = in_reader.read(chunk_size)
 *             if data.empty():
 *                 break             # <<<<<<<<<<<<<<
 *             bytes_written += out_stream_wrapped.write(data.data(), data.size())
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 */
      goto __pyx_L11_break;

      /* "fastwarc/warc.pyx":517
 *         while True:
 *             data = in_reader.read(chunk_size)
 *             if data.empty():             # <<<<<<<<<<<<<<
 *                 break
 *             bytes_written += out_stream_wrapped.write(data.data(), data.size())
 */
    }

    /* "fastwarc/warc.pyx":519
 *             if data.empty():
 *                 break
 *             bytes_written += out_stream_wrapped.write(data.data(), data.size())             # <<<<<<<<<<<<<<
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_out_stream_wrapped->__pyx_vtab)->write(__pyx_v_out_stream_wrapped, __pyx_v_data.data(), __pyx_v_data.size()));
  }
  __pyx_L11_break:;

  /* "fastwarc/warc.pyx":520
 *                 break
 *             bytes_written += out_stream_wrapped.write(data.data(), data.size())
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)             # <<<<<<<<<<<<<<
 * 
 *         if compress_member_started:
 */
  __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream *)__pyx_v_out_stream_wrapped->__pyx_vtab)->write(__pyx_v_out_stream_wrapped, ((char const *)"\r\n"), 2));

  /* "fastwarc/warc.pyx":522
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 *         if compress_member_started:             # <<<<<<<<<<<<<<
 *             bytes_written += (<CompressingStream> out_stream_wrapped).end_member()
 * 
 */
  __pyx_t_1 = (__pyx_v_compress_member_started != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":523
 * 
 *         if compress_member_started:
 *             bytes_written += (<CompressingStream> out_stream_wrapped).end_member()             # <<<<<<<<<<<<<<
 * 
 *         return bytes_written
 */
    __pyx_v_bytes_written = (__pyx_v_bytes_written + ((struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream *)((struct __pyx_obj_8fastwarc_9stream_io_CompressingStream *)__pyx_v_out_stream_wrapped)->__pyx_base.__pyx_vtab)->end_member(((struct __pyx_obj_8fastwarc_9stream_io_CompressingStream *)__pyx_v_out_stream_wrapped)));

    /* "fastwarc/warc.pyx":522
 *         bytes_written += out_stream_wrapped.write(b'\r\n', 2)
 * 
 *         if compress_member_started:             # <<<<<<<<<<<<<<
 *             bytes_written += (<CompressingStream> out_stream_wrapped).end_member()
 * 
 */
  }

  /* "fastwarc/warc.pyx":525
 *             bytes_written += (<CompressingStream> out_stream_wrapped).end_member()
 * 
 *         return bytes_written             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume):
 */
  __pyx_r = __pyx_v_bytes_written;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":489
 *         return self._write_impl(block_buf, stream, False, chunk_size)
 * 
 *     cdef size_t _write_impl(self, in_reader, out_stream, bint write_payload_headers, size_t chunk_size):             # <<<<<<<<<<<<<<
 *         cdef IOStream out_stream_wrapped
 *         cdef size_t bytes_written = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord._write_impl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out_stream_wrapped);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":527
 *         return bytes_written
 * 
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume):             # <<<<<<<<<<<<<<
 *         cdef size_t sep_pos = base32_digest.find(b':')
 *         if sep_pos == strnpos:
 */

static int __pyx_f_8fastwarc_4warc_10WarcRecord__verify_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, std::string const &__pyx_v_base32_digest, int __pyx_v_consume) {
  size_t __pyx_v_sep_pos;
  std::string __pyx_v_alg;
  PyObject *__pyx_v_digest = 0;
  PyObject *__pyx_v_h = NULL;
  std::string __pyx_v_block;
  struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *__pyx_v_tee_stream = 0;
  int __pyx_v_consume_override;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  std::string __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_t_9;
  struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_read __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_verify_digest", 0);

  /* "fastwarc/warc.pyx":528
 * 
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume):
 *         cdef size_t sep_pos = base32_digest.find(b':')             # <<<<<<<<<<<<<<
 *         if sep_pos == strnpos:
 *             return False
 */
  __pyx_v_sep_pos = __pyx_v_base32_digest.find(((char const *)":"));

  /* "fastwarc/warc.pyx":529
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume):
 *         cdef size_t sep_pos = base32_digest.find(b':')
 *         if sep_pos == strnpos:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_1 = ((__pyx_v_sep_pos == std::string::npos) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":530
 *         cdef size_t sep_pos = base32_digest.find(b':')
 *         if sep_pos == strnpos:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         cdef string alg = base32_digest.substr(0, sep_pos)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":529
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume):
 *         cdef size_t sep_pos = base32_digest.find(b':')
 *         if sep_pos == strnpos:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "fastwarc/warc.pyx":532
 *             return False
 * 
 *         cdef string alg = base32_digest.substr(0, sep_pos)             # <<<<<<<<<<<<<<
 *         cdef bytes digest = base64.b32decode(base32_digest.substr(sep_pos + 1))
 * 
 */
  try {
    __pyx_t_2 = __pyx_v_base32_digest.substr(0, __pyx_v_sep_pos);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 532, __pyx_L1_error)
  }
  __pyx_v_alg = __pyx_t_2;

  /* "fastwarc/warc.pyx":533
 * 
 *         cdef string alg = base32_digest.substr(0, sep_pos)
 *         cdef bytes digest = base64.b32decode(base32_digest.substr(sep_pos + 1))             # <<<<<<<<<<<<<<
 * 
 *         if alg == b'sha1':
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_base64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_b32decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  try {
    __pyx_t_2 = __pyx_v_base32_digest.substr((__pyx_v_sep_pos + 1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 533, __pyx_L1_error)
  }
  __pyx_t_4 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 533, __pyx_L1_error)
  __pyx_v_digest = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastwarc/warc.pyx":535
 *         cdef bytes digest = base64.b32decode(base32_digest.substr(sep_pos + 1))
 * 
 *         if alg == b'sha1':             # <<<<<<<<<<<<<<
 *             h = hashlib.sha1()
 *         elif alg == b'md5':
 */
  __pyx_t_1 = ((__pyx_v_alg == ((char const *)"sha1")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":536
 * 
 *         if alg == b'sha1':
 *             h = hashlib.sha1()             # <<<<<<<<<<<<<<
 *         elif alg == b'md5':
 *             h = hashlib.md5()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sha1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_h = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":535
 *         cdef bytes digest = base64.b32decode(base32_digest.substr(sep_pos + 1))
 * 
 *         if alg == b'sha1':             # <<<<<<<<<<<<<<
 *             h = hashlib.sha1()
 *         elif alg == b'md5':
 */
    goto __pyx_L4;
  }

  /* "fastwarc/warc.pyx":537
 *         if alg == b'sha1':
 *             h = hashlib.sha1()
 *         elif alg == b'md5':             # <<<<<<<<<<<<<<
 *             h = hashlib.md5()
 *         elif alg == b'sha256':
 */
  __pyx_t_1 = ((__pyx_v_alg == ((char const *)"md5")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":538
 *             h = hashlib.sha1()
 *         elif alg == b'md5':
 *             h = hashlib.md5()             # <<<<<<<<<<<<<<
 *         elif alg == b'sha256':
 *             h = hashlib.sha256()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_md5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_h = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":537
 *         if alg == b'sha1':
 *             h = hashlib.sha1()
 *         elif alg == b'md5':             # <<<<<<<<<<<<<<
 *             h = hashlib.md5()
 *         elif alg == b'sha256':
 */
    goto __pyx_L4;
  }

  /* "fastwarc/warc.pyx":539
 *         elif alg == b'md5':
 *             h = hashlib.md5()
 *         elif alg == b'sha256':             # <<<<<<<<<<<<<<
 *             h = hashlib.sha256()
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_alg == ((char const *)"sha256")) != 0);
  if (__pyx_t_1) {

    /* "fastwarc/warc.pyx":540
 *             h = hashlib.md5()
 *         elif alg == b'sha256':
 *             h = hashlib.sha256()             # <<<<<<<<<<<<<<
 *         else:
 *             warnings.warn(f'Unsupported hash algorithm "{alg.decode()}".')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_sha256); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_h = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":539
 *         elif alg == b'md5':
 *             h = hashlib.md5()
 *         elif alg == b'sha256':             # <<<<<<<<<<<<<<
 *             h = hashlib.sha256()
 *         else:
 */
    goto __pyx_L4;
  }

  /* "fastwarc/warc.pyx":542
 *             h = hashlib.sha256()
 *         else:
 *             warnings.warn(f'Unsupported hash algorithm "{alg.decode()}".')             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_warnings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_Unsupported_hash_algorithm);
    __pyx_t_7 += 28;
    __Pyx_GIVEREF(__pyx_kp_u_Unsupported_hash_algorithm);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Unsupported_hash_algorithm);
    __pyx_t_6 = __Pyx_decode_cpp_string(__pyx_v_alg, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__10);
    __pyx_t_7 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__10);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u__10);
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":543
 *         else:
 *             warnings.warn(f'Unsupported hash algorithm "{alg.decode()}".')
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         cdef string block
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L4:;

  /* "fastwarc/warc.pyx":547
 *         cdef string block
 *         cdef BytesIOStream tee_stream
 *         cdef bint consume_override = consume             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(self._reader.stream, BytesIOStream):
 */
  __pyx_v_consume_override = __pyx_v_consume;

  /* "fastwarc/warc.pyx":549
 *         cdef bint consume_override = consume
 * 
 *         if isinstance(self._reader.stream, BytesIOStream):             # <<<<<<<<<<<<<<
 *             # Stream is already a BytesIOStream, so we don't need to create another copy
 *             consume_override = True
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_reader->stream);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_t_3, __pyx_ptype_8fastwarc_9stream_io_BytesIOStream); 
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = (__pyx_t_1 != 0);
  if (__pyx_t_9) {

    /* "fastwarc/warc.pyx":551
 *         if isinstance(self._reader.stream, BytesIOStream):
 *             # Stream is already a BytesIOStream, so we don't need to create another copy
 *             consume_override = True             # <<<<<<<<<<<<<<
 *             tee_stream = <BytesIOStream>self._reader.stream
 *         elif not consume:
 */
    __pyx_v_consume_override = 1;

    /* "fastwarc/warc.pyx":552
 *             # Stream is already a BytesIOStream, so we don't need to create another copy
 *             consume_override = True
 *             tee_stream = <BytesIOStream>self._reader.stream             # <<<<<<<<<<<<<<
 *         elif not consume:
 *             tee_stream = BytesIOStream()
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_self->_reader->stream);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_tee_stream = ((struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":549
 *         cdef bint consume_override = consume
 * 
 *         if isinstance(self._reader.stream, BytesIOStream):             # <<<<<<<<<<<<<<
 *             # Stream is already a BytesIOStream, so we don't need to create another copy
 *             consume_override = True
 */
    goto __pyx_L5;
  }

  /* "fastwarc/warc.pyx":553
 *             consume_override = True
 *             tee_stream = <BytesIOStream>self._reader.stream
 *         elif not consume:             # <<<<<<<<<<<<<<
 *             tee_stream = BytesIOStream()
 * 
 */
  __pyx_t_9 = ((!(__pyx_v_consume != 0)) != 0);
  if (__pyx_t_9) {

    /* "fastwarc/warc.pyx":554
 *             tee_stream = <BytesIOStream>self._reader.stream
 *         elif not consume:
 *             tee_stream = BytesIOStream()             # <<<<<<<<<<<<<<
 * 
 *         while True:
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_BytesIOStream)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_tee_stream = ((struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":553
 *             consume_override = True
 *             tee_stream = <BytesIOStream>self._reader.stream
 *         elif not consume:             # <<<<<<<<<<<<<<
 *             tee_stream = BytesIOStream()
 * 
 */
  }
  __pyx_L5:;

  /* "fastwarc/warc.pyx":556
 *             tee_stream = BytesIOStream()
 * 
 *         while True:             # <<<<<<<<<<<<<<
 *             block = self._reader.read(1024)
 *             if block.empty():
 */
  while (1) {

    /* "fastwarc/warc.pyx":557
 * 
 *         while True:
 *             block = self._reader.read(1024)             # <<<<<<<<<<<<<<
 *             if block.empty():
 *                 break
 */
    __pyx_t_10.__pyx_n = 1;
    __pyx_t_10.size = 0x400;
    __pyx_t_2 = ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->_reader->__pyx_vtab)->read(__pyx_v_self->_reader, 0, &__pyx_t_10); 
    __pyx_v_block = __pyx_t_2;

    /* "fastwarc/warc.pyx":558
 *         while True:
 *             block = self._reader.read(1024)
 *             if block.empty():             # <<<<<<<<<<<<<<
 *                 break
 *             h.update(block)
 */
    __pyx_t_9 = (__pyx_v_block.empty() != 0);
    if (__pyx_t_9) {

      /* "fastwarc/warc.pyx":559
 *             block = self._reader.read(1024)
 *             if block.empty():
 *                 break             # <<<<<<<<<<<<<<
 *             h.update(block)
 *             if not consume_override:
 */
      goto __pyx_L7_break;

      /* "fastwarc/warc.pyx":558
 *         while True:
 *             block = self._reader.read(1024)
 *             if block.empty():             # <<<<<<<<<<<<<<
 *                 break
 *             h.update(block)
 */
    }

    /* "fastwarc/warc.pyx":560
 *             if block.empty():
 *                 break
 *             h.update(block)             # <<<<<<<<<<<<<<
 *             if not consume_override:
 *                 tee_stream.write(block.data(), block.size())
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_h, __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_v_block); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fastwarc/warc.pyx":561
 *                 break
 *             h.update(block)
 *             if not consume_override:             # <<<<<<<<<<<<<<
 *                 tee_stream.write(block.data(), block.size())
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_consume_override != 0)) != 0);
    if (__pyx_t_9) {

      /* "fastwarc/warc.pyx":562
 *             h.update(block)
 *             if not consume_override:
 *                 tee_stream.write(block.data(), block.size())             # <<<<<<<<<<<<<<
 * 
 *         if not consume:
 */
      if (unlikely(!__pyx_v_tee_stream)) { __Pyx_RaiseUnboundLocalError("tee_stream"); __PYX_ERR(0, 562, __pyx_L1_error) }
      (void)(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_tee_stream->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_v_tee_stream), __pyx_v_block.data(), __pyx_v_block.size()));

      /* "fastwarc/warc.pyx":561
 *                 break
 *             h.update(block)
 *             if not consume_override:             # <<<<<<<<<<<<<<
 *                 tee_stream.write(block.data(), block.size())
 * 
 */
    }
  }
  __pyx_L7_break:;

  /* "fastwarc/warc.pyx":564
 *                 tee_stream.write(block.data(), block.size())
 * 
 *         if not consume:             # <<<<<<<<<<<<<<
 *             tee_stream.seek(0)
 *             self._reader = BufferedReader.__new__(BufferedReader, tee_stream)
 */
  __pyx_t_9 = ((!(__pyx_v_consume != 0)) != 0);
  if (__pyx_t_9) {

    /* "fastwarc/warc.pyx":565
 * 
 *         if not consume:
 *             tee_stream.seek(0)             # <<<<<<<<<<<<<<
 *             self._reader = BufferedReader.__new__(BufferedReader, tee_stream)
 * 
 */
    if (unlikely(!__pyx_v_tee_stream)) { __Pyx_RaiseUnboundLocalError("tee_stream"); __PYX_ERR(0, 565, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream *)__pyx_v_tee_stream->__pyx_base.__pyx_vtab)->seek(__pyx_v_tee_stream, 0);

    /* "fastwarc/warc.pyx":566
 *         if not consume:
 *             tee_stream.seek(0)
 *             self._reader = BufferedReader.__new__(BufferedReader, tee_stream)             # <<<<<<<<<<<<<<
 * 
 *         return h.digest() == digest
 */
    if (unlikely(!__pyx_v_tee_stream)) { __Pyx_RaiseUnboundLocalError("tee_stream"); __PYX_ERR(0, 566, __pyx_L1_error) }
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_v_tee_stream));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_tee_stream));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_tee_stream));
    __pyx_t_5 = __Pyx_tp_new(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_BufferedReader), ((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_8fastwarc_9stream_io_BufferedReader)))) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->_reader);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->_reader));
    __pyx_v_self->_reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "fastwarc/warc.pyx":564
 *                 tee_stream.write(block.data(), block.size())
 * 
 *         if not consume:             # <<<<<<<<<<<<<<
 *             tee_stream.seek(0)
 *             self._reader = BufferedReader.__new__(BufferedReader, tee_stream)
 */
  }

  /* "fastwarc/warc.pyx":568
 *             self._reader = BufferedReader.__new__(BufferedReader, tee_stream)
 * 
 *         return h.digest() == digest             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint verify_block_digest(self, bint consume=False):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_h, __pyx_n_s_digest); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_v_digest, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_9;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":527
 *         return bytes_written
 * 
 *     cdef bint _verify_digest(self, const string& base32_digest, bint consume):             # <<<<<<<<<<<<<<
 *         cdef size_t sep_pos = base32_digest.find(b':')
 *         if sep_pos == strnpos:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord._verify_digest", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_digest);
  __Pyx_XDECREF(__pyx_v_h);
  __Pyx_XDECREF((PyObject *)__pyx_v_tee_stream);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":570
 *         return h.digest() == digest
 * 
 *     cpdef bint verify_block_digest(self, bint consume=False):             # <<<<<<<<<<<<<<
 *         """
 *         Verify whether record block digest is valid.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_11verify_block_digest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_8fastwarc_4warc_10WarcRecord_verify_block_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest *__pyx_optional_args) {
  int __pyx_v_consume = ((int)0);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("verify_block_digest", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_consume = __pyx_optional_args->consume;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_verify_block_digest); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_10WarcRecord_11verify_block_digest)) {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":579
 *         :return: ``True`` if digest exists and is valid
 *         """
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b''), consume)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint verify_payload_digest(self, bint consume=False):
 */
  __pyx_r = ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *)__pyx_v_self->__pyx_vtab)->_verify_digest(__pyx_v_self, ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_headers->__pyx_vtab)->find_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Block-Digest")), ((char *)((char *)""))), __pyx_v_consume);
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":570
 *         return h.digest() == digest
 * 
 *     cpdef bint verify_block_digest(self, bint consume=False):             # <<<<<<<<<<<<<<
 *         """
 *         Verify whether record block digest is valid.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord.verify_block_digest", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_11verify_block_digest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10WarcRecord_10verify_block_digest[] = "\n        Verify whether record block digest is valid.\n        \n        :param consume: do not create an in-memory copy of the record stream\n                        (will fully consume the rest of the record)\n        :type consume: bool, optional, default: False\n        :return: ``True`` if digest exists and is valid\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_11verify_block_digest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_consume;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("verify_block_digest (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_consume,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_consume);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "verify_block_digest") < 0)) __PYX_ERR(0, 570, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_consume = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_consume == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    } else {
      __pyx_v_consume = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("verify_block_digest", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 570, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.verify_block_digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_10verify_block_digest(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), __pyx_v_consume);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_10verify_block_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_v_consume) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("verify_block_digest", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.consume = __pyx_v_consume;
  __pyx_t_1 = __pyx_vtabptr_8fastwarc_4warc_WarcRecord->verify_block_digest(__pyx_v_self, 1, &__pyx_t_2); 
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.verify_block_digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":581
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b''), consume)
 * 
 *     cpdef bint verify_payload_digest(self, bint consume=False):             # <<<<<<<<<<<<<<
 *         """
 *         Verify whether record payload digest is valid.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_13verify_payload_digest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_8fastwarc_4warc_10WarcRecord_verify_payload_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest *__pyx_optional_args) {
  int __pyx_v_consume = ((int)0);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("verify_payload_digest", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_consume = __pyx_optional_args->consume;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_verify_payload_digest); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_10WarcRecord_13verify_payload_digest)) {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_consume); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":590
 *         :return: ``True`` if record is HTTP record and digest exists and is valid
 *         """
 *         if not self._http_parsed:             # <<<<<<<<<<<<<<
 *             return False
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b''), consume)
 */
  __pyx_t_6 = ((!(__pyx_v_self->_http_parsed != 0)) != 0);
  if (__pyx_t_6) {

    /* "fastwarc/warc.pyx":591
 *         """
 *         if not self._http_parsed:
 *             return False             # <<<<<<<<<<<<<<
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b''), consume)
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":590
 *         :return: ``True`` if record is HTTP record and digest exists and is valid
 *         """
 *         if not self._http_parsed:             # <<<<<<<<<<<<<<
 *             return False
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b''), consume)
 */
  }

  /* "fastwarc/warc.pyx":592
 *         if not self._http_parsed:
 *             return False
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b''), consume)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *)__pyx_v_self->__pyx_vtab)->_verify_digest(__pyx_v_self, ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_self->_headers->__pyx_vtab)->find_header(__pyx_v_self->_headers, ((char *)((char *)"WARC-Payload-Digest")), ((char *)((char *)""))), __pyx_v_consume);
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":581
 *         return self._verify_digest(self._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b''), consume)
 * 
 *     cpdef bint verify_payload_digest(self, bint consume=False):             # <<<<<<<<<<<<<<
 *         """
 *         Verify whether record payload digest is valid.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("fastwarc.warc.WarcRecord.verify_payload_digest", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_13verify_payload_digest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10WarcRecord_12verify_payload_digest[] = "\n        Verify whether record payload digest is valid.\n        \n        :param consume: do not create an in-memory copy of the record stream\n                        (will fully consume the rest of the record)\n        :type consume: bool, optional, default: False\n        :return: ``True`` if record is HTTP record and digest exists and is valid\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_13verify_payload_digest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_consume;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("verify_payload_digest (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_consume,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_consume);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "verify_payload_digest") < 0)) __PYX_ERR(0, 581, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_consume = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_consume == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
    } else {
      __pyx_v_consume = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("verify_payload_digest", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 581, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.verify_payload_digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_12verify_payload_digest(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), __pyx_v_consume);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_12verify_payload_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, int __pyx_v_consume) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("verify_payload_digest", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.consume = __pyx_v_consume;
  __pyx_t_1 = __pyx_vtabptr_8fastwarc_4warc_WarcRecord->verify_payload_digest(__pyx_v_self, 1, &__pyx_t_2); 
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.verify_payload_digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_14__reduce_cython__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_10WarcRecord_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_10WarcRecord_16__setstate_cython__(((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10WarcRecord_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.WarcRecord.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":596
 * 
 * # noinspection PyProtectedMember
 * cdef size_t parse_header_block(BufferedReader reader, WarcHeaderMap target, bint has_status_line=False):             # <<<<<<<<<<<<<<
 *     """
 *     Helper function for parsing WARC or HTTP header blocks.
 */

static size_t __pyx_f_8fastwarc_4warc_parse_header_block(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *__pyx_v_reader, struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *__pyx_v_target, struct __pyx_opt_args_8fastwarc_4warc_parse_header_block *__pyx_optional_args) {
  int __pyx_v_has_status_line = ((int)0);
  std::string __pyx_v_line;
  std::string __pyx_v_header_key;
  std::string __pyx_v_header_value;
  size_t __pyx_v_delim_pos;
  size_t __pyx_v_bytes_consumed;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  std::string __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_header_block", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_has_status_line = __pyx_optional_args->has_status_line;
    }
  }

  /* "fastwarc/warc.pyx":610
 *     cdef string line
 *     cdef string header_key, header_value
 *     cdef size_t delim_pos = 0             # <<<<<<<<<<<<<<
 *     cdef size_t bytes_consumed = 0
 * 
 */
  __pyx_v_delim_pos = 0;

  /* "fastwarc/warc.pyx":611
 *     cdef string header_key, header_value
 *     cdef size_t delim_pos = 0
 *     cdef size_t bytes_consumed = 0             # <<<<<<<<<<<<<<
 * 
 *     while True:
 */
  __pyx_v_bytes_consumed = 0;

  /* "fastwarc/warc.pyx":613
 *     cdef size_t bytes_consumed = 0
 * 
 *     while True:             # <<<<<<<<<<<<<<
 *         line = reader.readline()
 *         bytes_consumed += line.size()
 */
  while (1) {

    /* "fastwarc/warc.pyx":614
 * 
 *     while True:
 *         line = reader.readline()             # <<<<<<<<<<<<<<
 *         bytes_consumed += line.size()
 * 
 */
    __pyx_v_line = ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_reader->__pyx_vtab)->readline(__pyx_v_reader, 0, NULL);

    /* "fastwarc/warc.pyx":615
 *     while True:
 *         line = reader.readline()
 *         bytes_consumed += line.size()             # <<<<<<<<<<<<<<
 * 
 *         if line == b'\r\n' or line == b'':
 */
    __pyx_v_bytes_consumed = (__pyx_v_bytes_consumed + __pyx_v_line.size());

    /* "fastwarc/warc.pyx":617
 *         bytes_consumed += line.size()
 * 
 *         if line == b'\r\n' or line == b'':             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_2 = ((__pyx_v_line == ((char const *)"\r\n")) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_line == ((char const *)"")) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":618
 * 
 *         if line == b'\r\n' or line == b'':
 *             break             # <<<<<<<<<<<<<<
 * 
 *         if isspace(line[0]):
 */
      goto __pyx_L4_break;

      /* "fastwarc/warc.pyx":617
 *         bytes_consumed += line.size()
 * 
 *         if line == b'\r\n' or line == b'':             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "fastwarc/warc.pyx":620
 *             break
 * 
 *         if isspace(line[0]):             # <<<<<<<<<<<<<<
 *             # Continuation line
 *             target.add_continuation(strip_str(line))
 */
    __pyx_t_1 = (std::isspace((__pyx_v_line[0])) != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":622
 *         if isspace(line[0]):
 *             # Continuation line
 *             target.add_continuation(strip_str(line))             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
      ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_target->__pyx_vtab)->add_continuation(__pyx_v_target, __pyx_f_8fastwarc_4warc_strip_str(__pyx_v_line));

      /* "fastwarc/warc.pyx":623
 *             # Continuation line
 *             target.add_continuation(strip_str(line))
 *             continue             # <<<<<<<<<<<<<<
 * 
 *         if has_status_line:
 */
      goto __pyx_L3_continue;

      /* "fastwarc/warc.pyx":620
 *             break
 * 
 *         if isspace(line[0]):             # <<<<<<<<<<<<<<
 *             # Continuation line
 *             target.add_continuation(strip_str(line))
 */
    }

    /* "fastwarc/warc.pyx":625
 *             continue
 * 
 *         if has_status_line:             # <<<<<<<<<<<<<<
 *             target.set_status_line(strip_str(line))
 *             has_status_line = False
 */
    __pyx_t_1 = (__pyx_v_has_status_line != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":626
 * 
 *         if has_status_line:
 *             target.set_status_line(strip_str(line))             # <<<<<<<<<<<<<<
 *             has_status_line = False
 *             continue
 */
      __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line(__pyx_v_target, __pyx_f_8fastwarc_4warc_strip_str(__pyx_v_line));

      /* "fastwarc/warc.pyx":627
 *         if has_status_line:
 *             target.set_status_line(strip_str(line))
 *             has_status_line = False             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
      __pyx_v_has_status_line = 0;

      /* "fastwarc/warc.pyx":628
 *             target.set_status_line(strip_str(line))
 *             has_status_line = False
 *             continue             # <<<<<<<<<<<<<<
 * 
 *         delim_pos = line.find(b':')
 */
      goto __pyx_L3_continue;

      /* "fastwarc/warc.pyx":625
 *             continue
 * 
 *         if has_status_line:             # <<<<<<<<<<<<<<
 *             target.set_status_line(strip_str(line))
 *             has_status_line = False
 */
    }

    /* "fastwarc/warc.pyx":630
 *             continue
 * 
 *         delim_pos = line.find(b':')             # <<<<<<<<<<<<<<
 *         if delim_pos == strnpos:
 *             # Invalid header, try to preserve it as best we can
 */
    __pyx_v_delim_pos = __pyx_v_line.find(((char const *)":"));

    /* "fastwarc/warc.pyx":631
 * 
 *         delim_pos = line.find(b':')
 *         if delim_pos == strnpos:             # <<<<<<<<<<<<<<
 *             # Invalid header, try to preserve it as best we can
 *             target.add_continuation(strip_str(line))
 */
    __pyx_t_1 = ((__pyx_v_delim_pos == std::string::npos) != 0);
    if (__pyx_t_1) {

      /* "fastwarc/warc.pyx":633
 *         if delim_pos == strnpos:
 *             # Invalid header, try to preserve it as best we can
 *             target.add_continuation(strip_str(line))             # <<<<<<<<<<<<<<
 *         else:
 *             header_key = strip_str(line.substr(0, delim_pos))
 */
      ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_target->__pyx_vtab)->add_continuation(__pyx_v_target, __pyx_f_8fastwarc_4warc_strip_str(__pyx_v_line));

      /* "fastwarc/warc.pyx":631
 * 
 *         delim_pos = line.find(b':')
 *         if delim_pos == strnpos:             # <<<<<<<<<<<<<<
 *             # Invalid header, try to preserve it as best we can
 *             target.add_continuation(strip_str(line))
 */
      goto __pyx_L10;
    }

    /* "fastwarc/warc.pyx":635
 *             target.add_continuation(strip_str(line))
 *         else:
 *             header_key = strip_str(line.substr(0, delim_pos))             # <<<<<<<<<<<<<<
 *             if delim_pos >= line.size():
 *                 header_value = string()
 */
    /*else*/ {
      try {
        __pyx_t_3 = __pyx_v_line.substr(0, __pyx_v_delim_pos);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 635, __pyx_L1_error)
      }
      __pyx_v_header_key = __pyx_f_8fastwarc_4warc_strip_str(__pyx_t_3);

      /* "fastwarc/warc.pyx":636
 *         else:
 *             header_key = strip_str(line.substr(0, delim_pos))
 *             if delim_pos >= line.size():             # <<<<<<<<<<<<<<
 *                 header_value = string()
 *             else:
 */
      __pyx_t_1 = ((__pyx_v_delim_pos >= __pyx_v_line.size()) != 0);
      if (__pyx_t_1) {

        /* "fastwarc/warc.pyx":637
 *             header_key = strip_str(line.substr(0, delim_pos))
 *             if delim_pos >= line.size():
 *                 header_value = string()             # <<<<<<<<<<<<<<
 *             else:
 *                 header_value = strip_str(line.substr(delim_pos + 1))
 */
        try {
          __pyx_t_3 = std::string();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 637, __pyx_L1_error)
        }
        __pyx_v_header_value = __pyx_t_3;

        /* "fastwarc/warc.pyx":636
 *         else:
 *             header_key = strip_str(line.substr(0, delim_pos))
 *             if delim_pos >= line.size():             # <<<<<<<<<<<<<<
 *                 header_value = string()
 *             else:
 */
        goto __pyx_L11;
      }

      /* "fastwarc/warc.pyx":639
 *                 header_value = string()
 *             else:
 *                 header_value = strip_str(line.substr(delim_pos + 1))             # <<<<<<<<<<<<<<
 *             target.append_header(header_key, header_value)
 * 
 */
      /*else*/ {
        try {
          __pyx_t_3 = __pyx_v_line.substr((__pyx_v_delim_pos + 1));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 639, __pyx_L1_error)
        }
        __pyx_v_header_value = __pyx_f_8fastwarc_4warc_strip_str(__pyx_t_3);
      }
      __pyx_L11:;

      /* "fastwarc/warc.pyx":640
 *             else:
 *                 header_value = strip_str(line.substr(delim_pos + 1))
 *             target.append_header(header_key, header_value)             # <<<<<<<<<<<<<<
 * 
 *     return bytes_consumed
 */
      __pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header(__pyx_v_target, __pyx_v_header_key, __pyx_v_header_value);
    }
    __pyx_L10:;
    __pyx_L3_continue:;
  }
  __pyx_L4_break:;

  /* "fastwarc/warc.pyx":642
 *             target.append_header(header_key, header_value)
 * 
 *     return bytes_consumed             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_bytes_consumed;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":596
 * 
 * # noinspection PyProtectedMember
 * cdef size_t parse_header_block(BufferedReader reader, WarcHeaderMap target, bint has_status_line=False):             # <<<<<<<<<<<<<<
 *     """
 *     Helper function for parsing WARC or HTTP header blocks.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fastwarc.warc.parse_header_block", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":650
 *     """WARC record stream iterator."""
 * 
 *     def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize WARC record iterator.
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_15ArchiveIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8fastwarc_4warc_15ArchiveIterator___init__[] = "\n        Initialize WARC record iterator.\n\n        :param stream: input stream (preferably an :class:`~fastwarc.stream_io.IOStream`,\n                       but any file-like Python object is fine)\n        :param parse_http: whether to parse HTTP records automatically (disable for better performance if not needed)\n        :type parse_http: bool, optional, default: True\n        :param record_types: bitmask of :class:`WarcRecordType` record types to return (others will be skipped)\n        :type record_types: int, optional, default: any_type\n        :param min_content_length: skip records with Content-Length less than this\n        :type min_content_length: int, optional\n        :param max_content_length: skip records with Content-Length large than this\n        :type max_content_length: int, optional\n        :param func_filter: Python callable taking a :class:`WarcRecord` and returning a ``bool``\n                            for further record filtering\n        :type func_filter: Callable, optional\n        :param verify_digests: skip records which have no or an invalid block digest\n        :type verify_digests: bool, optional, default: False\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_8fastwarc_4warc_15ArchiveIterator___init__;
#endif
static int __pyx_pw_8fastwarc_4warc_15ArchiveIterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 1))) return -1;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_8fastwarc_4warc_15ArchiveIterator___init__(((struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_15ArchiveIterator___init__(CYTHON_UNUSED struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":671
 *         """
 * 
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,             # <<<<<<<<<<<<<<
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):
 */

/* Python wrapper */
static int __pyx_pw_8fastwarc_4warc_15ArchiveIterator_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8fastwarc_4warc_15ArchiveIterator_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_stream = 0;
  uint16_t __pyx_v_record_types;
  int __pyx_v_parse_http;
  size_t __pyx_v_min_content_length;
  size_t __pyx_v_max_content_length;
  PyObject *__pyx_v_func_filter = 0;
  int __pyx_v_verify_digests;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_stream,&__pyx_n_s_record_types,&__pyx_n_s_parse_http,&__pyx_n_s_min_content_length,&__pyx_n_s_max_content_length,&__pyx_n_s_func_filter,&__pyx_n_s_verify_digests,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "fastwarc/warc.pyx":673
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):             # <<<<<<<<<<<<<<
 *         self._set_stream(stream)
 *         self.record = None
 */
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_record_types);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parse_http);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_content_length);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_content_length);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func_filter);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verify_digests);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 671, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_stream = values[0];
    if (values[1]) {
      __pyx_v_record_types = __Pyx_PyInt_As_uint16_t(values[1]); if (unlikely((__pyx_v_record_types == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    } else {
      __pyx_v_record_types = __pyx_k__13;
    }
    if (values[2]) {
      __pyx_v_parse_http = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_parse_http == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    } else {

      /* "fastwarc/warc.pyx":671
 *         """
 * 
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,             # <<<<<<<<<<<<<<
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):
 */
      __pyx_v_parse_http = ((int)1);
    }
    if (values[3]) {
      __pyx_v_min_content_length = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_min_content_length == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    } else {
      __pyx_v_min_content_length = __pyx_k__14;
    }
    if (values[4]) {
      __pyx_v_max_content_length = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_max_content_length == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    } else {
      __pyx_v_max_content_length = __pyx_k__15;
    }
    __pyx_v_func_filter = values[5];
    if (values[6]) {
      __pyx_v_verify_digests = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_verify_digests == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 673, __pyx_L3_error)
    } else {

      /* "fastwarc/warc.pyx":673
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):             # <<<<<<<<<<<<<<
 *         self._set_stream(stream)
 *         self.record = None
 */
      __pyx_v_verify_digests = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 671, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fastwarc.warc.ArchiveIterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8fastwarc_4warc_15ArchiveIterator_2__cinit__(((struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)__pyx_v_self), __pyx_v_stream, __pyx_v_record_types, __pyx_v_parse_http, __pyx_v_min_content_length, __pyx_v_max_content_length, __pyx_v_func_filter, __pyx_v_verify_digests);

  /* "fastwarc/warc.pyx":671
 *         """
 * 
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,             # <<<<<<<<<<<<<<
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8fastwarc_4warc_15ArchiveIterator_2__cinit__(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, PyObject *__pyx_v_stream, uint16_t __pyx_v_record_types, int __pyx_v_parse_http, size_t __pyx_v_min_content_length, size_t __pyx_v_max_content_length, PyObject *__pyx_v_func_filter, int __pyx_v_verify_digests) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "fastwarc/warc.pyx":674
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):
 *         self._set_stream(stream)             # <<<<<<<<<<<<<<
 *         self.record = None
 *         self.parse_http = parse_http
 */
  ((struct __pyx_vtabstruct_8fastwarc_4warc_ArchiveIterator *)__pyx_v_self->__pyx_vtab)->_set_stream(__pyx_v_self, __pyx_v_stream, 0);

  /* "fastwarc/warc.pyx":675
 *                   func_filter=None, bint verify_digests=False):
 *         self._set_stream(stream)
 *         self.record = None             # <<<<<<<<<<<<<<
 *         self.parse_http = parse_http
 *         self.verify_digests = verify_digests
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->record);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->record));
  __pyx_v_self->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)Py_None);

  /* "fastwarc/warc.pyx":676
 *         self._set_stream(stream)
 *         self.record = None
 *         self.parse_http = parse_http             # <<<<<<<<<<<<<<
 *         self.verify_digests = verify_digests
 *         self.min_content_length = min_content_length
 */
  __pyx_v_self->parse_http = __pyx_v_parse_http;

  /* "fastwarc/warc.pyx":677
 *         self.record = None
 *         self.parse_http = parse_http
 *         self.verify_digests = verify_digests             # <<<<<<<<<<<<<<
 *         self.min_content_length = min_content_length
 *         self.max_content_length = max_content_length
 */
  __pyx_v_self->verify_digests = __pyx_v_verify_digests;

  /* "fastwarc/warc.pyx":678
 *         self.parse_http = parse_http
 *         self.verify_digests = verify_digests
 *         self.min_content_length = min_content_length             # <<<<<<<<<<<<<<
 *         self.max_content_length = max_content_length
 *         self.func_filter = func_filter
 */
  __pyx_v_self->min_content_length = __pyx_v_min_content_length;

  /* "fastwarc/warc.pyx":679
 *         self.verify_digests = verify_digests
 *         self.min_content_length = min_content_length
 *         self.max_content_length = max_content_length             # <<<<<<<<<<<<<<
 *         self.func_filter = func_filter
 *         self.record_type_filter = record_types
 */
  __pyx_v_self->max_content_length = __pyx_v_max_content_length;

  /* "fastwarc/warc.pyx":680
 *         self.min_content_length = min_content_length
 *         self.max_content_length = max_content_length
 *         self.func_filter = func_filter             # <<<<<<<<<<<<<<
 *         self.record_type_filter = record_types
 * 
 */
  __Pyx_INCREF(__pyx_v_func_filter);
  __Pyx_GIVEREF(__pyx_v_func_filter);
  __Pyx_GOTREF(__pyx_v_self->func_filter);
  __Pyx_DECREF(__pyx_v_self->func_filter);
  __pyx_v_self->func_filter = __pyx_v_func_filter;

  /* "fastwarc/warc.pyx":681
 *         self.max_content_length = max_content_length
 *         self.func_filter = func_filter
 *         self.record_type_filter = record_types             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self) -> Iterable[WarcRecord]:
 */
  __pyx_v_self->record_type_filter = __pyx_v_record_types;

  /* "fastwarc/warc.pyx":671
 *         """
 * 
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,             # <<<<<<<<<<<<<<
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8fastwarc_4warc_15ArchiveIterator_6generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "fastwarc/warc.pyx":683
 *         self.record_type_filter = record_types
 * 
 *     def __iter__(self) -> Iterable[WarcRecord]:             # <<<<<<<<<<<<<<
 *         cdef _NextRecStatus status
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_15ArchiveIterator_5__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8fastwarc_4warc_15ArchiveIterator_5__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_15ArchiveIterator_4__iter__(((struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_15ArchiveIterator_4__iter__(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self) {
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *)__pyx_tp_new_8fastwarc_4warc___pyx_scope_struct_1___iter__(__pyx_ptype_8fastwarc_4warc___pyx_scope_struct_1___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 683, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8fastwarc_4warc_15ArchiveIterator_6generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_ArchiveIterator___iter, __pyx_n_s_fastwarc_warc); if (unlikely(!gen)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fastwarc.warc.ArchiveIterator.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8fastwarc_4warc_15ArchiveIterator_6generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *__pyx_cur_scope = ((struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  enum __pyx_t_8fastwarc_4warc__NextRecStatus __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)

  /* "fastwarc/warc.pyx":686
 *         cdef _NextRecStatus status
 * 
 *         while True:             # <<<<<<<<<<<<<<
 *             status = self._read_next_record()
 *             if status == has_next:
 */
  while (1) {

    /* "fastwarc/warc.pyx":687
 * 
 *         while True:
 *             status = self._read_next_record()             # <<<<<<<<<<<<<<
 *             if status == has_next:
 *                 yield self.record
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8fastwarc_4warc_ArchiveIterator *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_read_next_record(__pyx_cur_scope->__pyx_v_self); if (unlikely(__pyx_t_1 == ((enum __pyx_t_8fastwarc_4warc__NextRecStatus)__pyx_e_8fastwarc_4warc_error))) __PYX_ERR(0, 687, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_status = __pyx_t_1;

    /* "fastwarc/warc.pyx":688
 *         while True:
 *             status = self._read_next_record()
 *             if status == has_next:             # <<<<<<<<<<<<<<
 *                 yield self.record
 *             elif status == eof:
 */
    switch (__pyx_cur_scope->__pyx_v_status) {
      case __pyx_e_8fastwarc_4warc_has_next:

      /* "fastwarc/warc.pyx":689
 *             status = self._read_next_record()
 *             if status == has_next:
 *                 yield self.record             # <<<<<<<<<<<<<<
 *             elif status == eof:
 *                 self.reader.close()
 */
      __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self->record));
      __pyx_r = ((PyObject *)__pyx_cur_scope->__pyx_v_self->record);
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L6_resume_from_yield:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 689, __pyx_L1_error)

      /* "fastwarc/warc.pyx":688
 *         while True:
 *             status = self._read_next_record()
 *             if status == has_next:             # <<<<<<<<<<<<<<
 *                 yield self.record
 *             elif status == eof:
 */
      break;
      case __pyx_e_8fastwarc_4warc_eof:

      /* "fastwarc/warc.pyx":691
 *                 yield self.record
 *             elif status == eof:
 *                 self.reader.close()             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_cur_scope->__pyx_v_self->reader->__pyx_vtab)->close(__pyx_cur_scope->__pyx_v_self->reader, 0);

      /* "fastwarc/warc.pyx":692
 *             elif status == eof:
 *                 self.reader.close()
 *                 return             # <<<<<<<<<<<<<<
 * 
 *     cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = NULL;
      goto __pyx_L0;

      /* "fastwarc/warc.pyx":690
 *             if status == has_next:
 *                 yield self.record
 *             elif status == eof:             # <<<<<<<<<<<<<<
 *                 self.reader.close()
 *                 return
 */
      break;
      default: break;
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "fastwarc/warc.pyx":683
 *         self.record_type_filter = record_types
 * 
 *     def __iter__(self) -> Iterable[WarcRecord]:             # <<<<<<<<<<<<<<
 *         cdef _NextRecStatus status
 * 
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":694
 *                 return
 * 
 *     cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error:             # <<<<<<<<<<<<<<
 *         self.reader.detect_stream_type()
 * 
 */

static enum __pyx_t_8fastwarc_4warc__NextRecStatus __pyx_f_8fastwarc_4warc_15ArchiveIterator__read_next_record(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self) {
  std::string __pyx_v_version_line;
  std::string __pyx_v_hkey;
  size_t __pyx_v_parse_count;
  __pyx_t_8fastwarc_4warc_str_pair __pyx_v_h;
  int __pyx_v_skip;
  enum __pyx_t_8fastwarc_4warc__NextRecStatus __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_t_8 = NULL;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair> ::iterator __pyx_t_9;
  std::vector<__pyx_t_8fastwarc_4warc_str_pair>  *__pyx_t_10;
  __pyx_t_8fastwarc_4warc_str_pair __pyx_t_11;
  struct __pyx_opt_args_8fastwarc_9stream_io_14BufferedReader_consume __pyx_t_12;
  struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_next_record", 0);

  /* "fastwarc/warc.pyx":695
 * 
 *     cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error:
 *         self.reader.detect_stream_type()             # <<<<<<<<<<<<<<
 * 
 *         if self.record is not None:
 */
  ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->detect_stream_type(__pyx_v_self->reader);

  /* "fastwarc/warc.pyx":697
 *         self.reader.detect_stream_type()
 * 
 *         if self.record is not None:             # <<<<<<<<<<<<<<
 *             self.reader.consume()
 *             self.reader.reset_limit()
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->record) != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":698
 * 
 *         if self.record is not None:
 *             self.reader.consume()             # <<<<<<<<<<<<<<
 *             self.reader.reset_limit()
 * 
 */
    ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->consume(__pyx_v_self->reader, 0, NULL);

    /* "fastwarc/warc.pyx":699
 *         if self.record is not None:
 *             self.reader.consume()
 *             self.reader.reset_limit()             # <<<<<<<<<<<<<<
 * 
 *         self.record = WarcRecord.__new__(WarcRecord)
 */
    ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->reset_limit(__pyx_v_self->reader);

    /* "fastwarc/warc.pyx":697
 *         self.reader.detect_stream_type()
 * 
 *         if self.record is not None:             # <<<<<<<<<<<<<<
 *             self.reader.consume()
 *             self.reader.reset_limit()
 */
  }

  /* "fastwarc/warc.pyx":701
 *             self.reader.reset_limit()
 * 
 *         self.record = WarcRecord.__new__(WarcRecord)             # <<<<<<<<<<<<<<
 *         self.record._stream_pos = self.reader.tell()
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_tp_new_8fastwarc_4warc_WarcRecord(((PyTypeObject *)__pyx_ptype_8fastwarc_4warc_WarcRecord), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  __Pyx_GOTREF(__pyx_v_self->record);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->record));
  __pyx_v_self->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fastwarc/warc.pyx":702
 * 
 *         self.record = WarcRecord.__new__(WarcRecord)
 *         self.record._stream_pos = self.reader.tell()             # <<<<<<<<<<<<<<
 * 
 *         cdef string version_line
 */
  __pyx_v_self->record->_stream_pos = ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->tell(__pyx_v_self->reader, 0);

  /* "fastwarc/warc.pyx":705
 * 
 *         cdef string version_line
 *         while True:             # <<<<<<<<<<<<<<
 *             version_line = self.reader.readline()
 *             if not self.reader.error().empty():
 */
  while (1) {

    /* "fastwarc/warc.pyx":706
 *         cdef string version_line
 *         while True:
 *             version_line = self.reader.readline()             # <<<<<<<<<<<<<<
 *             if not self.reader.error().empty():
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \
 */
    __pyx_v_version_line = ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->readline(__pyx_v_self->reader, 0, NULL);

    /* "fastwarc/warc.pyx":707
 *         while True:
 *             version_line = self.reader.readline()
 *             if not self.reader.error().empty():             # <<<<<<<<<<<<<<
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 */
    __pyx_t_2 = ((!(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->error(__pyx_v_self->reader, 0).empty() != 0)) != 0);
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":708
 *             version_line = self.reader.readline()
 *             if not self.reader.error().empty():
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \             # <<<<<<<<<<<<<<
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 *                     raise (<PythonIOStreamAdapter>self.reader.stream).exc
 */
      __pyx_t_3 = ((PyObject *)__pyx_v_self->reader->stream);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_TypeCheck(__pyx_t_3, __pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter); 
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_4 = (__pyx_t_1 != 0);
      if (__pyx_t_4) {
      } else {
        __pyx_t_2 = __pyx_t_4;
        goto __pyx_L8_bool_binop_done;
      }

      /* "fastwarc/warc.pyx":709
 *             if not self.reader.error().empty():
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:             # <<<<<<<<<<<<<<
 *                     raise (<PythonIOStreamAdapter>self.reader.stream).exc
 *                 raise StreamError(self.reader.error().decode())
 */
      __pyx_t_4 = (((struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *)__pyx_v_self->reader->stream)->exc != ((PyObject*)Py_None));
      __pyx_t_1 = (__pyx_t_4 != 0);
      __pyx_t_2 = __pyx_t_1;
      __pyx_L8_bool_binop_done:;

      /* "fastwarc/warc.pyx":708
 *             version_line = self.reader.readline()
 *             if not self.reader.error().empty():
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \             # <<<<<<<<<<<<<<
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 *                     raise (<PythonIOStreamAdapter>self.reader.stream).exc
 */
      if (unlikely(__pyx_t_2)) {

        /* "fastwarc/warc.pyx":710
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 *                     raise (<PythonIOStreamAdapter>self.reader.stream).exc             # <<<<<<<<<<<<<<
 *                 raise StreamError(self.reader.error().decode())
 * 
 */
        __Pyx_Raise(((struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *)__pyx_v_self->reader->stream)->exc, 0, 0, 0);
        __PYX_ERR(0, 710, __pyx_L1_error)

        /* "fastwarc/warc.pyx":708
 *             version_line = self.reader.readline()
 *             if not self.reader.error().empty():
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \             # <<<<<<<<<<<<<<
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 *                     raise (<PythonIOStreamAdapter>self.reader.stream).exc
 */
      }

      /* "fastwarc/warc.pyx":711
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 *                     raise (<PythonIOStreamAdapter>self.reader.stream).exc
 *                 raise StreamError(self.reader.error().decode())             # <<<<<<<<<<<<<<
 * 
 *             if version_line.empty():
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_StreamError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_decode_cpp_string(((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->error(__pyx_v_self->reader, 0), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 711, __pyx_L1_error)

      /* "fastwarc/warc.pyx":707
 *         while True:
 *             version_line = self.reader.readline()
 *             if not self.reader.error().empty():             # <<<<<<<<<<<<<<
 *                 if isinstance(self.reader.stream, PythonIOStreamAdapter) and \
 *                         (<PythonIOStreamAdapter>self.reader.stream).exc is not None:
 */
    }

    /* "fastwarc/warc.pyx":713
 *                 raise StreamError(self.reader.error().decode())
 * 
 *             if version_line.empty():             # <<<<<<<<<<<<<<
 *                 # EOF
 *                 return eof
 */
    __pyx_t_2 = (__pyx_v_version_line.empty() != 0);
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":715
 *             if version_line.empty():
 *                 # EOF
 *                 return eof             # <<<<<<<<<<<<<<
 * 
 *             if version_line == b'\r\n' or version_line == b'\n':
 */
      __pyx_r = __pyx_e_8fastwarc_4warc_eof;
      goto __pyx_L0;

      /* "fastwarc/warc.pyx":713
 *                 raise StreamError(self.reader.error().decode())
 * 
 *             if version_line.empty():             # <<<<<<<<<<<<<<
 *                 # EOF
 *                 return eof
 */
    }

    /* "fastwarc/warc.pyx":717
 *                 return eof
 * 
 *             if version_line == b'\r\n' or version_line == b'\n':             # <<<<<<<<<<<<<<
 *                 # Consume empty lines
 *                 if not self.reader.stream_is_compressed:
 */
    __pyx_t_1 = ((__pyx_v_version_line == ((char const *)"\r\n")) != 0);
    if (!__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_version_line == ((char const *)"\n")) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L12_bool_binop_done:;
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":719
 *             if version_line == b'\r\n' or version_line == b'\n':
 *                 # Consume empty lines
 *                 if not self.reader.stream_is_compressed:             # <<<<<<<<<<<<<<
 *                     self.record._stream_pos += version_line.size()
 *                 continue
 */
      __pyx_t_2 = ((!(__pyx_v_self->reader->stream_is_compressed != 0)) != 0);
      if (__pyx_t_2) {

        /* "fastwarc/warc.pyx":720
 *                 # Consume empty lines
 *                 if not self.reader.stream_is_compressed:
 *                     self.record._stream_pos += version_line.size()             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        __Pyx_INCREF(((PyObject *)__pyx_v_self->record));
        __pyx_t_8 = __pyx_v_self->record;
        __pyx_t_8->_stream_pos = (__pyx_t_8->_stream_pos + __pyx_v_version_line.size());
        __Pyx_DECREF(((PyObject *)__pyx_t_8)); __pyx_t_8 = 0;

        /* "fastwarc/warc.pyx":719
 *             if version_line == b'\r\n' or version_line == b'\n':
 *                 # Consume empty lines
 *                 if not self.reader.stream_is_compressed:             # <<<<<<<<<<<<<<
 *                     self.record._stream_pos += version_line.size()
 *                 continue
 */
      }

      /* "fastwarc/warc.pyx":721
 *                 if not self.reader.stream_is_compressed:
 *                     self.record._stream_pos += version_line.size()
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             version_line = strip_str(version_line)
 */
      goto __pyx_L4_continue;

      /* "fastwarc/warc.pyx":717
 *                 return eof
 * 
 *             if version_line == b'\r\n' or version_line == b'\n':             # <<<<<<<<<<<<<<
 *                 # Consume empty lines
 *                 if not self.reader.stream_is_compressed:
 */
    }

    /* "fastwarc/warc.pyx":723
 *                 continue
 * 
 *             version_line = strip_str(version_line)             # <<<<<<<<<<<<<<
 *             if version_line == b'WARC/1.0' or version_line == b'WARC/1.1':
 *                 # OK, continue with parsing headers
 */
    __pyx_v_version_line = __pyx_f_8fastwarc_4warc_strip_str(__pyx_v_version_line);

    /* "fastwarc/warc.pyx":724
 * 
 *             version_line = strip_str(version_line)
 *             if version_line == b'WARC/1.0' or version_line == b'WARC/1.1':             # <<<<<<<<<<<<<<
 *                 # OK, continue with parsing headers
 *                 self.record._headers.set_status_line(version_line)
 */
    __pyx_t_1 = ((__pyx_v_version_line == ((char const *)"WARC/1.0")) != 0);
    if (!__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L16_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_version_line == ((char const *)"WARC/1.1")) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L16_bool_binop_done:;
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":726
 *             if version_line == b'WARC/1.0' or version_line == b'WARC/1.1':
 *                 # OK, continue with parsing headers
 *                 self.record._headers.set_status_line(version_line)             # <<<<<<<<<<<<<<
 *                 break
 *             else:
 */
      __pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line(__pyx_v_self->record->_headers, __pyx_v_version_line);

      /* "fastwarc/warc.pyx":727
 *                 # OK, continue with parsing headers
 *                 self.record._headers.set_status_line(version_line)
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 # Not a WARC file or unsupported version
 */
      goto __pyx_L5_break;

      /* "fastwarc/warc.pyx":724
 * 
 *             version_line = strip_str(version_line)
 *             if version_line == b'WARC/1.0' or version_line == b'WARC/1.1':             # <<<<<<<<<<<<<<
 *                 # OK, continue with parsing headers
 *                 self.record._headers.set_status_line(version_line)
 */
    }

    /* "fastwarc/warc.pyx":730
 *             else:
 *                 # Not a WARC file or unsupported version
 *                 return eof             # <<<<<<<<<<<<<<
 * 
 *         parse_header_block(self.reader, self.record._headers)
 */
    /*else*/ {
      __pyx_r = __pyx_e_8fastwarc_4warc_eof;
      goto __pyx_L0;
    }
    __pyx_L4_continue:;
  }
  __pyx_L5_break:;

  /* "fastwarc/warc.pyx":732
 *                 return eof
 * 
 *         parse_header_block(self.reader, self.record._headers)             # <<<<<<<<<<<<<<
 * 
 *         cdef string hkey
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->reader);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = ((PyObject *)__pyx_v_self->record->_headers);
  __Pyx_INCREF(__pyx_t_5);
  (void)(__pyx_f_8fastwarc_4warc_parse_header_block(((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)__pyx_t_3), ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)__pyx_t_5), NULL));
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":735
 * 
 *         cdef string hkey
 *         cdef size_t parse_count = 0             # <<<<<<<<<<<<<<
 *         cdef str_pair h
 *         cdef bint skip = False
 */
  __pyx_v_parse_count = 0;

  /* "fastwarc/warc.pyx":737
 *         cdef size_t parse_count = 0
 *         cdef str_pair h
 *         cdef bint skip = False             # <<<<<<<<<<<<<<
 *         for h in self.record._headers._headers:
 *             hkey = str_to_lower(h[0])
 */
  __pyx_v_skip = 0;

  /* "fastwarc/warc.pyx":738
 *         cdef str_pair h
 *         cdef bint skip = False
 *         for h in self.record._headers._headers:             # <<<<<<<<<<<<<<
 *             hkey = str_to_lower(h[0])
 *             if hkey == b'warc-type':
 */
  __pyx_t_10 = &__pyx_v_self->record->_headers->_headers;
  __pyx_t_9 = __pyx_t_10->begin();
  for (;;) {
    if (!(__pyx_t_9 != __pyx_t_10->end())) break;
    __pyx_t_11 = *__pyx_t_9;
    ++__pyx_t_9;
    __pyx_v_h = __pyx_t_11;

    /* "fastwarc/warc.pyx":739
 *         cdef bint skip = False
 *         for h in self.record._headers._headers:
 *             hkey = str_to_lower(h[0])             # <<<<<<<<<<<<<<
 *             if hkey == b'warc-type':
 *                 self.record._record_type = _str_record_type_to_enum(h[1])
 */
    __pyx_v_hkey = __pyx_f_8fastwarc_4warc_str_to_lower(__pyx_v_h.f0);

    /* "fastwarc/warc.pyx":740
 *         for h in self.record._headers._headers:
 *             hkey = str_to_lower(h[0])
 *             if hkey == b'warc-type':             # <<<<<<<<<<<<<<
 *                 self.record._record_type = _str_record_type_to_enum(h[1])
 *                 parse_count += 1
 */
    __pyx_t_2 = ((__pyx_v_hkey == ((char const *)"warc-type")) != 0);
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":741
 *             hkey = str_to_lower(h[0])
 *             if hkey == b'warc-type':
 *                 self.record._record_type = _str_record_type_to_enum(h[1])             # <<<<<<<<<<<<<<
 *                 parse_count += 1
 *             elif hkey == b'content-length':
 */
      __pyx_v_self->record->_record_type = __pyx_f_8fastwarc_4warc__str_record_type_to_enum(__pyx_v_h.f1);

      /* "fastwarc/warc.pyx":742
 *             if hkey == b'warc-type':
 *                 self.record._record_type = _str_record_type_to_enum(h[1])
 *                 parse_count += 1             # <<<<<<<<<<<<<<
 *             elif hkey == b'content-length':
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)
 */
      __pyx_v_parse_count = (__pyx_v_parse_count + 1);

      /* "fastwarc/warc.pyx":740
 *         for h in self.record._headers._headers:
 *             hkey = str_to_lower(h[0])
 *             if hkey == b'warc-type':             # <<<<<<<<<<<<<<
 *                 self.record._record_type = _str_record_type_to_enum(h[1])
 *                 parse_count += 1
 */
      goto __pyx_L20;
    }

    /* "fastwarc/warc.pyx":743
 *                 self.record._record_type = _str_record_type_to_enum(h[1])
 *                 parse_count += 1
 *             elif hkey == b'content-length':             # <<<<<<<<<<<<<<
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)
 *                 parse_count += 1
 */
    __pyx_t_2 = ((__pyx_v_hkey == ((char const *)"content-length")) != 0);
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":744
 *                 parse_count += 1
 *             elif hkey == b'content-length':
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)             # <<<<<<<<<<<<<<
 *                 parse_count += 1
 *             elif hkey == b'content-type' and h[1].find(b'application/http') == 0:
 */
      __pyx_v_self->record->_content_length = std::strtol(__pyx_v_h.f1.c_str(), NULL, 10);

      /* "fastwarc/warc.pyx":745
 *             elif hkey == b'content-length':
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)
 *                 parse_count += 1             # <<<<<<<<<<<<<<
 *             elif hkey == b'content-type' and h[1].find(b'application/http') == 0:
 *                 self.record._is_http = True
 */
      __pyx_v_parse_count = (__pyx_v_parse_count + 1);

      /* "fastwarc/warc.pyx":743
 *                 self.record._record_type = _str_record_type_to_enum(h[1])
 *                 parse_count += 1
 *             elif hkey == b'content-length':             # <<<<<<<<<<<<<<
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)
 *                 parse_count += 1
 */
      goto __pyx_L20;
    }

    /* "fastwarc/warc.pyx":746
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)
 *                 parse_count += 1
 *             elif hkey == b'content-type' and h[1].find(b'application/http') == 0:             # <<<<<<<<<<<<<<
 *                 self.record._is_http = True
 *                 parse_count += 1
 */
    __pyx_t_1 = ((__pyx_v_hkey == ((char const *)"content-type")) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L21_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_h.f1.find(((char const *)"application/http")) == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L21_bool_binop_done:;
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":747
 *                 parse_count += 1
 *             elif hkey == b'content-type' and h[1].find(b'application/http') == 0:
 *                 self.record._is_http = True             # <<<<<<<<<<<<<<
 *                 parse_count += 1
 * 
 */
      __pyx_v_self->record->_is_http = 1;

      /* "fastwarc/warc.pyx":748
 *             elif hkey == b'content-type' and h[1].find(b'application/http') == 0:
 *                 self.record._is_http = True
 *                 parse_count += 1             # <<<<<<<<<<<<<<
 * 
 *             if parse_count >= 3:
 */
      __pyx_v_parse_count = (__pyx_v_parse_count + 1);

      /* "fastwarc/warc.pyx":746
 *                 self.record._content_length = strtol(h[1].c_str(), NULL, 10)
 *                 parse_count += 1
 *             elif hkey == b'content-type' and h[1].find(b'application/http') == 0:             # <<<<<<<<<<<<<<
 *                 self.record._is_http = True
 *                 parse_count += 1
 */
    }
    __pyx_L20:;

    /* "fastwarc/warc.pyx":750
 *                 parse_count += 1
 * 
 *             if parse_count >= 3:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_2 = ((__pyx_v_parse_count >= 3) != 0);
    if (__pyx_t_2) {

      /* "fastwarc/warc.pyx":751
 * 
 *             if parse_count >= 3:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         # Check if record is to be skipped
 */
      goto __pyx_L19_break;

      /* "fastwarc/warc.pyx":750
 *                 parse_count += 1
 * 
 *             if parse_count >= 3:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    }

    /* "fastwarc/warc.pyx":738
 *         cdef str_pair h
 *         cdef bint skip = False
 *         for h in self.record._headers._headers:             # <<<<<<<<<<<<<<
 *             hkey = str_to_lower(h[0])
 *             if hkey == b'warc-type':
 */
  }
  __pyx_L19_break:;

  /* "fastwarc/warc.pyx":754
 * 
 *         # Check if record is to be skipped
 *         skip |= (self.record._record_type & self.record_type_filter) == 0             # <<<<<<<<<<<<<<
 *         skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length
 *         skip |= self.min_content_length != strnpos and self.record._content_length < self.max_content_length
 */
  __pyx_v_skip = (__pyx_v_skip | ((__pyx_v_self->record->_record_type & __pyx_v_self->record_type_filter) == 0));

  /* "fastwarc/warc.pyx":755
 *         # Check if record is to be skipped
 *         skip |= (self.record._record_type & self.record_type_filter) == 0
 *         skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length             # <<<<<<<<<<<<<<
 *         skip |= self.min_content_length != strnpos and self.record._content_length < self.max_content_length
 *         if not skip and self.func_filter is not None:
 */
  __pyx_t_1 = ((__pyx_v_self->max_content_length != std::string::npos) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L24_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_self->record->_content_length > __pyx_v_self->max_content_length) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L24_bool_binop_done:;
  __pyx_v_skip = (__pyx_v_skip | __pyx_t_2);

  /* "fastwarc/warc.pyx":756
 *         skip |= (self.record._record_type & self.record_type_filter) == 0
 *         skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length
 *         skip |= self.min_content_length != strnpos and self.record._content_length < self.max_content_length             # <<<<<<<<<<<<<<
 *         if not skip and self.func_filter is not None:
 *             # Execute expensive filters last and only if skip is not already true
 */
  __pyx_t_1 = ((__pyx_v_self->min_content_length != std::string::npos) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L26_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_self->record->_content_length < __pyx_v_self->max_content_length) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L26_bool_binop_done:;
  __pyx_v_skip = (__pyx_v_skip | __pyx_t_2);

  /* "fastwarc/warc.pyx":757
 *         skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length
 *         skip |= self.min_content_length != strnpos and self.record._content_length < self.max_content_length
 *         if not skip and self.func_filter is not None:             # <<<<<<<<<<<<<<
 *             # Execute expensive filters last and only if skip is not already true
 *             skip |= not self.func_filter(self.record)
 */
  __pyx_t_1 = ((!(__pyx_v_skip != 0)) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L29_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_v_self->func_filter != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L29_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":759
 *         if not skip and self.func_filter is not None:
 *             # Execute expensive filters last and only if skip is not already true
 *             skip |= not self.func_filter(self.record)             # <<<<<<<<<<<<<<
 * 
 *         if skip:
 */
    __Pyx_INCREF(__pyx_v_self->func_filter);
    __pyx_t_3 = __pyx_v_self->func_filter; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, ((PyObject *)__pyx_v_self->record)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_self->record));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 759, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_skip = (__pyx_v_skip | (!__pyx_t_2));

    /* "fastwarc/warc.pyx":757
 *         skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length
 *         skip |= self.min_content_length != strnpos and self.record._content_length < self.max_content_length
 *         if not skip and self.func_filter is not None:             # <<<<<<<<<<<<<<
 *             # Execute expensive filters last and only if skip is not already true
 *             skip |= not self.func_filter(self.record)
 */
  }

  /* "fastwarc/warc.pyx":761
 *             skip |= not self.func_filter(self.record)
 * 
 *         if skip:             # <<<<<<<<<<<<<<
 *             self.reader.consume(self.record._content_length)
 *             self.record = None
 */
  __pyx_t_2 = (__pyx_v_skip != 0);
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":762
 * 
 *         if skip:
 *             self.reader.consume(self.record._content_length)             # <<<<<<<<<<<<<<
 *             self.record = None
 *             return skip_next
 */
    __pyx_t_12.__pyx_n = 1;
    __pyx_t_12.size = __pyx_v_self->record->_content_length;
    ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->consume(__pyx_v_self->reader, 0, &__pyx_t_12); 

    /* "fastwarc/warc.pyx":763
 *         if skip:
 *             self.reader.consume(self.record._content_length)
 *             self.record = None             # <<<<<<<<<<<<<<
 *             return skip_next
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->record);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->record));
    __pyx_v_self->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)Py_None);

    /* "fastwarc/warc.pyx":764
 *             self.reader.consume(self.record._content_length)
 *             self.record = None
 *             return skip_next             # <<<<<<<<<<<<<<
 * 
 *         self.reader.set_limit(self.record._content_length)
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_skip_next;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":761
 *             skip |= not self.func_filter(self.record)
 * 
 *         if skip:             # <<<<<<<<<<<<<<
 *             self.reader.consume(self.record._content_length)
 *             self.record = None
 */
  }

  /* "fastwarc/warc.pyx":766
 *             return skip_next
 * 
 *         self.reader.set_limit(self.record._content_length)             # <<<<<<<<<<<<<<
 *         self.record._reader = self.reader
 * 
 */
  ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->set_limit(__pyx_v_self->reader, __pyx_v_self->record->_content_length);

  /* "fastwarc/warc.pyx":767
 * 
 *         self.reader.set_limit(self.record._content_length)
 *         self.record._reader = self.reader             # <<<<<<<<<<<<<<
 * 
 *         if self.verify_digests and not self.record.verify_block_digest(False):
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_self->reader);
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->record->_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->record->_reader));
  __pyx_v_self->record->_reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":769
 *         self.record._reader = self.reader
 * 
 *         if self.verify_digests and not self.record.verify_block_digest(False):             # <<<<<<<<<<<<<<
 *             self.reader.reset_limit()
 *             self.reader.consume(self.record._content_length)
 */
  __pyx_t_4 = (__pyx_v_self->verify_digests != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L33_bool_binop_done;
  }
  __pyx_t_13.__pyx_n = 1;
  __pyx_t_13.consume = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *)__pyx_v_self->record->__pyx_vtab)->verify_block_digest(__pyx_v_self->record, 0, &__pyx_t_13); 
  __pyx_t_1 = ((!(__pyx_t_4 != 0)) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L33_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":770
 * 
 *         if self.verify_digests and not self.record.verify_block_digest(False):
 *             self.reader.reset_limit()             # <<<<<<<<<<<<<<
 *             self.reader.consume(self.record._content_length)
 *             self.record = None
 */
    ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->reset_limit(__pyx_v_self->reader);

    /* "fastwarc/warc.pyx":771
 *         if self.verify_digests and not self.record.verify_block_digest(False):
 *             self.reader.reset_limit()
 *             self.reader.consume(self.record._content_length)             # <<<<<<<<<<<<<<
 *             self.record = None
 *             return skip_next
 */
    __pyx_t_12.__pyx_n = 1;
    __pyx_t_12.size = __pyx_v_self->record->_content_length;
    ((struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader *)__pyx_v_self->reader->__pyx_vtab)->consume(__pyx_v_self->reader, 0, &__pyx_t_12); 

    /* "fastwarc/warc.pyx":772
 *             self.reader.reset_limit()
 *             self.reader.consume(self.record._content_length)
 *             self.record = None             # <<<<<<<<<<<<<<
 *             return skip_next
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->record);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->record));
    __pyx_v_self->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)Py_None);

    /* "fastwarc/warc.pyx":773
 *             self.reader.consume(self.record._content_length)
 *             self.record = None
 *             return skip_next             # <<<<<<<<<<<<<<
 * 
 *         if self.parse_http and self.record._is_http:
 */
    __pyx_r = __pyx_e_8fastwarc_4warc_skip_next;
    goto __pyx_L0;

    /* "fastwarc/warc.pyx":769
 *         self.record._reader = self.reader
 * 
 *         if self.verify_digests and not self.record.verify_block_digest(False):             # <<<<<<<<<<<<<<
 *             self.reader.reset_limit()
 *             self.reader.consume(self.record._content_length)
 */
  }

  /* "fastwarc/warc.pyx":775
 *             return skip_next
 * 
 *         if self.parse_http and self.record._is_http:             # <<<<<<<<<<<<<<
 *             self.record.parse_http()
 * 
 */
  __pyx_t_1 = (__pyx_v_self->parse_http != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L36_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_v_self->record->_is_http != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L36_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fastwarc/warc.pyx":776
 * 
 *         if self.parse_http and self.record._is_http:
 *             self.record.parse_http()             # <<<<<<<<<<<<<<
 * 
 *         return has_next
 */
    ((struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord *)__pyx_v_self->record->__pyx_vtab)->parse_http(__pyx_v_self->record, 0);

    /* "fastwarc/warc.pyx":775
 *             return skip_next
 * 
 *         if self.parse_http and self.record._is_http:             # <<<<<<<<<<<<<<
 *             self.record.parse_http()
 * 
 */
  }

  /* "fastwarc/warc.pyx":778
 *             self.record.parse_http()
 * 
 *         return has_next             # <<<<<<<<<<<<<<
 * 
 *     cpdef void _set_stream(self, stream):
 */
  __pyx_r = __pyx_e_8fastwarc_4warc_has_next;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":694
 *                 return
 * 
 *     cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error:             # <<<<<<<<<<<<<<
 *         self.reader.detect_stream_type()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(((PyObject *)__pyx_t_8));
  __Pyx_AddTraceback("fastwarc.warc.ArchiveIterator._read_next_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = __pyx_e_8fastwarc_4warc_error;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":780
 *         return has_next
 * 
 *     cpdef void _set_stream(self, stream):             # <<<<<<<<<<<<<<
 *         """
 *         Replace underlying input stream.
 */

static PyObject *__pyx_pw_8fastwarc_4warc_15ArchiveIterator_8_set_stream(PyObject *__pyx_v_self, PyObject *__pyx_v_stream); /*proto*/
static void __pyx_f_8fastwarc_4warc_15ArchiveIterator__set_stream(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, PyObject *__pyx_v_stream, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_attr = NULL;
  struct __pyx_obj_8fastwarc_9stream_io_IOStream *__pyx_v_stream_ = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_stream", 0);
  __Pyx_INCREF(__pyx_v_stream);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_stream); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_8fastwarc_4warc_15ArchiveIterator_8_set_stream)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_stream) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_stream);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fastwarc/warc.pyx":786
 *         This method is for internal use and should not be called by external users.
 *         """
 *         if not isinstance(stream, IOStream):             # <<<<<<<<<<<<<<
 *             for attr in ('read', 'tell', 'close'):
 *                 if not hasattr(stream, attr):
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_stream, __pyx_ptype_8fastwarc_9stream_io_IOStream); 
  __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
  if (__pyx_t_6) {

    /* "fastwarc/warc.pyx":787
 *         """
 *         if not isinstance(stream, IOStream):
 *             for attr in ('read', 'tell', 'close'):             # <<<<<<<<<<<<<<
 *                 if not hasattr(stream, attr):
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
 */
    __pyx_t_1 = __pyx_tuple__16; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= 3) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 787, __pyx_L1_error)
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_attr, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "fastwarc/warc.pyx":788
 *         if not isinstance(stream, IOStream):
 *             for attr in ('read', 'tell', 'close'):
 *                 if not hasattr(stream, attr):             # <<<<<<<<<<<<<<
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
 *             stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)
 */
      __pyx_t_6 = __Pyx_HasAttr(__pyx_v_stream, __pyx_v_attr); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 788, __pyx_L1_error)
      __pyx_t_5 = ((!(__pyx_t_6 != 0)) != 0);
      if (unlikely(__pyx_t_5)) {

        /* "fastwarc/warc.pyx":789
 *             for attr in ('read', 'tell', 'close'):
 *                 if not hasattr(stream, attr):
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")             # <<<<<<<<<<<<<<
 *             stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)
 * 
 */
        __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = 0;
        __pyx_t_9 = 127;
        __Pyx_INCREF(__pyx_kp_u_Object_of_type);
        __pyx_t_8 += 16;
        __Pyx_GIVEREF(__pyx_kp_u_Object_of_type);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Object_of_type);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_stream)), __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u_has_no_attribute);
        __pyx_t_8 += 20;
        __Pyx_GIVEREF(__pyx_kp_u_has_no_attribute);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_has_no_attribute);
        __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_attr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u__17);
        __pyx_t_8 += 2;
        __Pyx_GIVEREF(__pyx_kp_u__17);
        PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_kp_u__17);
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_2, 5, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 789, __pyx_L1_error)

        /* "fastwarc/warc.pyx":788
 *         if not isinstance(stream, IOStream):
 *             for attr in ('read', 'tell', 'close'):
 *                 if not hasattr(stream, attr):             # <<<<<<<<<<<<<<
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
 *             stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)
 */
      }

      /* "fastwarc/warc.pyx":787
 *         """
 *         if not isinstance(stream, IOStream):
 *             for attr in ('read', 'tell', 'close'):             # <<<<<<<<<<<<<<
 *                 if not hasattr(stream, attr):
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fastwarc/warc.pyx":790
 *                 if not hasattr(stream, attr):
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
 *             stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)             # <<<<<<<<<<<<<<
 * 
 *         cdef IOStream stream_ = <IOStream>stream
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_stream);
    __Pyx_GIVEREF(__pyx_v_stream);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_stream);
    __pyx_t_2 = __Pyx_tp_new(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter), ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_stream, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "fastwarc/warc.pyx":786
 *         This method is for internal use and should not be called by external users.
 *         """
 *         if not isinstance(stream, IOStream):             # <<<<<<<<<<<<<<
 *             for attr in ('read', 'tell', 'close'):
 *                 if not hasattr(stream, attr):
 */
  }

  /* "fastwarc/warc.pyx":792
 *             stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)
 * 
 *         cdef IOStream stream_ = <IOStream>stream             # <<<<<<<<<<<<<<
 *         self.reader = BufferedReader.__new__(BufferedReader, stream_)
 *         self.record = None
 */
  __pyx_t_2 = __pyx_v_stream;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_stream_ = ((struct __pyx_obj_8fastwarc_9stream_io_IOStream *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":793
 * 
 *         cdef IOStream stream_ = <IOStream>stream
 *         self.reader = BufferedReader.__new__(BufferedReader, stream_)             # <<<<<<<<<<<<<<
 *         self.record = None
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_stream_));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_stream_));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_stream_));
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_8fastwarc_9stream_io_BufferedReader), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8fastwarc_9stream_io_BufferedReader)))) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->reader));
  __pyx_v_self->reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":794
 *         cdef IOStream stream_ = <IOStream>stream
 *         self.reader = BufferedReader.__new__(BufferedReader, stream_)
 *         self.record = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->record);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->record));
  __pyx_v_self->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)Py_None);

  /* "fastwarc/warc.pyx":780
 *         return has_next
 * 
 *     cpdef void _set_stream(self, stream):             # <<<<<<<<<<<<<<
 *         """
 *         Replace underlying input stream.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("fastwarc.warc.ArchiveIterator._set_stream", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_attr);
  __Pyx_XDECREF((PyObject *)__pyx_v_stream_);
  __Pyx_XDECREF(__pyx_v_stream);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_15ArchiveIterator_8_set_stream(PyObject *__pyx_v_self, PyObject *__pyx_v_stream); /*proto*/
static char __pyx_doc_8fastwarc_4warc_15ArchiveIterator_7_set_stream[] = "\n        Replace underlying input stream.\n        \n        This method is for internal use and should not be called by external users.\n        ";
static PyObject *__pyx_pw_8fastwarc_4warc_15ArchiveIterator_8_set_stream(PyObject *__pyx_v_self, PyObject *__pyx_v_stream) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_stream (wrapper)", 0);
  __pyx_r = __pyx_pf_8fastwarc_4warc_15ArchiveIterator_7_set_stream(((struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)__pyx_v_self), ((PyObject *)__pyx_v_stream));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_15ArchiveIterator_7_set_stream(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *__pyx_v_self, PyObject *__pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_stream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_8fastwarc_4warc_15ArchiveIterator__set_stream(__pyx_v_self, __pyx_v_stream, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.ArchiveIterator._set_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":798
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_warc_10(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is a WARC/1.0 record."""
 *     return record._headers._status_line == <char*>b'WARC/1.0'
 */

static PyObject *__pyx_pw_8fastwarc_4warc_1is_warc_10(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_warc_10(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_warc_10", 0);

  /* "fastwarc/warc.pyx":800
 * cpdef bint is_warc_10(WarcRecord record):
 *     """Filter predicate for checking if record is a WARC/1.0 record."""
 *     return record._headers._status_line == <char*>b'WARC/1.0'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_record->_headers->_status_line == ((char *)((char *)"WARC/1.0")));
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":798
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_warc_10(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is a WARC/1.0 record."""
 *     return record._headers._status_line == <char*>b'WARC/1.0'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_1is_warc_10(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_8fastwarc_4warc_is_warc_10[] = "Filter predicate for checking if record is a WARC/1.0 record.";
static PyObject *__pyx_pw_8fastwarc_4warc_1is_warc_10(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_warc_10 (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record), __pyx_ptype_8fastwarc_4warc_WarcRecord, 1, "record", 0))) __PYX_ERR(0, 798, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_is_warc_10(__pyx_self, ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_record));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_is_warc_10(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_warc_10", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_8fastwarc_4warc_is_warc_10(__pyx_v_record, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.is_warc_10", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":804
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_warc_11(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is a WARC/1.1 record."""
 *     return record._headers._status_line == <char*>b'WARC/1.1'
 */

static PyObject *__pyx_pw_8fastwarc_4warc_3is_warc_11(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_warc_11(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_warc_11", 0);

  /* "fastwarc/warc.pyx":806
 * cpdef bint is_warc_11(WarcRecord record):
 *     """Filter predicate for checking if record is a WARC/1.1 record."""
 *     return record._headers._status_line == <char*>b'WARC/1.1'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_record->_headers->_status_line == ((char *)((char *)"WARC/1.1")));
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":804
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_warc_11(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is a WARC/1.1 record."""
 *     return record._headers._status_line == <char*>b'WARC/1.1'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_3is_warc_11(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_8fastwarc_4warc_2is_warc_11[] = "Filter predicate for checking if record is a WARC/1.1 record.";
static PyObject *__pyx_pw_8fastwarc_4warc_3is_warc_11(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_warc_11 (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record), __pyx_ptype_8fastwarc_4warc_WarcRecord, 1, "record", 0))) __PYX_ERR(0, 804, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_2is_warc_11(__pyx_self, ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_record));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_2is_warc_11(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_warc_11", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_8fastwarc_4warc_is_warc_11(__pyx_v_record, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.is_warc_11", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":810
 * 
 * # noinspection PyProtectedMember
 * cpdef bint has_block_digest(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record has a block digest."""
 *     return not record._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b'').empty()
 */

static PyObject *__pyx_pw_8fastwarc_4warc_5has_block_digest(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static int __pyx_f_8fastwarc_4warc_has_block_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_block_digest", 0);

  /* "fastwarc/warc.pyx":812
 * cpdef bint has_block_digest(WarcRecord record):
 *     """Filter predicate for checking if record has a block digest."""
 *     return not record._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b'').empty()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (!(((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_record->_headers->__pyx_vtab)->find_header(__pyx_v_record->_headers, ((char *)((char *)"WARC-Block-Digest")), ((char *)((char *)""))).empty() != 0));
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":810
 * 
 * # noinspection PyProtectedMember
 * cpdef bint has_block_digest(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record has a block digest."""
 *     return not record._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b'').empty()
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_5has_block_digest(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_8fastwarc_4warc_4has_block_digest[] = "Filter predicate for checking if record has a block digest.";
static PyObject *__pyx_pw_8fastwarc_4warc_5has_block_digest(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_block_digest (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record), __pyx_ptype_8fastwarc_4warc_WarcRecord, 1, "record", 0))) __PYX_ERR(0, 810, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_4has_block_digest(__pyx_self, ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_record));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_4has_block_digest(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_block_digest", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_8fastwarc_4warc_has_block_digest(__pyx_v_record, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.has_block_digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":816
 * 
 * # noinspection PyProtectedMember
 * cpdef bint has_payload_digest(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record has a payload digest."""
 *     return not record._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b'').empty()
 */

static PyObject *__pyx_pw_8fastwarc_4warc_7has_payload_digest(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static int __pyx_f_8fastwarc_4warc_has_payload_digest(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_payload_digest", 0);

  /* "fastwarc/warc.pyx":818
 * cpdef bint has_payload_digest(WarcRecord record):
 *     """Filter predicate for checking if record has a payload digest."""
 *     return not record._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b'').empty()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (!(((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_record->_headers->__pyx_vtab)->find_header(__pyx_v_record->_headers, ((char *)((char *)"WARC-Payload-Digest")), ((char *)((char *)""))).empty() != 0));
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":816
 * 
 * # noinspection PyProtectedMember
 * cpdef bint has_payload_digest(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record has a payload digest."""
 *     return not record._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b'').empty()
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_7has_payload_digest(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_8fastwarc_4warc_6has_payload_digest[] = "Filter predicate for checking if record has a payload digest.";
static PyObject *__pyx_pw_8fastwarc_4warc_7has_payload_digest(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_payload_digest (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record), __pyx_ptype_8fastwarc_4warc_WarcRecord, 1, "record", 0))) __PYX_ERR(0, 816, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_6has_payload_digest(__pyx_self, ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_record));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_6has_payload_digest(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_payload_digest", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_8fastwarc_4warc_has_payload_digest(__pyx_v_record, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.has_payload_digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":822
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_http(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is an HTTP record."""
 *     return record._is_http
 */

static PyObject *__pyx_pw_8fastwarc_4warc_9is_http(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_http(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_http", 0);

  /* "fastwarc/warc.pyx":824
 * cpdef bint is_http(WarcRecord record):
 *     """Filter predicate for checking if record is an HTTP record."""
 *     return record._is_http             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_record->_is_http;
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":822
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_http(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is an HTTP record."""
 *     return record._is_http
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_9is_http(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_8fastwarc_4warc_8is_http[] = "Filter predicate for checking if record is an HTTP record.";
static PyObject *__pyx_pw_8fastwarc_4warc_9is_http(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_http (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record), __pyx_ptype_8fastwarc_4warc_WarcRecord, 1, "record", 0))) __PYX_ERR(0, 822, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_8is_http(__pyx_self, ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_record));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_8is_http(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_http", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_8fastwarc_4warc_is_http(__pyx_v_record, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.is_http", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/warc.pyx":828
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_concurrent(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is concurrent to another record."""
 *     return not record._headers.find_header(<char*>b'WARC-Concurrent-To', <char*>b'').empty()
 */

static PyObject *__pyx_pw_8fastwarc_4warc_11is_concurrent(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static int __pyx_f_8fastwarc_4warc_is_concurrent(struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_concurrent", 0);

  /* "fastwarc/warc.pyx":830
 * cpdef bint is_concurrent(WarcRecord record):
 *     """Filter predicate for checking if record is concurrent to another record."""
 *     return not record._headers.find_header(<char*>b'WARC-Concurrent-To', <char*>b'').empty()             # <<<<<<<<<<<<<<
 */
  __pyx_r = (!(((struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap *)__pyx_v_record->_headers->__pyx_vtab)->find_header(__pyx_v_record->_headers, ((char *)((char *)"WARC-Concurrent-To")), ((char *)((char *)""))).empty() != 0));
  goto __pyx_L0;

  /* "fastwarc/warc.pyx":828
 * 
 * # noinspection PyProtectedMember
 * cpdef bint is_concurrent(WarcRecord record):             # <<<<<<<<<<<<<<
 *     """Filter predicate for checking if record is concurrent to another record."""
 *     return not record._headers.find_header(<char*>b'WARC-Concurrent-To', <char*>b'').empty()
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8fastwarc_4warc_11is_concurrent(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_8fastwarc_4warc_10is_concurrent[] = "Filter predicate for checking if record is concurrent to another record.";
static PyObject *__pyx_pw_8fastwarc_4warc_11is_concurrent(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_concurrent (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_record), __pyx_ptype_8fastwarc_4warc_WarcRecord, 1, "record", 0))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_r = __pyx_pf_8fastwarc_4warc_10is_concurrent(__pyx_self, ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)__pyx_v_record));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8fastwarc_4warc_10is_concurrent(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_8fastwarc_4warc_WarcRecord *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_concurrent", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_8fastwarc_4warc_is_concurrent(__pyx_v_record, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fastwarc.warc.is_concurrent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/stream_io.pxd":42
 *     cdef Exception exc
 * 
 *     cdef inline size_t tell(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.py_stream.tell()
 */

static size_t __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_tell(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  std::string __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tell", 0);

  /* "fastwarc/stream_io.pxd":43
 * 
 *     cdef inline size_t tell(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.tell()
 *         except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastwarc/stream_io.pxd":44
 *     cdef inline size_t tell(self):
 *         try:
 *             return self.py_stream.tell()             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_stream, __pyx_n_s_tell); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 44, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 44, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 44, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_7;
      goto __pyx_L7_try_return;

      /* "fastwarc/stream_io.pxd":43
 * 
 *     cdef inline size_t tell(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.tell()
 *         except Exception as e:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastwarc/stream_io.pxd":45
 *         try:
 *             return self.py_stream.tell()
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("fastwarc.stream_io.PythonIOStreamAdapter.tell", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(2, 45, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "fastwarc/stream_io.pxd":46
 *             return self.py_stream.tell()
 *         except Exception as e:
 *             self.errstr = str(e).encode()             # <<<<<<<<<<<<<<
 *             self.exc = e
 *             return 0
 */
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_e); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 46, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_9), NULL, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 46, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_11 = __pyx_convert_string_from_py_std__in_string(__pyx_t_10); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 46, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_v_self->__pyx_base.errstr = __pyx_t_11;

        /* "fastwarc/stream_io.pxd":47
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 *             self.exc = e             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
        if (!(likely(__Pyx_PyException_Check(__pyx_v_e))||((__pyx_v_e) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "Exception", Py_TYPE(__pyx_v_e)->tp_name), 0))) __PYX_ERR(2, 47, __pyx_L14_error)
        __pyx_t_10 = __pyx_v_e;
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_10);
        __Pyx_GOTREF(__pyx_v_self->exc);
        __Pyx_DECREF(__pyx_v_self->exc);
        __pyx_v_self->exc = ((PyObject*)__pyx_t_10);
        __pyx_t_10 = 0;

        /* "fastwarc/stream_io.pxd":48
 *             self.errstr = str(e).encode()
 *             self.exc = e
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline string read(self, size_t size):
 */
        __pyx_r = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L13_return;
      }

      /* "fastwarc/stream_io.pxd":45
 *         try:
 *             return self.py_stream.tell()
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_7 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_7;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastwarc/stream_io.pxd":43
 * 
 *     cdef inline size_t tell(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.tell()
 *         except Exception as e:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "fastwarc/stream_io.pxd":42
 *     cdef Exception exc
 * 
 *     cdef inline size_t tell(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.py_stream.tell()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("fastwarc.stream_io.PythonIOStreamAdapter.tell", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/stream_io.pxd":50
 *             return 0
 * 
 *     cdef inline string read(self, size_t size):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.py_stream.read(size)[:size]
 */

static std::string __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_read(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self, size_t __pyx_v_size) {
  PyObject *__pyx_v_e = NULL;
  std::string __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  std::string __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read", 0);

  /* "fastwarc/stream_io.pxd":51
 * 
 *     cdef inline string read(self, size_t size):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.read(size)[:size]
 *         except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastwarc/stream_io.pxd":52
 *     cdef inline string read(self, size_t size):
 *         try:
 *             return self.py_stream.read(size)[:size]             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_stream, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 52, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 52, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 52, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, __pyx_v_size, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 52, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __pyx_convert_string_from_py_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 52, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_8;
      goto __pyx_L7_try_return;

      /* "fastwarc/stream_io.pxd":51
 * 
 *     cdef inline string read(self, size_t size):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.read(size)[:size]
 *         except Exception as e:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastwarc/stream_io.pxd":53
 *         try:
 *             return self.py_stream.read(size)[:size]
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("fastwarc.stream_io.PythonIOStreamAdapter.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(2, 53, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_e = __pyx_t_4;
      /*try:*/ {

        /* "fastwarc/stream_io.pxd":54
 *             return self.py_stream.read(size)[:size]
 *         except Exception as e:
 *             self.errstr = str(e).encode()             # <<<<<<<<<<<<<<
 *             self.exc = e
 *             return string()
 */
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_e); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 54, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_7), NULL, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 54, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = __pyx_convert_string_from_py_std__in_string(__pyx_t_10); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 54, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_v_self->__pyx_base.errstr = __pyx_t_8;

        /* "fastwarc/stream_io.pxd":55
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 *             self.exc = e             # <<<<<<<<<<<<<<
 *             return string()
 * 
 */
        if (!(likely(__Pyx_PyException_Check(__pyx_v_e))||((__pyx_v_e) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "Exception", Py_TYPE(__pyx_v_e)->tp_name), 0))) __PYX_ERR(2, 55, __pyx_L14_error)
        __pyx_t_10 = __pyx_v_e;
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_10);
        __Pyx_GOTREF(__pyx_v_self->exc);
        __Pyx_DECREF(__pyx_v_self->exc);
        __pyx_v_self->exc = ((PyObject*)__pyx_t_10);
        __pyx_t_10 = 0;

        /* "fastwarc/stream_io.pxd":56
 *             self.errstr = str(e).encode()
 *             self.exc = e
 *             return string()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline size_t write(self, const char* data, size_t size):
 */
        try {
          __pyx_t_8 = std::string();
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(2, 56, __pyx_L14_error)
        }
        __pyx_r = __pyx_t_8;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L13_return;
      }

      /* "fastwarc/stream_io.pxd":53
 *         try:
 *             return self.py_stream.read(size)[:size]
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_9 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_8 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_8;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastwarc/stream_io.pxd":51
 * 
 *     cdef inline string read(self, size_t size):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.read(size)[:size]
 *         except Exception as e:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "fastwarc/stream_io.pxd":50
 *             return 0
 * 
 *     cdef inline string read(self, size_t size):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.py_stream.read(size)[:size]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("fastwarc.stream_io.PythonIOStreamAdapter.read", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/stream_io.pxd":58
 *             return string()
 * 
 *     cdef inline size_t write(self, const char* data, size_t size):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.py_stream.write(data[:size])
 */

static size_t __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_write(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self, char const *__pyx_v_data, size_t __pyx_v_size) {
  PyObject *__pyx_v_e = NULL;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  std::string __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 0);

  /* "fastwarc/stream_io.pxd":59
 * 
 *     cdef inline size_t write(self, const char* data, size_t size):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.write(data[:size])
 *         except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastwarc/stream_io.pxd":60
 *     cdef inline size_t write(self, const char* data, size_t size):
 *         try:
 *             return self.py_stream.write(data[:size])             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_stream, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_data + 0, __pyx_v_size - 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 60, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_8;
      goto __pyx_L7_try_return;

      /* "fastwarc/stream_io.pxd":59
 * 
 *     cdef inline size_t write(self, const char* data, size_t size):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.write(data[:size])
 *         except Exception as e:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fastwarc/stream_io.pxd":61
 *         try:
 *             return self.py_stream.write(data[:size])
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("fastwarc.stream_io.PythonIOStreamAdapter.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(2, 61, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "fastwarc/stream_io.pxd":62
 *             return self.py_stream.write(data[:size])
 *         except Exception as e:
 *             self.errstr = str(e).encode()             # <<<<<<<<<<<<<<
 *             self.exc = e
 *             return 0
 */
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_e); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 62, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_7), NULL, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 62, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __pyx_convert_string_from_py_std__in_string(__pyx_t_10); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 62, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_v_self->__pyx_base.errstr = __pyx_t_11;

        /* "fastwarc/stream_io.pxd":63
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 *             self.exc = e             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
        if (!(likely(__Pyx_PyException_Check(__pyx_v_e))||((__pyx_v_e) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "Exception", Py_TYPE(__pyx_v_e)->tp_name), 0))) __PYX_ERR(2, 63, __pyx_L14_error)
        __pyx_t_10 = __pyx_v_e;
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_10);
        __Pyx_GOTREF(__pyx_v_self->exc);
        __Pyx_DECREF(__pyx_v_self->exc);
        __pyx_v_self->exc = ((PyObject*)__pyx_t_10);
        __pyx_t_10 = 0;

        /* "fastwarc/stream_io.pxd":64
 *             self.errstr = str(e).encode()
 *             self.exc = e
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void flush(self):
 */
        __pyx_r = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L13_return;
      }

      /* "fastwarc/stream_io.pxd":61
 *         try:
 *             return self.py_stream.write(data[:size])
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_9 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
        __pyx_L13_return: {
          __pyx_t_8 = __pyx_r;
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          __pyx_r = __pyx_t_8;
          goto __pyx_L6_except_return;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastwarc/stream_io.pxd":59
 * 
 *     cdef inline size_t write(self, const char* data, size_t size):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.py_stream.write(data[:size])
 *         except Exception as e:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "fastwarc/stream_io.pxd":58
 *             return string()
 * 
 *     cdef inline size_t write(self, const char* data, size_t size):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.py_stream.write(data[:size])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("fastwarc.stream_io.PythonIOStreamAdapter.write", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fastwarc/stream_io.pxd":66
 *             return 0
 * 
 *     cdef inline void flush(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.py_stream.flush()
 */

static void __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_flush(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  std::string __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("flush", 0);

  /* "fastwarc/stream_io.pxd":67
 * 
 *     cdef inline void flush(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.py_stream.flush()
 *         except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fastwarc/stream_io.pxd":68
 *     cdef inline void flush(self):
 *         try:
 *             self.py_stream.flush()             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_stream, __pyx_n_s_flush); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 68, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 68, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fastwarc/stream_io.pxd":67
 * 
 *     cdef inline void flush(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.py_stream.flush()
 *         except Exception as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fastwarc/stream_io.pxd":69
 *         try:
 *             self.py_stream.flush()
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_AddTraceback("fastwarc.stream_io.PythonIOStreamAdapter.flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(2, 69, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "fastwarc/stream_io.pxd":70
 *             self.py_stream.flush()
 *         except Exception as e:
 *             self.errstr = str(e).encode()             # <<<<<<<<<<<<<<
 *             self.exc = e
 * 
 */
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_e); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 70, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_8), NULL, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 70, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_10 = __pyx_convert_string_from_py_std__in_string(__pyx_t_9); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 70, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_v_self->__pyx_base.errstr = __pyx_t_10;

        /* "fastwarc/stream_io.pxd":71
 *         except Exception as e:
 *             self.errstr = str(e).encode()
 *             self.exc = e             # <<<<<<<<<<<<<<
 * 
 *     cdef inline void close(self):
 */
        if (!(likely(__Pyx_PyException_Check(__pyx_v_e))||((__pyx_v_e) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "Exception", Py_TYPE(__pyx_v_e)->tp_name), 0))) __PYX_ERR(2, 71, __pyx_L14_error)
        __pyx_t_9 = __pyx_v_e;
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF(__pyx_v_self->exc);
        __Pyx_DECREF(__pyx_v_self->exc);
        __pyx_v_self->exc = ((PyObject*)__pyx_t_9);
        __pyx_t_9 = 0;
      }

      /* "fastwarc/stream_io.pxd":69
 *         try:
 *             self.py_stream.flush()
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.errstr = str(e).encode()
 *             self.exc = e
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_7 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fastwarc/stream_io.pxd":67
 * 
 *     cdef inline void flush(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.py_stream.flush()
 *         except Exception as e:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "fastwarc/stream_io.pxd":66
 *             return 0
 * 
 *     cdef inline void flush(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.py_stream.flush()
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("fastwarc.stream_io.PythonIOStreamAdapter.flush", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
}

/* "fastwarc/stream_io.pxd":73
 *             self.exc = e
 * 
 *     cdef inline void close(self):             # <<<<<<<<<<<<<<
 *         if self.py_stream is None:
 *             return
 */

static void __pyx_f_8fastwarc_9stream_io_21PythonIOStreamAdapter_close(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "fastwarc/stream_io.pxd":74
 * 
 *     cdef inline void close(self):
 *         if self.py_stream is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->py_stream == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "fastwarc/stream_io.pxd":75
 *     cdef inline void close(self):
 *         if self.py_stream is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # noinspection PyBroadException
 */
    goto __pyx_L0;

    /* "fastwarc/stream_io.pxd":74
 * 
 *     cdef inline void close(self):
 *         if self.py_stream is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "fastwarc/stream_io.pxd":78
 * 
 *         # noinspection PyBroadException
 *         try:             # <<<<<<<<<<<<<<
 *             self.py_stream.close()
 *         except:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "fastwarc/stream_io.pxd":79
 *         # noinspection PyBroadException
 *         try:
 *             self.py_stream.close()             # <<<<<<<<<<<<<<
 *         except:
 *             # Ignore exceptions during close()
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_stream, __pyx_n_s_close); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 79, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 79, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fastwarc/stream_io.pxd":78
 * 
 *         # noinspection PyBroadException
 *         try:             # <<<<<<<<<<<<<<
 *             self.py_stream.close()
 *         except:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "fastwarc/stream_io.pxd":80
 *         try:
 *             self.py_stream.close()
 *         except:             # <<<<<<<<<<<<<<
 *             # Ignore exceptions during close()
 *             pass
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L5_exception_handled;
    }
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L9_try_end:;
  }

  /* "fastwarc/stream_io.pxd":73
 *             self.exc = e
 * 
 *     cdef inline void close(self):             # <<<<<<<<<<<<<<
 *         if self.py_stream is None:
 *             return
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "string.to_py":31
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyObject_string_to_py_std__in_string", 0);

  /* "string.to_py":32
 * @cname("__pyx_convert_PyObject_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyUnicode_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":31
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyObject_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":37
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyUnicode_string_to_py_std__in_string", 0);

  /* "string.to_py":38
 * @cname("__pyx_convert_PyUnicode_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyStr_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":37
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyUnicode_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":43
 * 
 * @cname("__pyx_convert_PyStr_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyStr_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyStr_string_to_py_std__in_string", 0);

  /* "string.to_py":44
 * @cname("__pyx_convert_PyStr_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyBytes_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyStr_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":43
 * 
 * @cname("__pyx_convert_PyStr_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyStr_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":49
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyBytes_string_to_py_std__in_string", 0);

  /* "string.to_py":50
 * @cname("__pyx_convert_PyBytes_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyByteArray_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":49
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyBytes_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":55
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())
 * 
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyByteArray_string_to_py_std__in_string", 0);

  /* "string.to_py":56
 * @cname("__pyx_convert_PyByteArray_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyByteArray_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":55
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyByteArray_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

static std::string __pyx_convert_string_from_py_std__in_string(PyObject *__pyx_v_o) {
  Py_ssize_t __pyx_v_length;
  char const *__pyx_v_data;
  std::string __pyx_r;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_string_from_py_std__in_string", 0);

  /* "string.from_py":14
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0             # <<<<<<<<<<<<<<
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)
 */
  __pyx_v_length = 0;

  /* "string.from_py":15
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)             # <<<<<<<<<<<<<<
 *     return string(data, length)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_AsStringAndSize(__pyx_v_o, (&__pyx_v_length)); if (unlikely(__pyx_t_1 == ((char const *)NULL))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_v_data = __pyx_t_1;

  /* "string.from_py":16
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = std::string(__pyx_v_data, __pyx_v_length);
  goto __pyx_L0;

  /* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("string.from_py.__pyx_convert_string_from_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

/* Python wrapper */
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parents = 0;
  PyObject *__pyx_v_dct = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_parents,&__pyx_n_s_dct,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parents)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(1, 16, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dct)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(1, 16, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_name = values[0];
    __pyx_v_parents = values[1];
    __pyx_v_dct = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "EnumBase":17
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)             # <<<<<<<<<<<<<<
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_cls));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cls));
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_cls));
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_parents);
    __Pyx_GIVEREF(__pyx_v_parents);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_parents);
    __Pyx_INCREF(__pyx_v_dct);
    __Pyx_GIVEREF(__pyx_v_dct);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_dct);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":18
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()             # <<<<<<<<<<<<<<
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 */
  __Pyx_INCREF(__Pyx_OrderedDict);
  __pyx_t_2 = __Pyx_OrderedDict; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members, __pyx_t_1) < 0) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "EnumBase":20
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())             # <<<<<<<<<<<<<<
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "EnumBase":22
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle___Pyx_EnumMeta__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__ = {"__new__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__new__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_value,&__pyx_n_s_name_2,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, 1); __PYX_ERR(1, 28, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__new__") < 0)) __PYX_ERR(1, 28, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_value = values[1];
    __pyx_v_name = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 28, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(__pyx_self, __pyx_v_cls, __pyx_v_value, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__new__", 0);

  /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  if (likely(PyList_CheckExact(__pyx_v_cls)) || PyTuple_CheckExact(__pyx_v_cls)) {
    __pyx_t_1 = __pyx_v_cls; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 29, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 29, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 29, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 29, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "EnumBase":31
 *         for v in cls:
 *             if v == value:
 *                 return v             # <<<<<<<<<<<<<<
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_v);
      __pyx_r = __pyx_v_v;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    }

    /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  __pyx_t_5 = (__pyx_v_name == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_6)) {

    /* "EnumBase":33
 *                 return v
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)             # <<<<<<<<<<<<<<
 *         res = int.__new__(cls, value)
 *         res.name = name
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unknown_enum_value_s, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 33, __pyx_L1_error)

    /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  }

  /* "EnumBase":34
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)             # <<<<<<<<<<<<<<
 *         res.name = name
 *         setattr(cls, name, res)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_cls);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_value);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_res = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "EnumBase":35
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 *         res.name = name             # <<<<<<<<<<<<<<
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_res, __pyx_n_s_name_2, __pyx_v_name) < 0) __PYX_ERR(1, 35, __pyx_L1_error)

  /* "EnumBase":36
 *         res = int.__new__(cls, value)
 *         res.name = name
 *         setattr(cls, name, res)             # <<<<<<<<<<<<<<
 *         cls.__members__[name] = res
 *         return res
 */
  __pyx_t_10 = PyObject_SetAttr(__pyx_v_cls, __pyx_v_name, __pyx_v_res); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 36, __pyx_L1_error)

  /* "EnumBase":37
 *         res.name = name
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res             # <<<<<<<<<<<<<<
 *         return res
 *     def __repr__(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_v_name, __pyx_v_res) < 0)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":38
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 *         return res             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__ = {"__repr__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__, METH_O, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "EnumBase":40
 *         return res
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)             # <<<<<<<<<<<<<<
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__ = {"__str__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__, METH_O, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "EnumBase":42
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta = {"__pyx_unpickle___Pyx_EnumMeta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle___Pyx_EnumMeta") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle___Pyx_EnumMeta__set_state(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_8fastwarc_4warc_WarcHeaderMap __pyx_vtable_8fastwarc_4warc_WarcHeaderMap;

static PyObject *__pyx_tp_new_8fastwarc_4warc_WarcHeaderMap(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)o);
  p->__pyx_vtab = __pyx_vtabptr_8fastwarc_4warc_WarcHeaderMap;
  new((void*)&(p->_status_line)) std::string();
  new((void*)&(p->_headers)) std::vector<__pyx_t_8fastwarc_4warc_str_pair> ();
  p->_enc = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_dict_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_3__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8fastwarc_4warc_WarcHeaderMap(PyObject *o) {
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *p = (struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->_status_line);
  __Pyx_call_destructor(p->_headers);
  Py_CLEAR(p->_enc);
  Py_CLEAR(p->_dict_cache);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8fastwarc_4warc_WarcHeaderMap(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *p = (struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)o;
  if (p->_dict_cache) {
    e = (*v)(p->_dict_cache, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8fastwarc_4warc_WarcHeaderMap(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *p = (struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)o;
  tmp = ((PyObject*)p->_dict_cache);
  p->_dict_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_8fastwarc_4warc_WarcHeaderMap(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_8fastwarc_4warc_WarcHeaderMap(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_7__setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop_8fastwarc_4warc_13WarcHeaderMap_status_line(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_line_1__get__(o);
}

static int __pyx_setprop_8fastwarc_4warc_13WarcHeaderMap_status_line(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_line_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8fastwarc_4warc_13WarcHeaderMap_status_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_11status_code_1__get__(o);
}

static PyMethodDef __pyx_methods_8fastwarc_4warc_WarcHeaderMap[] = {
  {"append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_16append, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_15append},
  {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_18get, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_17get},
  {"items", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_20items, METH_NOARGS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_19items},
  {"keys", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_22keys, METH_NOARGS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_21keys},
  {"values", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_24values, METH_NOARGS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_23values},
  {"asdict", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_26asdict, METH_NOARGS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_25asdict},
  {"astuples", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_28astuples, METH_NOARGS, __pyx_doc_8fastwarc_4warc_13WarcHeaderMap_27astuples},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_30__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8fastwarc_4warc_13WarcHeaderMap_32__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8fastwarc_4warc_WarcHeaderMap[] = {
  {(char *)"status_line", __pyx_getprop_8fastwarc_4warc_13WarcHeaderMap_status_line, __pyx_setprop_8fastwarc_4warc_13WarcHeaderMap_status_line, (char *)"Header status line.", 0},
  {(char *)"status_code", __pyx_getprop_8fastwarc_4warc_13WarcHeaderMap_status_code, 0, (char *)"HTTP status code (unset if header block is not an HTTP header block).", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_WarcHeaderMap = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_8fastwarc_4warc_WarcHeaderMap, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_14__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_WarcHeaderMap = {
  0, /*mp_length*/
  __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_5__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_8fastwarc_4warc_WarcHeaderMap, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_8fastwarc_4warc_WarcHeaderMap = {
  PyVarObject_HEAD_INIT(0, 0)
  "fastwarc.warc.WarcHeaderMap", /*tp_name*/
  sizeof(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8fastwarc_4warc_WarcHeaderMap, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_12__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_WarcHeaderMap, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_WarcHeaderMap, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Dict-like type representing a WARC or HTTP header block", /*tp_doc*/
  __pyx_tp_traverse_8fastwarc_4warc_WarcHeaderMap, /*tp_traverse*/
  __pyx_tp_clear_8fastwarc_4warc_WarcHeaderMap, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_9__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8fastwarc_4warc_WarcHeaderMap, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8fastwarc_4warc_WarcHeaderMap, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8fastwarc_4warc_13WarcHeaderMap_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8fastwarc_4warc_WarcHeaderMap, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_8fastwarc_4warc_WarcRecord __pyx_vtable_8fastwarc_4warc_WarcRecord;

static PyObject *__pyx_tp_new_8fastwarc_4warc_WarcRecord(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8fastwarc_4warc_WarcRecord *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)o);
  p->__pyx_vtab = __pyx_vtabptr_8fastwarc_4warc_WarcRecord;
  new((void*)&(p->_http_charset)) std::string();
  p->_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)Py_None); Py_INCREF(Py_None);
  p->_http_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)Py_None); Py_INCREF(Py_None);
  p->_reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8fastwarc_4warc_10WarcRecord_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8fastwarc_4warc_WarcRecord(PyObject *o) {
  struct __pyx_obj_8fastwarc_4warc_WarcRecord *p = (struct __pyx_obj_8fastwarc_4warc_WarcRecord *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->_http_charset);
  Py_CLEAR(p->_headers);
  Py_CLEAR(p->_http_headers);
  Py_CLEAR(p->_reader);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8fastwarc_4warc_WarcRecord(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8fastwarc_4warc_WarcRecord *p = (struct __pyx_obj_8fastwarc_4warc_WarcRecord *)o;
  if (p->_headers) {
    e = (*v)(((PyObject *)p->_headers), a); if (e) return e;
  }
  if (p->_http_headers) {
    e = (*v)(((PyObject *)p->_http_headers), a); if (e) return e;
  }
  if (p->_reader) {
    e = (*v)(((PyObject *)p->_reader), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8fastwarc_4warc_WarcRecord(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8fastwarc_4warc_WarcRecord *p = (struct __pyx_obj_8fastwarc_4warc_WarcRecord *)o;
  tmp = ((PyObject*)p->_headers);
  p->_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_http_headers);
  p->_http_headers = ((struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_reader);
  p->_reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_record_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_9record_id_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_record_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_11record_type_1__get__(o);
}

static int __pyx_setprop_8fastwarc_4warc_10WarcRecord_record_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_8fastwarc_4warc_10WarcRecord_11record_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_headers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_7headers_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_is_http(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_7is_http_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_is_http_parsed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_14is_http_parsed_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_http_headers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_12http_headers_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_http_charset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_12http_charset_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_content_length(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_14content_length_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_reader(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_6reader_1__get__(o);
}

static PyObject *__pyx_getprop_8fastwarc_4warc_10WarcRecord_stream_pos(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8fastwarc_4warc_10WarcRecord_10stream_pos_1__get__(o);
}

static PyMethodDef __pyx_methods_8fastwarc_4warc_WarcRecord[] = {
  {"init_headers", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_10WarcRecord_3init_headers, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8fastwarc_4warc_10WarcRecord_2init_headers},
  {"set_bytes_content", (PyCFunction)__pyx_pw_8fastwarc_4warc_10WarcRecord_5set_bytes_content, METH_O, __pyx_doc_8fastwarc_4warc_10WarcRecord_4set_bytes_content},
  {"parse_http", (PyCFunction)__pyx_pw_8fastwarc_4warc_10WarcRecord_7parse_http, METH_NOARGS, __pyx_doc_8fastwarc_4warc_10WarcRecord_6parse_http},
  {"write", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_10WarcRecord_9write, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8fastwarc_4warc_10WarcRecord_8write},
  {"verify_block_digest", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_10WarcRecord_11verify_block_digest, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8fastwarc_4warc_10WarcRecord_10verify_block_digest},
  {"verify_payload_digest", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8fastwarc_4warc_10WarcRecord_13verify_payload_digest, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8fastwarc_4warc_10WarcRecord_12verify_payload_digest},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8fastwarc_4warc_10WarcRecord_15__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8fastwarc_4warc_10WarcRecord_17__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8fastwarc_4warc_WarcRecord[] = {
  {(char *)"record_id", __pyx_getprop_8fastwarc_4warc_10WarcRecord_record_id, 0, (char *)"Record ID (same as ``headers['WARC-Record'ID']``.", 0},
  {(char *)"record_type", __pyx_getprop_8fastwarc_4warc_10WarcRecord_record_type, __pyx_setprop_8fastwarc_4warc_10WarcRecord_record_type, (char *)"Record type (same as ``headers['WARC-Type']``.", 0},
  {(char *)"headers", __pyx_getprop_8fastwarc_4warc_10WarcRecord_headers, 0, (char *)"WARC record headers.", 0},
  {(char *)"is_http", __pyx_getprop_8fastwarc_4warc_10WarcRecord_is_http, 0, (char *)"Whether record is an HTTP record", 0},
  {(char *)"is_http_parsed", __pyx_getprop_8fastwarc_4warc_10WarcRecord_is_http_parsed, 0, (char *)"Whether HTTP headers have been parsed.", 0},
  {(char *)"http_headers", __pyx_getprop_8fastwarc_4warc_10WarcRecord_http_headers, 0, (char *)"HTTP headers if record is an HTTP record and HTTP headers have been parsed yet.", 0},
  {(char *)"http_charset", __pyx_getprop_8fastwarc_4warc_10WarcRecord_http_charset, 0, (char *)"\n        HTTP charset/encoding as returned by the server or ``None`` if no valid charset is set.\n        A returned string is guaranteed to be a valid Python encoding name.\n        ", 0},
  {(char *)"content_length", __pyx_getprop_8fastwarc_4warc_10WarcRecord_content_length, 0, (char *)"Remaining WARC length in bytes (not necessarily the same as the ``Content-Length`` header).", 0},
  {(char *)"reader", __pyx_getprop_8fastwarc_4warc_10WarcRecord_reader, 0, (char *)"Reader for the remaining WARC record content.", 0},
  {(char *)"stream_pos", __pyx_getprop_8fastwarc_4warc_10WarcRecord_stream_pos, 0, (char *)"WARC record start offset in the original (uncompressed) stream.", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8fastwarc_4warc_WarcRecord = {
  PyVarObject_HEAD_INIT(0, 0)
  "fastwarc.warc.WarcRecord", /*tp_name*/
  sizeof(struct __pyx_obj_8fastwarc_4warc_WarcRecord), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8fastwarc_4warc_WarcRecord, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "A WARC record.", /*tp_doc*/
  __pyx_tp_traverse_8fastwarc_4warc_WarcRecord, /*tp_traverse*/
  __pyx_tp_clear_8fastwarc_4warc_WarcRecord, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8fastwarc_4warc_WarcRecord, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8fastwarc_4warc_WarcRecord, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8fastwarc_4warc_WarcRecord, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_8fastwarc_4warc_ArchiveIterator __pyx_vtable_8fastwarc_4warc_ArchiveIterator;

static PyObject *__pyx_tp_new_8fastwarc_4warc_ArchiveIterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)o);
  p->__pyx_vtab = __pyx_vtabptr_8fastwarc_4warc_ArchiveIterator;
  p->reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)Py_None); Py_INCREF(Py_None);
  p->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)Py_None); Py_INCREF(Py_None);
  p->func_filter = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8fastwarc_4warc_15ArchiveIterator_3__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8fastwarc_4warc_ArchiveIterator(PyObject *o) {
  struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *p = (struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->reader);
  Py_CLEAR(p->record);
  Py_CLEAR(p->func_filter);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8fastwarc_4warc_ArchiveIterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *p = (struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)o;
  if (p->reader) {
    e = (*v)(((PyObject *)p->reader), a); if (e) return e;
  }
  if (p->record) {
    e = (*v)(((PyObject *)p->record), a); if (e) return e;
  }
  if (p->func_filter) {
    e = (*v)(p->func_filter, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8fastwarc_4warc_ArchiveIterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *p = (struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *)o;
  tmp = ((PyObject*)p->reader);
  p->reader = ((struct __pyx_obj_8fastwarc_9stream_io_BufferedReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->record);
  p->record = ((struct __pyx_obj_8fastwarc_4warc_WarcRecord *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->func_filter);
  p->func_filter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8fastwarc_4warc_ArchiveIterator[] = {
  {"_set_stream", (PyCFunction)__pyx_pw_8fastwarc_4warc_15ArchiveIterator_8_set_stream, METH_O, __pyx_doc_8fastwarc_4warc_15ArchiveIterator_7_set_stream},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8fastwarc_4warc_ArchiveIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "fastwarc.warc.ArchiveIterator", /*tp_name*/
  sizeof(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8fastwarc_4warc_ArchiveIterator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "WARC record stream iterator.", /*tp_doc*/
  __pyx_tp_traverse_8fastwarc_4warc_ArchiveIterator, /*tp_traverse*/
  __pyx_tp_clear_8fastwarc_4warc_ArchiveIterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8fastwarc_4warc_15ArchiveIterator_5__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8fastwarc_4warc_ArchiveIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8fastwarc_4warc_15ArchiveIterator_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8fastwarc_4warc_ArchiveIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *__pyx_freelist_8fastwarc_4warc___pyx_scope_struct____iter__[8];
static int __pyx_freecount_8fastwarc_4warc___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_8fastwarc_4warc___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8fastwarc_4warc___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_8fastwarc_4warc___pyx_scope_struct____iter__[--__pyx_freecount_8fastwarc_4warc___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8fastwarc_4warc___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *p = (struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8fastwarc_4warc___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__)))) {
    __pyx_freelist_8fastwarc_4warc___pyx_scope_struct____iter__[__pyx_freecount_8fastwarc_4warc___pyx_scope_struct____iter__++] = ((struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8fastwarc_4warc___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *p = (struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "fastwarc.warc.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8fastwarc_4warc___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8fastwarc_4warc___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8fastwarc_4warc___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *__pyx_freelist_8fastwarc_4warc___pyx_scope_struct_1___iter__[8];
static int __pyx_freecount_8fastwarc_4warc___pyx_scope_struct_1___iter__ = 0;

static PyObject *__pyx_tp_new_8fastwarc_4warc___pyx_scope_struct_1___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8fastwarc_4warc___pyx_scope_struct_1___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__)))) {
    o = (PyObject*)__pyx_freelist_8fastwarc_4warc___pyx_scope_struct_1___iter__[--__pyx_freecount_8fastwarc_4warc___pyx_scope_struct_1___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8fastwarc_4warc___pyx_scope_struct_1___iter__(PyObject *o) {
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8fastwarc_4warc___pyx_scope_struct_1___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__)))) {
    __pyx_freelist_8fastwarc_4warc___pyx_scope_struct_1___iter__[__pyx_freecount_8fastwarc_4warc___pyx_scope_struct_1___iter__++] = ((struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8fastwarc_4warc___pyx_scope_struct_1___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "fastwarc.warc.__pyx_scope_struct_1___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_8fastwarc_4warc___pyx_scope_struct_1___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8fastwarc_4warc___pyx_scope_struct_1___iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8fastwarc_4warc___pyx_scope_struct_1___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8fastwarc_4warc___pyx_scope_struct_1___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = (&PyType_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc___Pyx_EnumMeta(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  (&PyType_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse___Pyx_EnumMeta(PyObject *o, visitproc v, void *a) {
  int e;
  if (!(&PyType_Type)->tp_traverse); else { e = (&PyType_Type)->tp_traverse(o,v,a); if (e) return e; }
  return 0;
}

static int __pyx_tp_clear___Pyx_EnumMeta(PyObject *o) {
  if (!(&PyType_Type)->tp_clear); else (&PyType_Type)->tp_clear(o);
  return 0;
}
static PyObject *__pyx_sq_item___Pyx_EnumMeta(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods___Pyx_EnumMeta[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence___Pyx_EnumMeta = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item___Pyx_EnumMeta, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping___Pyx_EnumMeta = {
  0, /*mp_length*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __Pyx_EnumMeta = {
  PyVarObject_HEAD_INIT(0, 0)
  "fastwarc.warc.__Pyx_EnumMeta", /*tp_name*/
  sizeof(struct __pyx_obj___Pyx_EnumMeta), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___Pyx_EnumMeta, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence___Pyx_EnumMeta, /*tp_as_sequence*/
  &__pyx_tp_as_mapping___Pyx_EnumMeta, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse___Pyx_EnumMeta, /*tp_traverse*/
  __pyx_tp_clear___Pyx_EnumMeta, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods___Pyx_EnumMeta, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___Pyx_EnumMeta, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"is_warc_10", (PyCFunction)__pyx_pw_8fastwarc_4warc_1is_warc_10, METH_O, __pyx_doc_8fastwarc_4warc_is_warc_10},
  {"is_warc_11", (PyCFunction)__pyx_pw_8fastwarc_4warc_3is_warc_11, METH_O, __pyx_doc_8fastwarc_4warc_2is_warc_11},
  {"has_block_digest", (PyCFunction)__pyx_pw_8fastwarc_4warc_5has_block_digest, METH_O, __pyx_doc_8fastwarc_4warc_4has_block_digest},
  {"has_payload_digest", (PyCFunction)__pyx_pw_8fastwarc_4warc_7has_payload_digest, METH_O, __pyx_doc_8fastwarc_4warc_6has_payload_digest},
  {"is_http", (PyCFunction)__pyx_pw_8fastwarc_4warc_9is_http, METH_O, __pyx_doc_8fastwarc_4warc_8is_http},
  {"is_concurrent", (PyCFunction)__pyx_pw_8fastwarc_4warc_11is_concurrent, METH_O, __pyx_doc_8fastwarc_4warc_10is_concurrent},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_warc(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_warc},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "warc",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_b_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 0, 0},
  {&__pyx_n_s_ArchiveIterator, __pyx_k_ArchiveIterator, sizeof(__pyx_k_ArchiveIterator), 0, 0, 1, 1},
  {&__pyx_n_s_ArchiveIterator___iter, __pyx_k_ArchiveIterator___iter, sizeof(__pyx_k_ArchiveIterator___iter), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStr, __pyx_k_CaseInsensitiveStr, sizeof(__pyx_k_CaseInsensitiveStr), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict, __pyx_k_CaseInsensitiveStrDict, sizeof(__pyx_k_CaseInsensitiveStrDict), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict___contain, __pyx_k_CaseInsensitiveStrDict___contain, sizeof(__pyx_k_CaseInsensitiveStrDict___contain), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict___getitem, __pyx_k_CaseInsensitiveStrDict___getitem, sizeof(__pyx_k_CaseInsensitiveStrDict___getitem), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict___setitem, __pyx_k_CaseInsensitiveStrDict___setitem, sizeof(__pyx_k_CaseInsensitiveStrDict___setitem), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict_get, __pyx_k_CaseInsensitiveStrDict_get, sizeof(__pyx_k_CaseInsensitiveStrDict_get), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict_pop, __pyx_k_CaseInsensitiveStrDict_pop, sizeof(__pyx_k_CaseInsensitiveStrDict_pop), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict_setdefaul, __pyx_k_CaseInsensitiveStrDict_setdefaul, sizeof(__pyx_k_CaseInsensitiveStrDict_setdefaul), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStrDict_update, __pyx_k_CaseInsensitiveStrDict_update, sizeof(__pyx_k_CaseInsensitiveStrDict_update), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStr___eq, __pyx_k_CaseInsensitiveStr___eq, sizeof(__pyx_k_CaseInsensitiveStr___eq), 0, 0, 1, 1},
  {&__pyx_n_s_CaseInsensitiveStr___hash, __pyx_k_CaseInsensitiveStr___hash, sizeof(__pyx_k_CaseInsensitiveStr___hash), 0, 0, 1, 1},
  {&__pyx_kp_s_Case_insensitive_str_dict, __pyx_k_Case_insensitive_str_dict, sizeof(__pyx_k_Case_insensitive_str_dict), 0, 0, 1, 0},
  {&__pyx_kp_s_Case_insensitive_str_implementat, __pyx_k_Case_insensitive_str_implementat, sizeof(__pyx_k_Case_insensitive_str_implementat), 0, 0, 1, 0},
  {&__pyx_n_s_EnumBase, __pyx_k_EnumBase, sizeof(__pyx_k_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_EnumType, __pyx_k_EnumType, sizeof(__pyx_k_EnumType), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
  {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
  {&__pyx_n_s_LookupError, __pyx_k_LookupError, sizeof(__pyx_k_LookupError), 0, 0, 1, 1},
  {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
  {&__pyx_kp_u_Object_of_type, __pyx_k_Object_of_type, sizeof(__pyx_k_Object_of_type), 0, 1, 0, 0},
  {&__pyx_n_s_Optional, __pyx_k_Optional, sizeof(__pyx_k_Optional), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase, __pyx_k_Pyx_EnumBase, sizeof(__pyx_k_Pyx_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___new, __pyx_k_Pyx_EnumBase___new, sizeof(__pyx_k_Pyx_EnumBase___new), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___repr, __pyx_k_Pyx_EnumBase___repr, sizeof(__pyx_k_Pyx_EnumBase___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___str, __pyx_k_Pyx_EnumBase___str, sizeof(__pyx_k_Pyx_EnumBase___str), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeWarning, __pyx_k_RuntimeWarning, sizeof(__pyx_k_RuntimeWarning), 0, 0, 1, 1},
  {&__pyx_n_s_StreamError, __pyx_k_StreamError, sizeof(__pyx_k_StreamError), 0, 0, 1, 1},
  {&__pyx_n_s_Tuple, __pyx_k_Tuple, sizeof(__pyx_k_Tuple), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_enum_value_s, __pyx_k_Unknown_enum_value_s, sizeof(__pyx_k_Unknown_enum_value_s), 0, 0, 1, 0},
  {&__pyx_kp_u_Unsupported_hash_algorithm, __pyx_k_Unsupported_hash_algorithm, sizeof(__pyx_k_Unsupported_hash_algorithm), 0, 1, 0, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u_WARC_Record_ID, __pyx_k_WARC_Record_ID, sizeof(__pyx_k_WARC_Record_ID), 0, 1, 0, 0},
  {&__pyx_kp_u_WARC_Type, __pyx_k_WARC_Type, sizeof(__pyx_k_WARC_Type), 0, 1, 0, 0},
  {&__pyx_n_s_WarcHeaderMap, __pyx_k_WarcHeaderMap, sizeof(__pyx_k_WarcHeaderMap), 0, 0, 1, 1},
  {&__pyx_n_s_WarcHeaderMap___iter, __pyx_k_WarcHeaderMap___iter, sizeof(__pyx_k_WarcHeaderMap___iter), 0, 0, 1, 1},
  {&__pyx_n_s_WarcRecord, __pyx_k_WarcRecord, sizeof(__pyx_k_WarcRecord), 0, 0, 1, 1},
  {&__pyx_n_s_WarcRecordType, __pyx_k_WarcRecordType, sizeof(__pyx_k_WarcRecordType), 0, 0, 1, 1},
  {&__pyx_kp_u_Y_m_dT_H_M_SZ, __pyx_k_Y_m_dT_H_M_SZ, sizeof(__pyx_k_Y_m_dT_H_M_SZ), 0, 1, 0, 0},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_u__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 1, 0, 0},
  {&__pyx_kp_b__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_b__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 0, 0, 0},
  {&__pyx_kp_b__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 0, 0},
  {&__pyx_n_s_any_type, __pyx_k_any_type, sizeof(__pyx_k_any_type), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_asdict, __pyx_k_asdict, sizeof(__pyx_k_asdict), 0, 0, 1, 1},
  {&__pyx_n_s_b32decode, __pyx_k_b32decode, sizeof(__pyx_k_b32decode), 0, 0, 1, 1},
  {&__pyx_n_s_b32encode, __pyx_k_b32encode, sizeof(__pyx_k_b32encode), 0, 0, 1, 1},
  {&__pyx_n_s_base64, __pyx_k_base64, sizeof(__pyx_k_base64), 0, 0, 1, 1},
  {&__pyx_n_s_casefold, __pyx_k_casefold, sizeof(__pyx_k_casefold), 0, 0, 1, 1},
  {&__pyx_n_s_checksum_data, __pyx_k_checksum_data, sizeof(__pyx_k_checksum_data), 0, 0, 1, 1},
  {&__pyx_n_s_chunk_size, __pyx_k_chunk_size, sizeof(__pyx_k_chunk_size), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_u_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 1, 0, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_consume, __pyx_k_consume, sizeof(__pyx_k_consume), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_content_length, __pyx_k_content_length, sizeof(__pyx_k_content_length), 0, 0, 1, 1},
  {&__pyx_n_s_continuation, __pyx_k_continuation, sizeof(__pyx_k_continuation), 0, 0, 1, 1},
  {&__pyx_n_s_conversion, __pyx_k_conversion, sizeof(__pyx_k_conversion), 0, 0, 1, 1},
  {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
  {&__pyx_n_s_dct, __pyx_k_dct, sizeof(__pyx_k_dct), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_digest, __pyx_k_digest, sizeof(__pyx_k_digest), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_errors, __pyx_k_errors, sizeof(__pyx_k_errors), 0, 0, 1, 1},
  {&__pyx_n_s_fastwarc_stream_io, __pyx_k_fastwarc_stream_io, sizeof(__pyx_k_fastwarc_stream_io), 0, 0, 1, 1},
  {&__pyx_n_s_fastwarc_warc, __pyx_k_fastwarc_warc, sizeof(__pyx_k_fastwarc_warc), 0, 0, 1, 1},
  {&__pyx_kp_s_fastwarc_warc_pyx, __pyx_k_fastwarc_warc_pyx, sizeof(__pyx_k_fastwarc_warc_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
  {&__pyx_n_s_func_filter, __pyx_k_func_filter, sizeof(__pyx_k_func_filter), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getitem, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_u_has_no_attribute, __pyx_k_has_no_attribute, sizeof(__pyx_k_has_no_attribute), 0, 1, 0, 0},
  {&__pyx_n_s_hash, __pyx_k_hash, sizeof(__pyx_k_hash), 0, 0, 1, 1},
  {&__pyx_n_s_hashlib, __pyx_k_hashlib, sizeof(__pyx_k_hashlib), 0, 0, 1, 1},
  {&__pyx_n_s_http_parsed, __pyx_k_http_parsed, sizeof(__pyx_k_http_parsed), 0, 0, 1, 1},
  {&__pyx_n_u_ignore, __pyx_k_ignore, sizeof(__pyx_k_ignore), 0, 1, 0, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_init_headers, __pyx_k_init_headers, sizeof(__pyx_k_init_headers), 0, 0, 1, 1},
  {&__pyx_kp_u_is_not_a_valid_stream, __pyx_k_is_not_a_valid_stream, sizeof(__pyx_k_is_not_a_valid_stream), 0, 1, 0, 0},
  {&__pyx_n_s_isdigit, __pyx_k_isdigit, sizeof(__pyx_k_isdigit), 0, 0, 1, 1},
  {&__pyx_kp_u_iso_8859_15, __pyx_k_iso_8859_15, sizeof(__pyx_k_iso_8859_15), 0, 1, 0, 0},
  {&__pyx_n_s_it, __pyx_k_it, sizeof(__pyx_k_it), 0, 0, 1, 1},
  {&__pyx_n_s_item, __pyx_k_item, sizeof(__pyx_k_item), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_u_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 1, 0, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max_content_length, __pyx_k_max_content_length, sizeof(__pyx_k_max_content_length), 0, 0, 1, 1},
  {&__pyx_n_s_md5, __pyx_k_md5, sizeof(__pyx_k_md5), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_metadata, __pyx_k_metadata, sizeof(__pyx_k_metadata), 0, 0, 1, 1},
  {&__pyx_n_s_min_content_length, __pyx_k_min_content_length, sizeof(__pyx_k_min_content_length), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_no_type, __pyx_k_no_type, sizeof(__pyx_k_no_type), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
  {&__pyx_n_s_parents, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {&__pyx_n_s_parse_http, __pyx_k_parse_http, sizeof(__pyx_k_parse_http), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_k_pyx_unpickle___Pyx_EnumMeta, sizeof(__pyx_k_pyx_unpickle___Pyx_EnumMeta), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_u_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 1, 0, 1},
  {&__pyx_n_s_reader, __pyx_k_reader, sizeof(__pyx_k_reader), 0, 0, 1, 1},
  {&__pyx_n_s_record_type, __pyx_k_record_type, sizeof(__pyx_k_record_type), 0, 0, 1, 1},
  {&__pyx_n_s_record_types, __pyx_k_record_types, sizeof(__pyx_k_record_types), 0, 0, 1, 1},
  {&__pyx_n_s_record_urn, __pyx_k_record_urn, sizeof(__pyx_k_record_urn), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_request, __pyx_k_request, sizeof(__pyx_k_request), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_resource, __pyx_k_resource, sizeof(__pyx_k_resource), 0, 0, 1, 1},
  {&__pyx_n_s_response, __pyx_k_response, sizeof(__pyx_k_response), 0, 0, 1, 1},
  {&__pyx_n_s_reversed, __pyx_k_reversed, sizeof(__pyx_k_reversed), 0, 0, 1, 1},
  {&__pyx_n_s_revisit, __pyx_k_revisit, sizeof(__pyx_k_revisit), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d, __pyx_k_s_s_d, sizeof(__pyx_k_s_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set_bytes_content, __pyx_k_set_bytes_content, sizeof(__pyx_k_set_bytes_content), 0, 0, 1, 1},
  {&__pyx_n_s_set_stream, __pyx_k_set_stream, sizeof(__pyx_k_set_stream), 0, 0, 1, 1},
  {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_setitem, __pyx_k_setitem, sizeof(__pyx_k_setitem), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sha1, __pyx_k_sha1, sizeof(__pyx_k_sha1), 0, 0, 1, 1},
  {&__pyx_kp_b_sha1_2, __pyx_k_sha1_2, sizeof(__pyx_k_sha1_2), 0, 0, 0, 0},
  {&__pyx_n_s_sha256, __pyx_k_sha256, sizeof(__pyx_k_sha256), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_strftime, __pyx_k_strftime, sizeof(__pyx_k_strftime), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_tell, __pyx_k_tell, sizeof(__pyx_k_tell), 0, 0, 1, 1},
  {&__pyx_n_u_tell, __pyx_k_tell, sizeof(__pyx_k_tell), 0, 1, 0, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
  {&__pyx_n_s_unknown, __pyx_k_unknown, sizeof(__pyx_k_unknown), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_urn, __pyx_k_urn, sizeof(__pyx_k_urn), 0, 0, 1, 1},
  {&__pyx_n_s_utcnow, __pyx_k_utcnow, sizeof(__pyx_k_utcnow), 0, 0, 1, 1},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 0, 1, 1},
  {&__pyx_n_s_uuid4, __pyx_k_uuid4, sizeof(__pyx_k_uuid4), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_verify_block_digest, __pyx_k_verify_block_digest, sizeof(__pyx_k_verify_block_digest), 0, 0, 1, 1},
  {&__pyx_n_s_verify_digests, __pyx_k_verify_digests, sizeof(__pyx_k_verify_digests), 0, 0, 1, 1},
  {&__pyx_n_s_verify_payload_digest, __pyx_k_verify_payload_digest, sizeof(__pyx_k_verify_payload_digest), 0, 0, 1, 1},
  {&__pyx_n_s_warcinfo, __pyx_k_warcinfo, sizeof(__pyx_k_warcinfo), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_u_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 1, 0, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_builtin_reversed = __Pyx_GetBuiltinName(__pyx_n_s_reversed); if (!__pyx_builtin_reversed) __PYX_ERR(0, 45, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 115, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_LookupError = __Pyx_GetBuiltinName(__pyx_n_s_LookupError); if (!__pyx_builtin_LookupError) __PYX_ERR(0, 368, __pyx_L1_error)
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_builtin_RuntimeWarning = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeWarning); if (!__pyx_builtin_RuntimeWarning) __PYX_ERR(0, 504, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 789, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 33, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "fastwarc/warc.pyx":304
 *         self._http_parsed = False
 *         self._content_length = 0
 *         self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')             # <<<<<<<<<<<<<<
 *         self._http_headers = None
 *         self._stream_pos = 0
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_utf_8); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "fastwarc/warc.pyx":434
 *         if self._http_parsed or not self._is_http:
 *             return
 *         self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')             # <<<<<<<<<<<<<<
 *         cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True)
 *         self._content_length = self._content_length - num_bytes
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_iso_8859_15); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "fastwarc/warc.pyx":787
 *         """
 *         if not isinstance(stream, IOStream):
 *             for attr in ('read', 'tell', 'close'):             # <<<<<<<<<<<<<<
 *                 if not hasattr(stream, attr):
 *                     raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
 */
  __pyx_tuple__16 = PyTuple_Pack(3, __pyx_n_u_read, __pyx_n_u_tell, __pyx_n_u_close); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "fastwarc/warc.pyx":104
 *     """Case insensitive str implementation for use as dict key."""
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.casefold())
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_hash, 104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 104, __pyx_L1_error)

  /* "fastwarc/warc.pyx":107
 *         return hash(self.casefold())
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self.casefold() == other.casefold()
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_eq, 107, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 107, __pyx_L1_error)

  /* "fastwarc/warc.pyx":114
 *     """Case-insensitive str dict."""
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_key); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_getitem, 114, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 114, __pyx_L1_error)

  /* "fastwarc/warc.pyx":117
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 *     def __setitem__(self, key, value):             # <<<<<<<<<<<<<<
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_setitem, 117, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 117, __pyx_L1_error)

  /* "fastwarc/warc.pyx":120
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_key); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_contains, 120, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 120, __pyx_L1_error)

  /* "fastwarc/warc.pyx":123
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 *     def get(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_get, 123, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 123, __pyx_L1_error)
  __pyx_tuple__30 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "fastwarc/warc.pyx":126
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 *     def setdefault(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_setdefault, 126, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_tuple__33 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "fastwarc/warc.pyx":129
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 *     def pop(self, key):             # <<<<<<<<<<<<<<
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_key); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_pop, 129, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 129, __pyx_L1_error)

  /* "fastwarc/warc.pyx":132
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 *     def update(self, it=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if hasattr(it, 'items'):
 *             it = it.items()
 */
  __pyx_tuple__36 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_it, __pyx_n_s_kwargs, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fastwarc_warc_pyx, __pyx_n_s_update, 132, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 132, __pyx_L1_error)
  __pyx_tuple__38 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_tuple__39 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_value, __pyx_n_s_name_2, __pyx_n_s_v, __pyx_n_s_res); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_new, 28, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(1, 28, __pyx_L1_error)
  __pyx_tuple__41 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_repr, 39, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 39, __pyx_L1_error)

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_str, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 41, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__46 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_OrderedDict = Py_None; Py_INCREF(Py_None);
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_globals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("is_warc_10", (void (*)(void))__pyx_f_8fastwarc_4warc_is_warc_10, "int (struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("is_warc_11", (void (*)(void))__pyx_f_8fastwarc_4warc_is_warc_11, "int (struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("has_block_digest", (void (*)(void))__pyx_f_8fastwarc_4warc_has_block_digest, "int (struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("has_payload_digest", (void (*)(void))__pyx_f_8fastwarc_4warc_has_payload_digest, "int (struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("is_http", (void (*)(void))__pyx_f_8fastwarc_4warc_is_http, "int (struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("is_concurrent", (void (*)(void))__pyx_f_8fastwarc_4warc_is_concurrent, "int (struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_8fastwarc_4warc_WarcHeaderMap = &__pyx_vtable_8fastwarc_4warc_WarcHeaderMap;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.write = (size_t (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, struct __pyx_obj_8fastwarc_9stream_io_IOStream *))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_write;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.clear = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_clear;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.set_status_line = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_status_line;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.find_header = (std::string (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_find_header;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.set_header = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_set_header;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.append_header = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &, std::string const &))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_append_header;
  __pyx_vtable_8fastwarc_4warc_WarcHeaderMap.add_continuation = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcHeaderMap *, std::string const &))__pyx_f_8fastwarc_4warc_13WarcHeaderMap_add_continuation;
  if (PyType_Ready(&__pyx_type_8fastwarc_4warc_WarcHeaderMap) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8fastwarc_4warc_WarcHeaderMap.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8fastwarc_4warc_WarcHeaderMap.tp_dictoffset && __pyx_type_8fastwarc_4warc_WarcHeaderMap.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8fastwarc_4warc_WarcHeaderMap.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_8fastwarc_4warc_WarcHeaderMap, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 143, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_8fastwarc_4warc_13WarcHeaderMap___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_8fastwarc_4warc_13WarcHeaderMap___init__.doc = __pyx_doc_8fastwarc_4warc_13WarcHeaderMap___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_8fastwarc_4warc_13WarcHeaderMap___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_8fastwarc_4warc_WarcHeaderMap.tp_dict, __pyx_vtabptr_8fastwarc_4warc_WarcHeaderMap) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WarcHeaderMap, (PyObject *)&__pyx_type_8fastwarc_4warc_WarcHeaderMap) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8fastwarc_4warc_WarcHeaderMap) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  __pyx_ptype_8fastwarc_4warc_WarcHeaderMap = &__pyx_type_8fastwarc_4warc_WarcHeaderMap;
  __pyx_vtabptr_8fastwarc_4warc_WarcRecord = &__pyx_vtable_8fastwarc_4warc_WarcRecord;
  __pyx_vtable_8fastwarc_4warc_WarcRecord.init_headers = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, size_t, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_init_headers *__pyx_optional_args))__pyx_f_8fastwarc_4warc_10WarcRecord_init_headers;
  __pyx_vtable_8fastwarc_4warc_WarcRecord.set_bytes_content = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, PyObject *, int __pyx_skip_dispatch))__pyx_f_8fastwarc_4warc_10WarcRecord_set_bytes_content;
  __pyx_vtable_8fastwarc_4warc_WarcRecord.parse_http = (void (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch))__pyx_f_8fastwarc_4warc_10WarcRecord_parse_http;
  __pyx_vtable_8fastwarc_4warc_WarcRecord.write = (size_t (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_write *__pyx_optional_args))__pyx_f_8fastwarc_4warc_10WarcRecord_write;
  __pyx_vtable_8fastwarc_4warc_WarcRecord.verify_block_digest = (int (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_block_digest *__pyx_optional_args))__pyx_f_8fastwarc_4warc_10WarcRecord_verify_block_digest;
  __pyx_vtable_8fastwarc_4warc_WarcRecord.verify_payload_digest = (int (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_8fastwarc_4warc_10WarcRecord_verify_payload_digest *__pyx_optional_args))__pyx_f_8fastwarc_4warc_10WarcRecord_verify_payload_digest;
  __pyx_vtable_8fastwarc_4warc_WarcRecord._verify_digest = (int (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, std::string const &, int))__pyx_f_8fastwarc_4warc_10WarcRecord__verify_digest;
  __pyx_vtable_8fastwarc_4warc_WarcRecord._write_impl = (size_t (*)(struct __pyx_obj_8fastwarc_4warc_WarcRecord *, PyObject *, PyObject *, int, size_t))__pyx_f_8fastwarc_4warc_10WarcRecord__write_impl;
  if (PyType_Ready(&__pyx_type_8fastwarc_4warc_WarcRecord) < 0) __PYX_ERR(0, 296, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8fastwarc_4warc_WarcRecord.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8fastwarc_4warc_WarcRecord.tp_dictoffset && __pyx_type_8fastwarc_4warc_WarcRecord.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8fastwarc_4warc_WarcRecord.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8fastwarc_4warc_WarcRecord.tp_dict, __pyx_vtabptr_8fastwarc_4warc_WarcRecord) < 0) __PYX_ERR(0, 296, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WarcRecord, (PyObject *)&__pyx_type_8fastwarc_4warc_WarcRecord) < 0) __PYX_ERR(0, 296, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8fastwarc_4warc_WarcRecord) < 0) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_ptype_8fastwarc_4warc_WarcRecord = &__pyx_type_8fastwarc_4warc_WarcRecord;
  __pyx_vtabptr_8fastwarc_4warc_ArchiveIterator = &__pyx_vtable_8fastwarc_4warc_ArchiveIterator;
  __pyx_vtable_8fastwarc_4warc_ArchiveIterator._read_next_record = (enum __pyx_t_8fastwarc_4warc__NextRecStatus (*)(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *))__pyx_f_8fastwarc_4warc_15ArchiveIterator__read_next_record;
  __pyx_vtable_8fastwarc_4warc_ArchiveIterator._set_stream = (void (*)(struct __pyx_obj_8fastwarc_4warc_ArchiveIterator *, PyObject *, int __pyx_skip_dispatch))__pyx_f_8fastwarc_4warc_15ArchiveIterator__set_stream;
  if (PyType_Ready(&__pyx_type_8fastwarc_4warc_ArchiveIterator) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8fastwarc_4warc_ArchiveIterator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8fastwarc_4warc_ArchiveIterator.tp_dictoffset && __pyx_type_8fastwarc_4warc_ArchiveIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8fastwarc_4warc_ArchiveIterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_8fastwarc_4warc_ArchiveIterator, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 647, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_8fastwarc_4warc_15ArchiveIterator___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_8fastwarc_4warc_15ArchiveIterator___init__.doc = __pyx_doc_8fastwarc_4warc_15ArchiveIterator___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_8fastwarc_4warc_15ArchiveIterator___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_8fastwarc_4warc_ArchiveIterator.tp_dict, __pyx_vtabptr_8fastwarc_4warc_ArchiveIterator) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ArchiveIterator, (PyObject *)&__pyx_type_8fastwarc_4warc_ArchiveIterator) < 0) __PYX_ERR(0, 647, __pyx_L1_error)
  __pyx_ptype_8fastwarc_4warc_ArchiveIterator = &__pyx_type_8fastwarc_4warc_ArchiveIterator;
  if (PyType_Ready(&__pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 164, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8fastwarc_4warc___pyx_scope_struct____iter__ = &__pyx_type_8fastwarc_4warc___pyx_scope_struct____iter__;
  if (PyType_Ready(&__pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__) < 0) __PYX_ERR(0, 683, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__.tp_dictoffset && __pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8fastwarc_4warc___pyx_scope_struct_1___iter__ = &__pyx_type_8fastwarc_4warc___pyx_scope_struct_1___iter__;
  __Pyx_EnumMeta.tp_base = (&PyType_Type);
  if (PyType_Ready(&__Pyx_EnumMeta) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __Pyx_EnumMeta.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__Pyx_EnumMeta.tp_dictoffset && __Pyx_EnumMeta.tp_getattro == PyObject_GenericGetAttr)) {
    __Pyx_EnumMeta.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__Pyx_EnumMeta) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_ptype___Pyx_EnumMeta = &__Pyx_EnumMeta;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("fastwarc.stream_io"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8fastwarc_9stream_io_IOStream = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "IOStream", sizeof(struct __pyx_obj_8fastwarc_9stream_io_IOStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_IOStream) __PYX_ERR(2, 24, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_IOStream = (struct __pyx_vtabstruct_8fastwarc_9stream_io_IOStream*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_IOStream->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_IOStream)) __PYX_ERR(2, 24, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "PythonIOStreamAdapter", sizeof(struct __pyx_obj_8fastwarc_9stream_io_PythonIOStreamAdapter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter) __PYX_ERR(2, 36, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_PythonIOStreamAdapter = (struct __pyx_vtabstruct_8fastwarc_9stream_io_PythonIOStreamAdapter*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_PythonIOStreamAdapter->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_PythonIOStreamAdapter)) __PYX_ERR(2, 36, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_BytesIOStream = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "BytesIOStream", sizeof(struct __pyx_obj_8fastwarc_9stream_io_BytesIOStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_BytesIOStream) __PYX_ERR(2, 85, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_BytesIOStream = (struct __pyx_vtabstruct_8fastwarc_9stream_io_BytesIOStream*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_BytesIOStream->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_BytesIOStream)) __PYX_ERR(2, 85, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_FileStream = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "FileStream", sizeof(struct __pyx_obj_8fastwarc_9stream_io_FileStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_FileStream) __PYX_ERR(2, 93, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_FileStream = (struct __pyx_vtabstruct_8fastwarc_9stream_io_FileStream*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_FileStream->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_FileStream)) __PYX_ERR(2, 93, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_CompressingStream = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "CompressingStream", sizeof(struct __pyx_obj_8fastwarc_9stream_io_CompressingStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_CompressingStream) __PYX_ERR(2, 100, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_CompressingStream = (struct __pyx_vtabstruct_8fastwarc_9stream_io_CompressingStream*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_CompressingStream->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_CompressingStream)) __PYX_ERR(2, 100, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_GZipStream = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "GZipStream", sizeof(struct __pyx_obj_8fastwarc_9stream_io_GZipStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_GZipStream) __PYX_ERR(2, 105, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_GZipStream = (struct __pyx_vtabstruct_8fastwarc_9stream_io_GZipStream*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_GZipStream->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_GZipStream)) __PYX_ERR(2, 105, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_LZ4Stream = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "LZ4Stream", sizeof(struct __pyx_obj_8fastwarc_9stream_io_LZ4Stream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_LZ4Stream) __PYX_ERR(2, 121, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_LZ4Stream = (struct __pyx_vtabstruct_8fastwarc_9stream_io_LZ4Stream*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_LZ4Stream->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_LZ4Stream)) __PYX_ERR(2, 121, __pyx_L1_error)
  __pyx_ptype_8fastwarc_9stream_io_BufferedReader = __Pyx_ImportType(__pyx_t_1, "fastwarc.stream_io", "BufferedReader", sizeof(struct __pyx_obj_8fastwarc_9stream_io_BufferedReader), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8fastwarc_9stream_io_BufferedReader) __PYX_ERR(2, 134, __pyx_L1_error)
  __pyx_vtabptr_8fastwarc_9stream_io_BufferedReader = (struct __pyx_vtabstruct_8fastwarc_9stream_io_BufferedReader*)__Pyx_GetVtable(__pyx_ptype_8fastwarc_9stream_io_BufferedReader->tp_dict); if (unlikely(!__pyx_vtabptr_8fastwarc_9stream_io_BufferedReader)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initwarc(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initwarc(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_warc(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_warc(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_warc(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'warc' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_warc(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("warc", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_fastwarc__warc) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "fastwarc.warc")) {
      if (unlikely(PyDict_SetItemString(modules, "fastwarc.warc", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "fastwarc/warc.pyx":22
 * from libcpp.vector cimport vector
 * 
 * import base64             # <<<<<<<<<<<<<<
 * from datetime import datetime
 * import hashlib
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_base64, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_base64, __pyx_t_1) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":23
 * 
 * import base64
 * from datetime import datetime             # <<<<<<<<<<<<<<
 * import hashlib
 * from typing import Iterable, List, Optional, Tuple
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_datetime);
  __Pyx_GIVEREF(__pyx_n_s_datetime);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_datetime);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_datetime, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_1) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":24
 * import base64
 * from datetime import datetime
 * import hashlib             # <<<<<<<<<<<<<<
 * from typing import Iterable, List, Optional, Tuple
 * import uuid
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hashlib, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hashlib, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":25
 * from datetime import datetime
 * import hashlib
 * from typing import Iterable, List, Optional, Tuple             # <<<<<<<<<<<<<<
 * import uuid
 * import warnings
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Iterable);
  __Pyx_GIVEREF(__pyx_n_s_Iterable);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Iterable);
  __Pyx_INCREF(__pyx_n_s_List);
  __Pyx_GIVEREF(__pyx_n_s_List);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_List);
  __Pyx_INCREF(__pyx_n_s_Optional);
  __Pyx_GIVEREF(__pyx_n_s_Optional);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_Optional);
  __Pyx_INCREF(__pyx_n_s_Tuple);
  __Pyx_GIVEREF(__pyx_n_s_Tuple);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_Tuple);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Iterable, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_List); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_List, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Optional, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Tuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Tuple, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":26
 * import hashlib
 * from typing import Iterable, List, Optional, Tuple
 * import uuid             # <<<<<<<<<<<<<<
 * import warnings
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_uuid, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid, __pyx_t_1) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":27
 * from typing import Iterable, List, Optional, Tuple
 * import uuid
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * from resiliparse_inc.cctype cimport isspace, tolower
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fastwarc/warc.pyx":34
 * 
 * from fastwarc.stream_io cimport BufferedReader, BytesIOStream, CompressingStream, IOStream, PythonIOStreamAdapter
 * from fastwarc.stream_io import StreamError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_StreamError);
  __Pyx_GIVEREF(__pyx_n_s_StreamError);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_StreamError);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fastwarc_stream_io, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_StreamError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StreamError, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":101
 * 
 * 
 * class CaseInsensitiveStr(str):             # <<<<<<<<<<<<<<
 *     """Case insensitive str implementation for use as dict key."""
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyUnicode_Type)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyUnicode_Type)));
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_CaseInsensitiveStr, __pyx_n_s_CaseInsensitiveStr, (PyObject *) NULL, __pyx_n_s_fastwarc_warc, __pyx_kp_s_Case_insensitive_str_implementat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "fastwarc/warc.pyx":104
 *     """Case insensitive str implementation for use as dict key."""
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.casefold())
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_18CaseInsensitiveStr_1__hash__, 0, __pyx_n_s_CaseInsensitiveStr___hash, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_hash, __pyx_t_4) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastwarc/warc.pyx":107
 *         return hash(self.casefold())
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         return self.casefold() == other.casefold()
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_18CaseInsensitiveStr_3__eq__, 0, __pyx_n_s_CaseInsensitiveStr___eq, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_eq, __pyx_t_4) < 0) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fastwarc/warc.pyx":101
 * 
 * 
 * class CaseInsensitiveStr(str):             # <<<<<<<<<<<<<<
 *     """Case insensitive str implementation for use as dict key."""
 * 
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_CaseInsensitiveStr, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CaseInsensitiveStr, __pyx_t_4) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":111
 * 
 * 
 * class CaseInsensitiveStrDict(dict):             # <<<<<<<<<<<<<<
 *     """Case-insensitive str dict."""
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)(&PyDict_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyDict_Type)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyDict_Type)));
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_CaseInsensitiveStrDict, __pyx_n_s_CaseInsensitiveStrDict, (PyObject *) NULL, __pyx_n_s_fastwarc_warc, __pyx_kp_s_Case_insensitive_str_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "fastwarc/warc.pyx":114
 *     """Case-insensitive str dict."""
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_1__getitem__, 0, __pyx_n_s_CaseInsensitiveStrDict___getitem, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_getitem, __pyx_t_5) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":117
 *         return super().__getitem__(CaseInsensitiveStr(key))
 * 
 *     def __setitem__(self, key, value):             # <<<<<<<<<<<<<<
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_3__setitem__, 0, __pyx_n_s_CaseInsensitiveStrDict___setitem, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_setitem, __pyx_t_5) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":120
 *         super().__setitem__(CaseInsensitiveStr(key), value)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_5__contains__, 0, __pyx_n_s_CaseInsensitiveStrDict___contain, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_contains, __pyx_t_5) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":123
 *         return super().__contains__(CaseInsensitiveStr(key))
 * 
 *     def get(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_7get, 0, __pyx_n_s_CaseInsensitiveStrDict_get, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__30);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_get, __pyx_t_5) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":126
 *         return super().get(CaseInsensitiveStr(key), value)
 * 
 *     def setdefault(self, key, value=None):             # <<<<<<<<<<<<<<
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_9setdefault, 0, __pyx_n_s_CaseInsensitiveStrDict_setdefaul, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__33);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_setdefault, __pyx_t_5) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":129
 *         return super().setdefault(CaseInsensitiveStr(key), value)
 * 
 *     def pop(self, key):             # <<<<<<<<<<<<<<
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_11pop, 0, __pyx_n_s_CaseInsensitiveStrDict_pop, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_pop, __pyx_t_5) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":132
 *         return super().pop(CaseInsensitiveStr(key))
 * 
 *     def update(self, it=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if hasattr(it, 'items'):
 *             it = it.items()
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8fastwarc_4warc_22CaseInsensitiveStrDict_13update, 0, __pyx_n_s_CaseInsensitiveStrDict_update, NULL, __pyx_n_s_fastwarc_warc, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__38);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_update, __pyx_t_5) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fastwarc/warc.pyx":111
 * 
 * 
 * class CaseInsensitiveStrDict(dict):             # <<<<<<<<<<<<<<
 *     """Case-insensitive str dict."""
 * 
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_CaseInsensitiveStrDict, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_4, __pyx_t_5) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CaseInsensitiveStrDict, __pyx_t_5) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fastwarc/warc.pyx":389
 *         return self._stream_pos
 * 
 *     cpdef void init_headers(self, size_t content_length, WarcRecordType record_type=no_type, bytes record_urn=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize mandatory headers in a fresh :class:`WarcRecord` instance.
 */
  __pyx_k__6 = __pyx_e_8fastwarc_4warc_no_type;
  __pyx_k__6 = __pyx_e_8fastwarc_4warc_no_type;

  /* "fastwarc/warc.pyx":671
 *         """
 * 
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,             # <<<<<<<<<<<<<<
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,
 *                   func_filter=None, bint verify_digests=False):
 */
  __pyx_k__13 = __pyx_e_8fastwarc_4warc_any_type;

  /* "fastwarc/warc.pyx":672
 * 
 *     def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,
 *                   size_t min_content_length=strnpos, size_t max_content_length=strnpos,             # <<<<<<<<<<<<<<
 *                   func_filter=None, bint verify_digests=False):
 *         self._set_stream(stream)
 */
  __pyx_k__14 = std::string::npos;
  __pyx_k__15 = std::string::npos;

  /* "fastwarc/warc.pyx":1
 * # Copyright 2021 Janek Bevendorff             # <<<<<<<<<<<<<<
 * #
 * # Licensed under the Apache License, Version 2.0 (the "License");
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */
  __pyx_t_6 = ((PY_VERSION_HEX >= 0x02070000) != 0);
  if (__pyx_t_6) {

    /* "EnumBase":10
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:
 *     from collections import OrderedDict as __Pyx_OrderedDict             # <<<<<<<<<<<<<<
 * else:
 *     __Pyx_OrderedDict = dict
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_OrderedDict);
    __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderedDict);
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XGOTREF(__Pyx_OrderedDict);
    __Pyx_DECREF_SET(__Pyx_OrderedDict, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */
    goto __pyx_L2;
  }

  /* "EnumBase":12
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 *     __Pyx_OrderedDict = dict             # <<<<<<<<<<<<<<
 * 
 * @cython.internal
 */
  /*else*/ {
    __Pyx_INCREF(((PyObject *)(&PyDict_Type)));
    __Pyx_XGOTREF(__Pyx_OrderedDict);
    __Pyx_DECREF_SET(__Pyx_OrderedDict, ((PyObject *)(&PyDict_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyDict_Type)));
  }
  __pyx_L2:;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyInt_Type)));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Pyx_EnumBase, __pyx_n_s_Pyx_EnumBase, (PyObject *) NULL, __pyx_n_s_EnumBase, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "EnumBase":27
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta             # <<<<<<<<<<<<<<
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 */
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_metaclass, ((PyObject *)__pyx_ptype___Pyx_EnumMeta)) < 0) __PYX_ERR(1, 27, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Pyx_EnumBase___new, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__41);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_new, __pyx_t_5) < 0) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__, 0, __pyx_n_s_Pyx_EnumBase___repr, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_repr, __pyx_t_5) < 0) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__, 0, __pyx_n_s_Pyx_EnumBase___str, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_str, __pyx_t_5) < 0) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Pyx_EnumBase, __pyx_t_1, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XGOTREF(__Pyx_EnumBase);
  __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  __pyx_t_6 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_6) {

    /* "EnumBase":45
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 *     from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_IntEnum);
    __Pyx_GIVEREF(__pyx_n_s_IntEnum);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_IntEnum);
    __pyx_t_2 = __Pyx_Import(__pyx_n_s_enum, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XGOTREF(__Pyx_EnumBase);
    __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, NULL, __pyx_n_s_EnumBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     WarcRecordType = __Pyx_EnumBase('WarcRecordType', __Pyx_OrderedDict([
 */
  __pyx_t_6 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_6) {

    /* "EnumType":54
 * 
 *     WarcRecordType = __Pyx_EnumBase('WarcRecordType', __Pyx_OrderedDict([
 *         ('warcinfo', warcinfo),             # <<<<<<<<<<<<<<
 *         ('response', response),
 *         ('resource', resource),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_warcinfo); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_warcinfo);
    __Pyx_GIVEREF(__pyx_n_s_warcinfo);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_warcinfo);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":55
 *     WarcRecordType = __Pyx_EnumBase('WarcRecordType', __Pyx_OrderedDict([
 *         ('warcinfo', warcinfo),
 *         ('response', response),             # <<<<<<<<<<<<<<
 *         ('resource', resource),
 *         ('request', request),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_response); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_response);
    __Pyx_GIVEREF(__pyx_n_s_response);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_response);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":56
 *         ('warcinfo', warcinfo),
 *         ('response', response),
 *         ('resource', resource),             # <<<<<<<<<<<<<<
 *         ('request', request),
 *         ('metadata', metadata),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_resource); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_resource);
    __Pyx_GIVEREF(__pyx_n_s_resource);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_resource);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":57
 *         ('response', response),
 *         ('resource', resource),
 *         ('request', request),             # <<<<<<<<<<<<<<
 *         ('metadata', metadata),
 *         ('revisit', revisit),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_request); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_request);
    __Pyx_GIVEREF(__pyx_n_s_request);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_request);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":58
 *         ('resource', resource),
 *         ('request', request),
 *         ('metadata', metadata),             # <<<<<<<<<<<<<<
 *         ('revisit', revisit),
 *         ('conversion', conversion),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_metadata); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_n_s_metadata);
    __Pyx_GIVEREF(__pyx_n_s_metadata);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_metadata);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":59
 *         ('request', request),
 *         ('metadata', metadata),
 *         ('revisit', revisit),             # <<<<<<<<<<<<<<
 *         ('conversion', conversion),
 *         ('continuation', continuation),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_revisit); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_revisit);
    __Pyx_GIVEREF(__pyx_n_s_revisit);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_revisit);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":60
 *         ('metadata', metadata),
 *         ('revisit', revisit),
 *         ('conversion', conversion),             # <<<<<<<<<<<<<<
 *         ('continuation', continuation),
 *         ('unknown', unknown),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_conversion); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_conversion);
    __Pyx_GIVEREF(__pyx_n_s_conversion);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_conversion);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":61
 *         ('revisit', revisit),
 *         ('conversion', conversion),
 *         ('continuation', continuation),             # <<<<<<<<<<<<<<
 *         ('unknown', unknown),
 *         ('any_type', any_type),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_continuation); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_continuation);
    __Pyx_GIVEREF(__pyx_n_s_continuation);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_continuation);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":62
 *         ('conversion', conversion),
 *         ('continuation', continuation),
 *         ('unknown', unknown),             # <<<<<<<<<<<<<<
 *         ('any_type', any_type),
 *         ('no_type', no_type),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_unknown); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_n_s_unknown);
    __Pyx_GIVEREF(__pyx_n_s_unknown);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_unknown);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":63
 *         ('continuation', continuation),
 *         ('unknown', unknown),
 *         ('any_type', any_type),             # <<<<<<<<<<<<<<
 *         ('no_type', no_type),
 *     ]))
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_any_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_any_type);
    __Pyx_GIVEREF(__pyx_n_s_any_type);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_any_type);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":64
 *         ('unknown', unknown),
 *         ('any_type', any_type),
 *         ('no_type', no_type),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['warcinfo'] = WarcRecordType.warcinfo
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_no_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_no_type);
    __Pyx_GIVEREF(__pyx_n_s_no_type);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_no_type);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     WarcRecordType = __Pyx_EnumBase('WarcRecordType', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('warcinfo', warcinfo),
 *         ('response', response),
 */
    __pyx_t_2 = PyList_New(11); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_2, 3, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyList_SET_ITEM(__pyx_t_2, 4, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_2, 5, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_2, 6, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_2, 7, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyList_SET_ITEM(__pyx_t_2, 8, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyList_SET_ITEM(__pyx_t_2, 9, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyList_SET_ITEM(__pyx_t_2, 10, __pyx_t_13);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_WarcRecordType);
    __Pyx_GIVEREF(__pyx_n_s_WarcRecordType);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_WarcRecordType);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_13);
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_2, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_WarcRecordType, __pyx_t_13) < 0) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":66
 *         ('no_type', no_type),
 *     ]))
 *     __Pyx_globals['warcinfo'] = WarcRecordType.warcinfo             # <<<<<<<<<<<<<<
 *     __Pyx_globals['response'] = WarcRecordType.response
 *     __Pyx_globals['resource'] = WarcRecordType.resource
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_warcinfo); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_warcinfo, __pyx_t_2) < 0)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":67
 *     ]))
 *     __Pyx_globals['warcinfo'] = WarcRecordType.warcinfo
 *     __Pyx_globals['response'] = WarcRecordType.response             # <<<<<<<<<<<<<<
 *     __Pyx_globals['resource'] = WarcRecordType.resource
 *     __Pyx_globals['request'] = WarcRecordType.request
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_response); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 67, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_response, __pyx_t_13) < 0)) __PYX_ERR(1, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":68
 *     __Pyx_globals['warcinfo'] = WarcRecordType.warcinfo
 *     __Pyx_globals['response'] = WarcRecordType.response
 *     __Pyx_globals['resource'] = WarcRecordType.resource             # <<<<<<<<<<<<<<
 *     __Pyx_globals['request'] = WarcRecordType.request
 *     __Pyx_globals['metadata'] = WarcRecordType.metadata
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_resource); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 68, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_resource, __pyx_t_2) < 0)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":69
 *     __Pyx_globals['response'] = WarcRecordType.response
 *     __Pyx_globals['resource'] = WarcRecordType.resource
 *     __Pyx_globals['request'] = WarcRecordType.request             # <<<<<<<<<<<<<<
 *     __Pyx_globals['metadata'] = WarcRecordType.metadata
 *     __Pyx_globals['revisit'] = WarcRecordType.revisit
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_request); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 69, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_request, __pyx_t_13) < 0)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":70
 *     __Pyx_globals['resource'] = WarcRecordType.resource
 *     __Pyx_globals['request'] = WarcRecordType.request
 *     __Pyx_globals['metadata'] = WarcRecordType.metadata             # <<<<<<<<<<<<<<
 *     __Pyx_globals['revisit'] = WarcRecordType.revisit
 *     __Pyx_globals['conversion'] = WarcRecordType.conversion
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_metadata); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 70, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_metadata, __pyx_t_2) < 0)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":71
 *     __Pyx_globals['request'] = WarcRecordType.request
 *     __Pyx_globals['metadata'] = WarcRecordType.metadata
 *     __Pyx_globals['revisit'] = WarcRecordType.revisit             # <<<<<<<<<<<<<<
 *     __Pyx_globals['conversion'] = WarcRecordType.conversion
 *     __Pyx_globals['continuation'] = WarcRecordType.continuation
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_revisit); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 71, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_revisit, __pyx_t_13) < 0)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":72
 *     __Pyx_globals['metadata'] = WarcRecordType.metadata
 *     __Pyx_globals['revisit'] = WarcRecordType.revisit
 *     __Pyx_globals['conversion'] = WarcRecordType.conversion             # <<<<<<<<<<<<<<
 *     __Pyx_globals['continuation'] = WarcRecordType.continuation
 *     __Pyx_globals['unknown'] = WarcRecordType.unknown
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_conversion); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 72, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_conversion, __pyx_t_2) < 0)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":73
 *     __Pyx_globals['revisit'] = WarcRecordType.revisit
 *     __Pyx_globals['conversion'] = WarcRecordType.conversion
 *     __Pyx_globals['continuation'] = WarcRecordType.continuation             # <<<<<<<<<<<<<<
 *     __Pyx_globals['unknown'] = WarcRecordType.unknown
 *     __Pyx_globals['any_type'] = WarcRecordType.any_type
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_continuation); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 73, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_continuation, __pyx_t_13) < 0)) __PYX_ERR(1, 73, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":74
 *     __Pyx_globals['conversion'] = WarcRecordType.conversion
 *     __Pyx_globals['continuation'] = WarcRecordType.continuation
 *     __Pyx_globals['unknown'] = WarcRecordType.unknown             # <<<<<<<<<<<<<<
 *     __Pyx_globals['any_type'] = WarcRecordType.any_type
 *     __Pyx_globals['no_type'] = WarcRecordType.no_type
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_unknown); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 74, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_unknown, __pyx_t_2) < 0)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":75
 *     __Pyx_globals['continuation'] = WarcRecordType.continuation
 *     __Pyx_globals['unknown'] = WarcRecordType.unknown
 *     __Pyx_globals['any_type'] = WarcRecordType.any_type             # <<<<<<<<<<<<<<
 *     __Pyx_globals['no_type'] = WarcRecordType.no_type
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_any_type); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 75, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_any_type, __pyx_t_13) < 0)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":76
 *     __Pyx_globals['unknown'] = WarcRecordType.unknown
 *     __Pyx_globals['any_type'] = WarcRecordType.any_type
 *     __Pyx_globals['no_type'] = WarcRecordType.no_type             # <<<<<<<<<<<<<<
 * else:
 *     class WarcRecordType(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_no_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 76, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_no_type, __pyx_t_2) < 0)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     WarcRecordType = __Pyx_EnumBase('WarcRecordType', __Pyx_OrderedDict([
 */
    goto __pyx_L4;
  }

  /* "EnumType":78
 *     __Pyx_globals['no_type'] = WarcRecordType.no_type
 * else:
 *     class WarcRecordType(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['warcinfo'] = WarcRecordType(warcinfo, 'warcinfo')
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __Pyx_EnumBase);
    __pyx_t_13 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = __Pyx_Py3MetaclassPrepare(__pyx_t_13, __pyx_t_2, __pyx_n_s_WarcRecordType, __pyx_n_s_WarcRecordType, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = __Pyx_Py3ClassCreate(__pyx_t_13, __pyx_n_s_WarcRecordType, __pyx_t_2, __pyx_t_12, NULL, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_WarcRecordType, __pyx_t_11) < 0) __PYX_ERR(1, 78, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":80
 *     class WarcRecordType(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['warcinfo'] = WarcRecordType(warcinfo, 'warcinfo')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['response'] = WarcRecordType(response, 'response')
 *     __Pyx_globals['resource'] = WarcRecordType(resource, 'resource')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_warcinfo); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_warcinfo);
    __Pyx_GIVEREF(__pyx_n_s_warcinfo);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_n_s_warcinfo);
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 80, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_warcinfo, __pyx_t_13) < 0)) __PYX_ERR(1, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":81
 *         pass
 *     __Pyx_globals['warcinfo'] = WarcRecordType(warcinfo, 'warcinfo')
 *     __Pyx_globals['response'] = WarcRecordType(response, 'response')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['resource'] = WarcRecordType(resource, 'resource')
 *     __Pyx_globals['request'] = WarcRecordType(request, 'request')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_response); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_response);
    __Pyx_GIVEREF(__pyx_n_s_response);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_response);
    __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 81, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_response, __pyx_t_12) < 0)) __PYX_ERR(1, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":82
 *     __Pyx_globals['warcinfo'] = WarcRecordType(warcinfo, 'warcinfo')
 *     __Pyx_globals['response'] = WarcRecordType(response, 'response')
 *     __Pyx_globals['resource'] = WarcRecordType(resource, 'resource')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['request'] = WarcRecordType(request, 'request')
 *     __Pyx_globals['metadata'] = WarcRecordType(metadata, 'metadata')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_resource); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_resource);
    __Pyx_GIVEREF(__pyx_n_s_resource);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_n_s_resource);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 82, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_resource, __pyx_t_2) < 0)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":83
 *     __Pyx_globals['response'] = WarcRecordType(response, 'response')
 *     __Pyx_globals['resource'] = WarcRecordType(resource, 'resource')
 *     __Pyx_globals['request'] = WarcRecordType(request, 'request')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['metadata'] = WarcRecordType(metadata, 'metadata')
 *     __Pyx_globals['revisit'] = WarcRecordType(revisit, 'revisit')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_request); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_request);
    __Pyx_GIVEREF(__pyx_n_s_request);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_n_s_request);
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 83, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_request, __pyx_t_13) < 0)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":84
 *     __Pyx_globals['resource'] = WarcRecordType(resource, 'resource')
 *     __Pyx_globals['request'] = WarcRecordType(request, 'request')
 *     __Pyx_globals['metadata'] = WarcRecordType(metadata, 'metadata')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['revisit'] = WarcRecordType(revisit, 'revisit')
 *     __Pyx_globals['conversion'] = WarcRecordType(conversion, 'conversion')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_metadata); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_metadata);
    __Pyx_GIVEREF(__pyx_n_s_metadata);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_metadata);
    __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 84, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_metadata, __pyx_t_12) < 0)) __PYX_ERR(1, 84, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":85
 *     __Pyx_globals['request'] = WarcRecordType(request, 'request')
 *     __Pyx_globals['metadata'] = WarcRecordType(metadata, 'metadata')
 *     __Pyx_globals['revisit'] = WarcRecordType(revisit, 'revisit')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['conversion'] = WarcRecordType(conversion, 'conversion')
 *     __Pyx_globals['continuation'] = WarcRecordType(continuation, 'continuation')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_revisit); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_revisit);
    __Pyx_GIVEREF(__pyx_n_s_revisit);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_n_s_revisit);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 85, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_revisit, __pyx_t_2) < 0)) __PYX_ERR(1, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":86
 *     __Pyx_globals['metadata'] = WarcRecordType(metadata, 'metadata')
 *     __Pyx_globals['revisit'] = WarcRecordType(revisit, 'revisit')
 *     __Pyx_globals['conversion'] = WarcRecordType(conversion, 'conversion')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['continuation'] = WarcRecordType(continuation, 'continuation')
 *     __Pyx_globals['unknown'] = WarcRecordType(unknown, 'unknown')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_conversion); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_conversion);
    __Pyx_GIVEREF(__pyx_n_s_conversion);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_n_s_conversion);
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 86, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_conversion, __pyx_t_13) < 0)) __PYX_ERR(1, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":87
 *     __Pyx_globals['revisit'] = WarcRecordType(revisit, 'revisit')
 *     __Pyx_globals['conversion'] = WarcRecordType(conversion, 'conversion')
 *     __Pyx_globals['continuation'] = WarcRecordType(continuation, 'continuation')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['unknown'] = WarcRecordType(unknown, 'unknown')
 *     __Pyx_globals['any_type'] = WarcRecordType(any_type, 'any_type')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_continuation); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_continuation);
    __Pyx_GIVEREF(__pyx_n_s_continuation);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_continuation);
    __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 87, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_continuation, __pyx_t_12) < 0)) __PYX_ERR(1, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":88
 *     __Pyx_globals['conversion'] = WarcRecordType(conversion, 'conversion')
 *     __Pyx_globals['continuation'] = WarcRecordType(continuation, 'continuation')
 *     __Pyx_globals['unknown'] = WarcRecordType(unknown, 'unknown')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['any_type'] = WarcRecordType(any_type, 'any_type')
 *     __Pyx_globals['no_type'] = WarcRecordType(no_type, 'no_type')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_unknown); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_unknown);
    __Pyx_GIVEREF(__pyx_n_s_unknown);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_n_s_unknown);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 88, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_unknown, __pyx_t_2) < 0)) __PYX_ERR(1, 88, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":89
 *     __Pyx_globals['continuation'] = WarcRecordType(continuation, 'continuation')
 *     __Pyx_globals['unknown'] = WarcRecordType(unknown, 'unknown')
 *     __Pyx_globals['any_type'] = WarcRecordType(any_type, 'any_type')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['no_type'] = WarcRecordType(no_type, 'no_type')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_any_type); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_any_type);
    __Pyx_GIVEREF(__pyx_n_s_any_type);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_n_s_any_type);
    __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 89, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_any_type, __pyx_t_13) < 0)) __PYX_ERR(1, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":90
 *     __Pyx_globals['unknown'] = WarcRecordType(unknown, 'unknown')
 *     __Pyx_globals['any_type'] = WarcRecordType(any_type, 'any_type')
 *     __Pyx_globals['no_type'] = WarcRecordType(no_type, 'no_type')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_WarcRecordType); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(__pyx_e_8fastwarc_4warc_no_type); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_no_type);
    __Pyx_GIVEREF(__pyx_n_s_no_type);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_no_type);
    __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_2, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 90, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_no_type, __pyx_t_12) < 0)) __PYX_ERR(1, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  }
  __pyx_L4:;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init fastwarc.warc", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init fastwarc.warc");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* GeneratorYieldFrom */
static void __PyxPyIter_CheckErrorAndDecref(PyObject *source) {
    PyErr_Format(PyExc_TypeError,
                 "iter() returned non-iterator of type '%.100s'",
                 Py_TYPE(source)->tp_name);
    Py_DECREF(source);
}
static CYTHON_INLINE PyObject* __Pyx_Generator_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *source_gen, *retval;
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(source)) {
        Py_INCREF(source);
        source_gen = source;
        retval = __Pyx_Generator_Next(source);
    } else
#endif
    {
#if CYTHON_USE_TYPE_SLOTS
        if (likely(Py_TYPE(source)->tp_iter)) {
            source_gen = Py_TYPE(source)->tp_iter(source);
            if (unlikely(!source_gen))
                return NULL;
            if (unlikely(!PyIter_Check(source_gen))) {
                __PyxPyIter_CheckErrorAndDecref(source_gen);
                return NULL;
            }
        } else
#endif
        {
            source_gen = PyObject_GetIter(source);
            if (unlikely(!source_gen))
                return NULL;
        }
#if CYTHON_USE_TYPE_SLOTS
        retval = Py_TYPE(source_gen)->tp_iternext(source_gen);
#else
        retval = PyIter_Next(source_gen);
#endif
    }
    if (likely(retval)) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CyFunctionClassCell */
static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj) {
    Py_ssize_t i, count = PyList_GET_SIZE(cyfunctions);
    for (i = 0; i < count; i++) {
        __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *)
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyList_GET_ITEM(cyfunctions, i);
#else
            PySequence_ITEM(cyfunctions, i);
        if (unlikely(!m))
            return -1;
#endif
        Py_INCREF(classobj);
        m->func_classobj = classobj;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF((PyObject*)m);
#endif
    }
    return 0;
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE enum __pyx_t_8fastwarc_4warc_WarcRecordType __Pyx_PyInt_As_enum____pyx_t_8fastwarc_4warc_WarcRecordType(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8fastwarc_4warc_WarcRecordType neg_one = (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1, const_zero = (enum __pyx_t_8fastwarc_4warc_WarcRecordType) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) (((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) (((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) (((((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) (((enum __pyx_t_8fastwarc_4warc_WarcRecordType)-1)*(((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) ((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) (((enum __pyx_t_8fastwarc_4warc_WarcRecordType)-1)*(((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) ((((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) (((enum __pyx_t_8fastwarc_4warc_WarcRecordType)-1)*(((((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_8fastwarc_4warc_WarcRecordType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) ((((((((((enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_8fastwarc_4warc_WarcRecordType)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8fastwarc_4warc_WarcRecordType, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_8fastwarc_4warc_WarcRecordType, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_8fastwarc_4warc_WarcRecordType val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1;
        }
    } else {
        enum __pyx_t_8fastwarc_4warc_WarcRecordType val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_8fastwarc_4warc_WarcRecordType(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_8fastwarc_4warc_WarcRecordType");
    return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_8fastwarc_4warc_WarcRecordType");
    return (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT) {
                            return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT) {
                            return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT) {
                            return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint16_t) -1;
        }
    } else {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyInt_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_8fastwarc_4warc_WarcRecordType(enum __pyx_t_8fastwarc_4warc_WarcRecordType value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_8fastwarc_4warc_WarcRecordType neg_one = (enum __pyx_t_8fastwarc_4warc_WarcRecordType) -1, const_zero = (enum __pyx_t_8fastwarc_4warc_WarcRecordType) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_8fastwarc_4warc_WarcRecordType),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
